/*
* ==============================================================================
*  Name        : BrowserControl_DataTypes.h
*  Part of     : Browser Control
*  Description : Holds all the structures, enums, constants, etc used in
*                browser control interfaces.
*  Version     :
*
*    Copyright (c) 2007,
*    All rights reserved.
* ==============================================================================
*/
#ifndef __BROWSERCONTROL_DATATYPES_H__
#define __BROWSERCONTROL_DATATYPES_H__

#ifdef __SHP__
#include "ImgCodec.h"
#include "types.h"
#else
#include <glib/gtypes.h>
#endif

#include <time.h>

#ifdef __SHP__
    #ifdef BUILDING_WEBKIT
        #define WEBKIT_SHP_API __declspec(dllexport)
    #else
        #define WEBKIT_SHP_API __declspec(dllimport)
    #endif
#else
    #define WEBKIT_SHP_API __attribute__((visibility("default")))
#endif

#ifndef WEBKIT_SHP_API
#define WEBKIT_SHP_API
#endif

#ifdef __SHP__
typedef int HWIN;
#else
typedef unsigned int HWIN;
typedef unsigned int HGC;
typedef bool BOOL;
#endif


/**
 * HANDLE.
 * A generic handle to represent Browser Control objects.
 * @ingroup General
 */
typedef  void *HANDLE;

typedef  guint IAPHANDLE;

/**
 * API Status enum.
 * This enumerates the status for an API call.
 * @ingroup General
 */
typedef enum
{
	ESuccess = 0,		/**< Indicates that the API succeeded. */
	EFail				/**< Indicates that the API failed. */
}ESTATUS;

/**
 * Error Codes enum.
 * This enumerates the error codes.
 * @ingroup General
 */
typedef enum
{
	_ESuccess = 0,		/**< Indicates that the API succeeded. */
	_EFail,				/**< Indicates that the API failed. */
	EInvalidHandle,		/**< The handle passed was not a valid handle. */
	ENullInputParameter, /**< The input parameter passed was not valid. */
	EInsufficientBuffer, /**< The buffer passed was not sufficient. */
	EInvalidInputParameter		/**< The input parameter passed was not a valid handle. */
}ERRORCODE;

/**
 * SDistinguishedNameInfo.
 * Structure to hold the DN (Distinguished Name) information. 
 * @ingroup General
 */
typedef struct 
{

	gchar iCountry[4];		/**< DN country string. */
	gchar iOrganizationUnit[64];	/**< DN organisation unit. */
	gchar iOrganization[64];	/**< DN organisation. */
	gchar iCommonName[64];		/**< DN common name. */
	gchar iLocality[128];		/**< DN locality. */
}SDistinguishedNameInfo;

/**
 * EPublicKeyAlg.
 * This enumerates the public key algorithms.
 * @ingroup General
 */
typedef enum
{
	ERsa, 	/**< RSA encryption. */
	EDh, 	/**< DH encryption. */
	EDsa 	/**< DSA encryption */
}EPublicKeyAlg;

/**
 * SCertInfo.
 * Structure to hold the certificate information. 
 * @ingroup General
 */
typedef struct
{
	gchar					iFingerprint[20];	/**< Certificate's fingerprint. */
	gchar					iSerialNo[32];		/**< Certificate's serial number. */
	gchar					iPublicKey[300];	/**< Certificate's public key. */
	gint					iVersionNo;			/**< Certificate's version. */
	EPublicKeyAlg			iPkAlg;				/**< Certificate's Public Key Algorithm, using the EPublicKeyAlg enum. */
	time_t					iStartValDate;		/**< Certificate's start validity date. */
	time_t					iEndValDate;		/**< Certificate's end validity date. */
	SDistinguishedNameInfo	iSubjectDNInfo;		/**< Subject DN Information. */
	SDistinguishedNameInfo	iIssuerDNInfo;		/**< Issuer DN Information. */
	gchar					iDNSName[128];		/**< DNS name. */
	gint					iDigAlg;              
}SCertInfo;

/**
 * EBrowserParameters.
 * This enumerates the browser paramaeters that can be set/retrieved.
 * @ingroup General
 */
typedef enum
{
	EHandleContentTypes,	/**< To set the content types to be handled by the browser core. */
	ERequestHeaders 		/**< To set the request headers to be added to each HTTP header. */
}EBrowserParameters;


//------------------------------------------------------------------------
// Settings API
//------------------------------------------------------------------------


/**
 * EAuthType.
 * This enumerates the authentication types for proxy.
 * @ingroup Settings
 */
typedef enum
{
	ENTLM,			/**< NTLM Authentication. */
	HTTP,			/**< HTTP Authentication. */
	SOCKS4,			/**< SOCK4 Authentication. */
	SOCKS5			/**< SOCK5 Authentication. */
}EAuthType;

/**
 * SProxyInfo.
 * This enumerates the status for an API call.
 * @ingroup Settings
 */
typedef struct
{
	gchar*			server_name;		/**< Proxy server name. */
	guint			port;				/**< Proxy port. */
	gchar*			user_name;			/**< User name. */
	gchar*			password;			/**< Passowrd. */
	EAuthType		auth_type;			/**< Proxy authentication type. */
} SProxyInfo;



/**
 * Browser settings property enum.
 * This enumerates the property flags that can be set/unset.
 * @ingroup Settings
 */
typedef enum 
{
	ESetCookies,			/**< To Enable/Disable Cookies. */
	ESetCache,				/**< To Enable/Disable Cache. */
	ESetSmartLink,			/**< To Enable/Disable SmartLinks. */
	ESetAnimationDisplay,	/**< To Enable/Disable Animation Display. */
	ESetImageSaveMode,		/**< To Enable/Disable Image saving. */
	ESetSplashHTML,			/**< To Enable/Disable Splash HTML. */
	ESetSSR,				/**< To Enable/Disable SSR. If SSR is disabled then it is Desktop view.*/
	EVirtualCursor,			/**< To Enable/Disable Virtual Cursor.*/
    EIsJavaEnabled,			/**< To Enable/Disable Java.*/
    ELoadsImagesAutomatically, /**< To Enable/Disable Loading of Images automatically .*/
    EPrivateBrowsingEnabled,/**< To Enable/Disable Private Browsing.*/
    EArePluginsEnabled,		/**< To Enable/Disable Plugin support.*/
    EIsJavaScriptEnabled,   /**< To Enable/Disable Javascript.*/
    EJavaScriptCanOpenWindowsAutomatically,/**< To Enable/Disable Opeing of windows through Javascript .*/ 
    EShouldPrintBackgrounds,/**< To Enable/Disable Background printing.*/	
    ETextAreasAreResizable, /**< To Enable/Disable Text areas resizing .*/
    EUsesDashboardBackwardCompatibilityMode, /**< To Enable/Disable Uses Dashboard Backward Compatibility Mode.*/
    ENeedsAdobeFrameReloadingQuirk,/**< To Enable/Disable Needs AdobeFrame Reloading Quirk .*/
    EIsDOMPasteAllowed,		/**< To Enable/Disable DOM Paste  .*/
    EShrinksStandaloneImagesToFit,/**< To Enable/Disable Shrinks Standalone Images To Fit .*/
    EUsesPageCache,			/**< To Enable/Disable Uses Page Cache .*/
    EShowsURLsInToolTips,	/**< To Enable/Disable Shows URLs In Tool Tips .*/
    EForceFTPDirectoryListings,/**< To Enable/Disable Force FTP Directory Listings .*/
    EDeveloperExtrasEnabled	/**< To Enable/Disable Developer Extras Enabled .*/
}EBrowserProperty;

/**
 * Browser Editable Links properties enum.
 * This enumerates the property flags of Links.
 * @ingroup Settings
 */
typedef enum 
{
    EditableLinkDefaultBehavior = 0,
    EditableLinkAlwaysLive,
    EditableLinkOnlyLiveWithShiftKey,
    EditableLinkLiveWhenNotFocused,
    EditableLinkNeverLive
}EEditableLinkBehavior ;

//------------------------------------------------------------------------
// Page API
//------------------------------------------------------------------------

/**
 * EPageItemType.
 * This enumerates the type of items a page can have.
 * @ingroup Page
 */
typedef enum 
{
	ENone					= 0x0000,	/**< No type. */
	EAll					= 0x0001,	/**< All. */
	EMultiple				= 0x0002,	/**< Multiple Items. */
	EImages					= 0x0004,	/**< Images. */
	EVideo					= 0x0008,	/**< Video. */
	ESound					= 0x0010,	/**< Sound. */
	ESmartlinksAll			= 0x0020,	/**< All Smartlinks. */
	ESmartlinkEmail			= 0x0040,	/**< Email Smartlink. */
	ESmartlinkPhoneNumber	= 0x0080,	/**< Phone Number Smartlink. */
	ESmartlinkContact		= 0x0100,	/**< Contact Smartlink. */
	EAnchor					= 0x0200	/**< Anchor Element. */
}EPageItemType;

/**
 * SFindInfo.
 * This structure holds the information needed to do a text search in a page.
 * @ingroup Page
 */
typedef struct
{
	gchar*			text;						/**< Text to be searched. */
	gboolean		from_current_position;		/**< Flag to specify whether to search from current position in the document. */
	gboolean		from_beginning;				/**< Flag to specify whether to search from beginning of the document. */
	gboolean		from_end;					/**< Flag to specify whether to search from end of the document. */
	gboolean		match_case;					/**< Flag to specify whether to match the case. */
	gboolean		match_whole_word;			/**< Flag to specify whether to match the whole word. */
#ifdef _SFB_SEARCH_STRING_IN_PAGE //Jan-16 Vamsi/Pramod
	gboolean		search_forward;				/**< Flag to specify which direction to search for. */
#endif
} SFindInfo;


#ifdef SFB_RSS_LINK //vamsi Jan 28 Rss Link Tag Support for SFB.
/**
 * SRssLink.
 * This structure holds the information about Rss Links in a page.
 * @ingroup Page
 */
typedef struct
{
	gchar*		szHref;						/**< Href Attribute. */
	gchar*		szType;						/**< Type Attribute. */
	gchar*		szTitle;					/**< Title Attribute. */

} SRssLink;
#endif

//------------------------------------------------------------------------
// Error API
//------------------------------------------------------------------------

/**
 * EErrorType.
 * This enumerates the type of possible errors.
 * @ingroup Error
 */
typedef enum 
{
	EUrlSchemeError,			/**<Url Scheme Error. */
	EUrlparsingError,		/**<Url parsing Error. */
	EHttpError,				/**<Http Error. */
	EJavaScriptError,		/**<JavaScript Error . */
	EMemoryError,			/**<Memory Error. */
	ENetworkError			/**<Network Error. */
}EErrorType;

/**
 * ErrorBase.
 * This structure is the base for all errors.
 * All error info structures must have an object of ErrorBase as their first member.
 * @ingroup Error
 */
typedef struct
{
	EErrorType error_type;	/**<Error type. */
} ErrorBase;


/**
 * UrlSchemeError.
 * Structure for url scheme error info.
 * @ingroup Error
 */
typedef struct
{
	ErrorBase error_base;	/**<Error base structure. */
	gchar* url;			/**<Url that has unsupported scheme. */
} UrlSchemeError;

/**
 * UrlparsingError.
 * Structure for Url parsing error info.
 * @ingroup Error
 */
typedef struct
{
	ErrorBase error_base;	/**<Error base structure. */
	gchar* url;			/**<Url that is malformed. */
} UrlparsingError;

/**
 * HttpError.
 * Structure for Http error info.
 * @ingroup Error
 */
typedef struct
{
	ErrorBase error_base;			/**<Error base structure. */
	guint http_error_code;   /**<Http error code. */
} HttpError;

/**
 * JavaScriptError.
 * Structure for Java script error info.
 * @ingroup Error
 */
typedef struct
{
	ErrorBase error_base;			/**<Error base structure. */
	guint line_number;		/**<Line number were javascript error occured. */
} JavaScriptError;

/**
 * MemoryError.
 * Structure for memory error info.
 * @ingroup Error
 */
typedef struct
{
	ErrorBase error_base;		/**<Error base structure. */
	/* <TODO - Add memory error code enum > */
} MemoryError;

/**
 * NetworkError.
 * Structure for network error info.
 * @ingroup Error
 */
typedef struct
{
	ErrorBase error_base;
	/* <TODO - Add network error code enum > */
} NetworkError;

/**
 * ErrorHandler.
 * Callback signature for error handler.
 * @ingroup Error
 */
//typedef void (*ErrorHandler)(gint aErrorCode, ErrorBase* aErrorInfo);

//------------------------------------------------------------------------
// Observer API
//------------------------------------------------------------------------

/**
 * LoadInfo.
 * Structure for resource load information.
 * @ingroup Observer
 */
typedef struct
{
	gchar* url;
	EPageItemType type;
	ESTATUS status;
} LoadInfo;

/**
 * LoadHandler.
 * Callback signature for load handler.
 * @ingroup Observer
 */
typedef void (*LoadHandler)(LoadInfo* aLoadInfo);

/**
 * ElementInfoBase.
 * This structure is the base for all element's info.
 * All element info structures must have an object of ElementInfoBase as their first member.
 * @ingroup Page
 */
typedef struct
{
	EPageItemType element_type;	/**<Element type. */
} ElementInfoBase;


/**
 * ImageInfo.
 * Structure for image info.
 * @ingroup Page
 */
typedef struct
{
	ElementInfoBase element_info_base;	/**<Element Info base structure. */
	gchar* url;							/**<Url of the image. */
	gint url_length;					/**<Size of Memory allocated to url. */
	gchar* extension;					/**<Extension of the image. */
	gint extension_length;				/**<Size of Memory allocated to extension. */
	gint size;							/**<Size of Image in bytes. */
} ImageInfo;

/**
 * ImageInfo.
 * Structure for image info.
 * @ingroup Page
 */
typedef struct
{
	gchar* url;							/**<Url of the image. */
	gchar* name;						/**<Name of the image. */
	const char* data;
	guint size;
} ImageSaveInfo;

/**
 * AnchorInfo.
 * Structure for anchor info.
 * @ingroup Page
 */
typedef struct
{
	ElementInfoBase element_info_base;	/**<Element Info base structure. */
	gchar* url;							/**<Url of the anchor. */
	gint length;						/**<Size of Memory allocated to url. */
	gint access_key;					/**<Access Key. */
} AnchorInfo;



#endif //__BROWSERCONTROL_DATATYPES_H__

