/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003, 2007 Apple Inc.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "interpreter.h"

#include "ExecState.h"
#include "JSGlobalObject.h"
#include "Parser.h"
#include "SavedBuiltins.h"
#include "array_object.h"
 
#include "bool_object.h"
#include "collector.h"
#include "date_object.h"
#include "debugger.h"
#include "error_object.h"
#include "function_object.h"
#include "internal.h"
#include "math_object.h"
#include "nodes.h"
#include "number_object.h"
#include "object.h"
#include "object_object.h"
#include "operations.h"
#include "regexp_object.h"
#include "runtime.h"
#include "string_object.h"
#include "types.h"
#include "value.h"
#if PLATFORM(SHP)
#include <TimeMgr.h>
#elif HAVE(SYS_TIME_H)
#include <sys/time.h>
#endif
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <wtf/Assertions.h>
#ifdef SFB_WIDGET
#include "widget_object.h"
#endif

#if !PLATFORM(WIN_OS)
#include <unistd.h>
#endif

#if defined(_SFB_PERF_LOG)
#include "SFBalProfiler.h"
#endif  

#if PLATFORM(SHP)
// TmGetTime has base year as 1900 instead of 1970
#define BASE_YEAR_OFFSET    70
// Leap years between 1900 to 1970 are 17
#define LEAP_YEAR_OFFSET	17
#endif // PLATFORM(SHP)
namespace KJS {

#if PLATFORM(SHP)
extern bool bDisableJS ;
#endif

Completion Interpreter::checkSyntax(ExecState* exec, const UString& sourceURL, int startingLineNumber, const UString& code)
{
    return checkSyntax(exec, sourceURL, startingLineNumber, code.data(), code.size());
}

Completion Interpreter::checkSyntax(ExecState* exec, const UString& sourceURL, int startingLineNumber, const UChar* code, int codeLength)
{
    JSLock lock;

    int errLine;
    UString errMsg;
    RefPtr<ProgramNode> progNode = parser().parse<ProgramNode>(sourceURL, startingLineNumber, code, codeLength, 0, &errLine, &errMsg);
    if (!progNode)
        return Completion(Throw, Error::create(exec, SyntaxError, errMsg, errLine, 0, sourceURL));
    return Completion(Normal);
}

Completion Interpreter::evaluate(ExecState* exec, const UString& sourceURL, int startingLineNumber, const UString& code, JSValue* thisV)
{
    return evaluate(exec, sourceURL, startingLineNumber, code.data(), code.size(), thisV);
}

Completion Interpreter::evaluate(ExecState* exec, const UString& sourceURL, int startingLineNumber, const UChar* code, int codeLength, JSValue* thisV)
{
#if defined(_SFB_PERF_LOG)
	SFBalProfiler sfbalProfiler( "JSCore", "Interpreter::evaluate", NULL );
#endif
    JSLock lock;
    
    JSGlobalObject* globalObject = exec->dynamicGlobalObject();

    if (globalObject->recursion() >= 20)
        return Completion(Throw, Error::create(exec, GeneralError, "Recursion too deep"));

    // parse the source code
    int sourceId;
    int errLine;
    UString errMsg;
    RefPtr<ProgramNode> progNode = parser().parse<ProgramNode>(sourceURL, startingLineNumber, code, codeLength, &sourceId, &errLine, &errMsg);
    
    // notify debugger that source has been parsed
    if (globalObject->debugger()) {
        bool cont = globalObject->debugger()->sourceParsed(exec, sourceId, sourceURL, UString(code, codeLength), startingLineNumber, errLine, errMsg);
        if (!cont)
            return Completion(Break);
    }

    // no program node means a syntax error occurred
    if (!progNode)
        return Completion(Throw, Error::create(exec, SyntaxError, errMsg, errLine, sourceId, sourceURL));
    
    exec->clearException();
    
    globalObject->incRecursion();
    
    JSObject* thisObj = globalObject;
    
    // "this" must be an object... use same rules as Function.prototype.apply()
    if (thisV && !thisV->isUndefinedOrNull())
        thisObj = thisV->toObject(exec);
    
    Completion res;
    if (exec->hadException())
        // the thisV->toObject() conversion above might have thrown an exception - if so, propagate it
        res = Completion(Throw, exec->exception());
    else {
        // execute the code

        InterpreterExecState newExec(globalObject, thisObj, progNode.get());
        JSValue* value = progNode->execute(&newExec);

#if PLATFORM(SHP)
		if (!value && KJS::bDisableJS )
		{
			res = Completion(Interrupted, NULL);
		}
		else
		{
        res = Completion(newExec.completionType(), value);
    }
#endif

    }

    globalObject->decRecursion();
    
    if (shouldPrintExceptions() && res.complType() == Throw) {
        JSLock lock;
        ExecState* exec = globalObject->globalExec();
        CString f = sourceURL.UTF8String();
        CString message = res.value()->toObject(exec)->toString(exec).UTF8String();
        int line = res.value()->toObject(exec)->get(exec, "line")->toUInt32(exec);

#if PLATFORM(SHP)
#pragma message("Need to figure out the way to print the stack on shp")
        printf("%s line %d: %s\n", f.c_str(), line, message.c_str());
#elif PLATFORM(WIN_OS)
        printf("%s line %d: %s\n", f.c_str(), line, message.c_str());
#else
        printf("[%d] %s line %d: %s\n", getpid(), f.c_str(), line, message.c_str());
#endif
    }

    return res;
}

static bool printExceptions = false;

bool Interpreter::shouldPrintExceptions()
{
    return printExceptions;
}

void Interpreter::setShouldPrintExceptions(bool print)
{
    printExceptions = print;
}
#ifdef SFB_GLOBAL_STATIC_HANDLING

void initInterpreter()
{
	printExceptions = false;
}
#endif//SFB_GLOBAL_STATIC_HANDLING

} // namespace KJS
