/*
* ==============================================================================
*  Name        : Allocator.h
*  Part of     : WEBKIT / MemMan
*  Interface   : 
*  Description : 
*  Version     : %Version%
*
* ==============================================================================
*/
#ifndef __ALLOCATOR_H__
#define __ALLOCATOR_H__

//  INCLUDES
#include "MemoryManager.h"
#include "Vector.h"


// CONSTANTS

// MACROS

// DATA TYPES
// FORWARD DECLARATIONS
enum ECMMCheckResultType;

// CLASS DECLARATION

/**
*
*  @lib memman.lib
*  @since 1.0
*/
class CAllocator
{
public:

	/**
	* create the allocator
	* @since 1.0
	* @param aSize bytes of memory to be allocated
	* @return pointer to the allocated memory
	*/
	virtual gbool Create();

	/**
	* allocate a memory buffer
	* @since 1.0
	* @param aSize bytes of memory to be allocated
	* @return pointer to the allocated memory
	*/
	virtual void* Allocate(guint32 aSize) = 0;

	/**
	* allocate a memory buffer and fill it with 0
	* @since 1.0
	* @param aElements number of elements to be allocated
	* @param aSize bytes of memory to be allocated
	* @return pointer to the allocated memory
	*/
	virtual void* Calloc(guint32 aElements, guint32 aSize) = 0;

	/**
	* realloc a memory buffer
	* @since 1.0
	* @param aSize bytes of memory to be allocated
	* @return pointer to the allocated memory
	*/
	virtual void* ReAllocate(void* aPtr, guint32 aSize) = 0;

	/**
	* free the memory
	* @since 1.0
	* @param aPtr pointer to the memory to be freed
	* @param 
	* @return 
	*/
	virtual void Free(void* aPtr) = 0;

	/**
	* Returns the available memory in bytes
	* @since 1.0
	* @param 
	* @return available memory in bytes
	*/
	virtual unsigned AvailableMemory() = 0;

    /**
	* register a memory collector, memory manager doesn't own this collector
	* @since 1.0
	* @param aCollector the collector to be registered
	* @param 
	* @return 
	*/
	void AddCollector(IMemoryCollector* aCollector);

	/**
	* set the CMM notifier, memory manager doesn't own this notifier
	* @since 1.0
	* @param aNotifier the notifier
	* @param 
	* @return 
	*/
	void SetNotifier(ICMMNotifier* aNotifier);

	/**
	* check if the memory manager is able to reserve enough memory for the coming operation.
	* @since 1.0
	* @param iRequiredSize Size to be allocated
    * @param bContiguous True if the required memory needs to be contiguous
	* @return result of prechecking, gtrue = successful
	*/
	gbool PreCheck(guint32 iRequiredSize, gbool bContiguous);

	/**
	* check if the memory manager is able to reserve enough memory without going OOM.
	* @since 1.0
	* @param iRequiredSize Size to be allocated
	* @return result of prechecking, gtrue = successful
	*/
	gbool PreCheckWithThreshold(guint32 iRequiredSize);

    /**
	* get status of memory manager
	* @since 1.0
	* @param
	* @return the status
	*/
	guint32 Status();

	/**
	* set status of memory manager
	* @since 1.0
	* @param 
	* @param
	* @return
	*/
	void SetStatus(ECMMCheckResultType aStatus);

    /**
	* Allocates memory from heap
	* @since 1.0
	* @param 
	* @return 
	*/   
	void* AllocFromHeap(guint32 sz);    

	/**
	* Returns true if OOM is sent
	* @since 1.0
	* @param 
	* @return 
	*/   
    gbool IsOOMSent();

    /**
    * Sets the Memory Manager in OOM state
    * @param
    * @return
    */
    void SetInOOMState();

    void ResetOOMState();

#ifdef SFB_GLOBAL_STATIC_HANDLING
    void DeInitState();
	void InitState();
#endif //SFB_GLOBAL_STATIC_HANDLING

protected:

	CAllocator()
    {
    }

	/**
	* Allocate a buffer
	* @since 1.0
	* @param aPtr the buffer to be allocated
	* @param
	* @return
	*/
	void* DoAlloc(guint32 aSize);

};

/**
*  A simple allocator only using User static allocating functions
*  @lib memman.lib
*  @since 1.0
*/
class CPlainAllocator : public CAllocator
{
public:
	void* Allocate(guint32 aSize) { return AllocFromHeap(aSize); }
	void* Calloc(guint32 aElements, guint32 aSize);
	void* ReAllocate(void* aPtr, guint32 aSize);
	void  Free(void* aPtr);
#ifndef NDEBUG
    unsigned AvailableMemory();
#endif // NDEBUG
};

#endif
