/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "Text.h"

#include "Document.h"
#include "ExceptionCode.h"
#include "RenderText.h"
#include "TextBreakIterator.h"

#if ENABLE(SVG)
#include "RenderSVGInlineText.h"
#endif // ENABLE(SVG)

#if defined(SHP_SUPPORT_WML)
#include "WMLDocument.h"
#include "WMLBrowserContext.h"
#include "PlatformString.h"
#include "CString.h"
#endif 

namespace WebCore {

// DOM Section 1.1.1

// ### allow having children in text nodes for entities, comments etc.

void substituteEntityReferences (String &tempString) {
    int i = 0 ;
    int size = tempString.length () ;
    String replaceValue ;
    bool isNoReplace = false ;
    while (i < size )   {
        switch (tempString[i]) {
        case 0x00c6:
            replaceValue = "&AElig;" ;
            break ;
        case 38:
            replaceValue = "&AMP;" ;
            break ;
        case 0x00c1:
            replaceValue = "&Aacute;" ;
            break ;
        case 0x00c2:
            replaceValue = "&Acirc;" ;
            break ;
        case 0x00c0:
            replaceValue = "&Agrave;" ;
            break ;
        case 0x0391:
            replaceValue = "&Alpha;" ;
            break ;
        case 0x00c5:
            replaceValue = "&Aring;" ;
            break ;
        case 0x00c3:
            replaceValue = "&Atilde;" ;
            break ;
        case 0x00c4:
            replaceValue = "&Auml;" ;
            break ;
        case 0x0392:
            replaceValue = "&Beta;" ;
            break ;
        case 0x00a9:
            replaceValue = "&COPY;" ;
            break ;
        case 0x00c7:
            replaceValue = "&Ccedil;" ;
            break ;
        case 0x03a7:
            replaceValue = "&Chi;" ;
            break ;
        case 0x2021:
            replaceValue = "&Dagger;" ;
            break ;
        case 0x0394:
            replaceValue = "&Delta;" ;
            break ;
        case 0x00d0:
            replaceValue = "&ETH;" ;
            break ;
        case 0x00c9:
            replaceValue = "&Eacute;" ;
            break ;
        case 0x00ca:
            replaceValue = "&Ecirc;" ;
            break ;
        case 0x00c8:
            replaceValue =  "&Egrave;" ;
            break ;
        case 0x0395:
            replaceValue = "&Epsilon;";
            break ;
        case 0x0397:
            replaceValue = "&Eta;" ;
            break ;
        case 0x00cb:
            replaceValue = "&Euml;" ;
            break ;
        case 62:
            replaceValue = "&GT;" ;
            break ;
        case 0x0393:
            replaceValue = "&Gamma;" ;
            break ;
        case 0x00cd:
            replaceValue = "&Iacute;" ;
            break ;
        case 0x00ce:
            replaceValue = "&Icirc;" ;
            break ;
        case 0x00cc:
            replaceValue = "&Igrave;" ;
            break ;
        case 0x0399:
            replaceValue = "&Iota;" ;
            break ;
        case 0x00cf:
            replaceValue = "&Iuml;" ;
            break ;
        case 0x039a:
            replaceValue = "&Kappa;" ;
            break ;
        case 60:
            replaceValue = "&LT;" ;
            break ;
        case 0x039b:
            replaceValue = "&Lambda;" ;
            break ;
        case 0x039c:
            replaceValue = "&Mu;" ;
            break ;
        case 0x00d1:
            replaceValue = "&Ntilde;" ;
            break ;
        case 0x039d:
            replaceValue = "&Nu;" ;
            break ;
        case 0x0152:
            replaceValue = "&OElig;" ;
            break ;
        case 0x00d3:
            replaceValue = "&Oacute;" ;
            break ;
        case 0x00d4:
            replaceValue = "&Ocirc;" ;
            break ;
        case 0x00d2:
            replaceValue = "&Ograve;" ;
            break ;
        case 0x03a9:
            replaceValue = "&Omega;" ;
            break ;
        case 0x039f:
            replaceValue = "&Omicron;" ;
            break ;
        case 0x00d8:
            replaceValue = "&Oslash;" ;
            break ;
        case 0x00d5:
            replaceValue = "&Otilde;" ;
            break ;
        case 0x00d6:
            replaceValue = "&Ouml;" ;
            break ;
        case 0x03a6:
            replaceValue = "&Phi;" ;
            break ;
        case 0x03a0:
            replaceValue = "&Pi;" ;
            break ;
        case 0x2033:
            replaceValue = "&Prime;" ;
            break ;
        case 0x03a8:
            replaceValue = "&Psi;" ;
            break ;
        case 34:
            replaceValue = "&QUOT;" ;
            break ;
        case 0x00ae:
            replaceValue = "&REG;" ;
            break ;
        case 0x03a1:
            replaceValue = "&Rho;" ;
            break ;
        case 0x0160:
            replaceValue = "&Scaron;" ;
            break ;
        case 0x03a3:
            replaceValue = "&Sigma;" ;
            break ;
        case 0x00de:
            replaceValue = "&THORN;" ;
            break ;
        case 0x03a4:
            replaceValue = "&Tau;" ;
            break ;
        case 0x0398:
            replaceValue = "&Theta;" ;
            break ;
        case 0x00da:
            replaceValue = "&Uacute;" ;
            break ;
        case 0x00db:
            replaceValue = "&Ucirc;" ;
            break ;
        case 0x00d9:
            replaceValue = "&Ugrave;" ;
            break ;
        case 0x03a5:
            replaceValue = "&Upsilon;" ;
            break ;
        case 0x00dc:
            replaceValue = "&Uuml;" ;
            break ;
        case 0x039e:
            replaceValue = "&Xi;" ;
            break ;
        case 0x00dd:
            replaceValue = "&Yacute;" ;
            break ;
        case 0x0178:
            replaceValue = "&Yuml;" ;
            break ;
        case 0x0396:
            replaceValue = "&Zeta;" ;
            break ;
        case 0x00e1:
            replaceValue = "&aacute;" ;
            break ;
        case 0x00e2:
            replaceValue = "&acirc;" ;
            break ;
        case 0x00b4:
            replaceValue = "&acute;" ;
            break ;
        case 0x00e6:
            replaceValue = "&aelig;" ;
            break ;
        case 0x00e0:
            replaceValue = "&agrave;" ;
            break ;
        case 0x2135:
            replaceValue = "&alefsym;" ;
            break ;
        case 0x03b1:
            replaceValue = "&alpha;" ;
            break ;
        case 0x2227:
            replaceValue = "&and;" ;
            break ;
        case 0x2220:
            replaceValue = "&ang;" ;
            break ;
        case 0x0027:
            replaceValue = "&apos;" ;
            break ;
        case 0x00e5:
            replaceValue = "&aring;" ;
            break ;
        case 0x2248:
            replaceValue = "&asymp;" ;
            break ;
        case 0x00e3:
            replaceValue = "&atilde;" ;
            break ;
        case 0x00e4:
            replaceValue = "&auml;" ;
            break ;
        case 0x201e:
            replaceValue = "&bdquo;" ;
            break ;
        case 0x03b2:
            replaceValue = "&beta;" ;
            break ;
        case 0x00a6:
            replaceValue = "&brvbar;" ;
            break ;
        case 0x2022:
            replaceValue = "&bull;" ;
            break ;
        case 0x2229:
            replaceValue = "&cap;" ;
            break ;
        case 0x00e7:
            replaceValue = "&ccedil;" ;
            break ;
        case 0x00b8:
            replaceValue = "&cedil;" ;
            break ;
        case 0x00a2:
            replaceValue = "&cent;" ;
            break ;
        case 0x03c7:
            replaceValue = "&chi;" ;
            break ;
        case 0x02c6:
            replaceValue = "&circ;" ;
            break ;
        case 0x2663:
            replaceValue = "&clubs;" ;
            break ;
        case 0x2245:
            replaceValue = "&cong;" ;
            break ;
        case 0x21b5:
            replaceValue = "&crarr;" ;
            break ;
        case 0x222a:
            replaceValue = "&cup;" ;
            break ;
        case 0x00a4:
            replaceValue = "&curren;" ;
            break ;
        case 0x21d3:
            replaceValue = "&dArr;" ;
            break ;
        case 0x2020:
            replaceValue = "&dagger;" ;
            break ;
        case 0x2193:
            replaceValue = "&darr;" ;
            break ;
        case 0x00b0:
            replaceValue = "&deg;" ;
            break ;
        case 0x03b4:
            replaceValue = "&delta;" ;
            break ;
        case 0x2666:
            replaceValue = "&diams;" ;
            break ;
        case 0x00f7:
            replaceValue = "&divide;" ;
            break ;
        case 0x00e9:
            replaceValue = "&eacute;" ;
            break ;
        case 0x00ea:
            replaceValue = "&ecirc;" ;
            break ;
        case 0x00e8:
            replaceValue = "&egrave;" ;
            break ;
        case 0x2205:
            replaceValue = "&empty;" ;
            break ;
        case 0x2003:
            replaceValue = "&emsp;" ;
            break ;
        case 0x2002:
            replaceValue = "&ensp;" ;
            break ;
        case 0x03b5:
            replaceValue = "&epsilon;" ;
            break ;
        case 0x2261: 
            replaceValue = "&equiv;" ;
            break ;
        case 0x03b7:
            replaceValue = "&eta;" ;
            break ;
        case 0x00f0:
            replaceValue = "&eth;" ;
            break ;
        case 0x00eb:
            replaceValue = "&euml;" ;
            break ;
        case 0x20ac:
            replaceValue = "&euro;" ;
            break ;
        case 0x2203:
            replaceValue = "&exists;" ;
            break ;
        case 0x0192:
            replaceValue = "&fnof;" ;
            break ;
        case 0x2200:
            replaceValue = "&forall;" ;
            break ;
        case 0x00bd:
            replaceValue = "&frac12;" ;
            break ;
        case 0x00bc:
            replaceValue = "&frac14;" ;
            break ;
        case 0x00be:
            replaceValue = "&frac34;" ;
            break ;
        case 0x2044:
            replaceValue = "&frasl;" ;
            break ;
        case 0x03b3:
            replaceValue = "&gamma;" ;
            break ;
        case 0x2265:
            replaceValue = "&ge;" ;
            break ;
        case 0x21d4:
            replaceValue = "&hArr;" ;
            break ;
        case 0x2194:
            replaceValue = "&harr;" ;
            break ;
        case 0x2665:
            replaceValue = "&hearts;" ;
            break ;
        case 0x2026:
            replaceValue = "&hellip;" ;
            break ;
        case 0x00ed:
            replaceValue = "&iacute;" ;
            break ;
        case 0x00ee:
            replaceValue = "&icirc;" ;
            break ;
        case 0x00a1:
            replaceValue = "&iexcl;" ;
            break ;
        case 0x00ec:
            replaceValue = "&igrave;" ;
            break ;
        case 0x2111:
            replaceValue = "&image;" ;
            break ;
        case 0x221e:
            replaceValue = "&infin;" ;
            break ;
        case 0x222b:
            replaceValue = "&int;" ;
            break ;
        case 0x03b9:
            replaceValue = "&iota;" ;
            break ;
        case 0x00bf:
            replaceValue = "&iquest;" ;
            break ;
        case 0x2208:
            replaceValue = "&isin;" ;
            break ;
        case 0x00ef:
            replaceValue = "&iuml;" ;
            break ;
        case 0x03ba:
            replaceValue = "&kappa;" ;
            break ;
        case 0x21d0:
            replaceValue = "&lArr;" ;
            break ;
        case 0x03bb:
            replaceValue = "&lambda;" ;
            break ;
        case 0x3008:
            replaceValue = "&lang;" ;
            break ;
        case 0x00ab:
            replaceValue = "&laquo;" ;
            break ;
        case 0x2190:
            replaceValue = "&larr;" ;
            break ;
        case 0x2308:
            replaceValue = "&lceil;" ;
            break ;
        case 0x201c:
            replaceValue = "&ldquo;" ;
            break ;
        case 0x2264:
            replaceValue = "&le;" ;
            break ;
        case 0x230a:
            replaceValue = "&lfloor;" ;
            break ;
        case 0x2217:
            replaceValue = "&lowast;";
            break ;
        case 0x25ca:
            replaceValue = "&loz;" ;
            break ;
        case 0x200e:
            replaceValue = "&lrm;" ;
            break ;
        case 0x2039:
            replaceValue = "&lsaquo;" ;
            break ;
        case 0x2018:
            replaceValue = "&lsquo;" ;
            break ;
        case 0x00af:
            replaceValue = "&macr;" ;
            break ;
        case 0x2014:
            replaceValue = "&mdash;" ;
            break ;
        case 0x00b5:
            replaceValue = "&micro;" ;
            break ;
        case 0x00b7:
            replaceValue = "&middot;" ;
            break ;
        case 0x2212:
            replaceValue = "&minus;" ;
            break ;
        case 0x03bc:
            replaceValue = "&mu;" ;
            break ;
        case 0x2207:
            replaceValue = "&nabla;" ;
            break ;
        case 0x00a0:
            replaceValue = "&nbsp;" ;
            break ;
        case 0x2013:
            replaceValue = "&ndash;" ;
            break ;
        case 0x2260:
            replaceValue = "&ne;" ;
            break ;
        case 0x220b:
            replaceValue = "&ni;" ;
            break ;
        case 0x00ac:
            replaceValue = "&not;" ;
            break ;
        case 0x2209:
            replaceValue = "&notin;" ;
            break ;
        case 0x2284:
            replaceValue = "&nsub;" ;
            break ;
        case 0x2285:
            replaceValue = "&nsup;" ;
            break ;
        case 0x00f1:
            replaceValue = "&ntilde;" ;
            break ;
        case 0x03bd:
            replaceValue = "&nu;" ;
            break ;
        case 0x00f3:
            replaceValue = "&oacute;" ;
            break ;
        case 0x00f4:
            replaceValue = "&ocirc;" ;
            break ;
        case 0x0153:
            replaceValue = "&oelig;" ;
            break ;
        case 0x00f2:
            replaceValue = "&ograve;" ;
            break ;
        case 0x203e:
            replaceValue = "&oline;" ;
            break ;
        case 0x03c9:
            replaceValue = "&omega;" ;
            break ;
        case 0x03bf:
            replaceValue = "&omicron;" ;
            break ;
        case 0x2295:
            replaceValue = "&oplus;" ;
            break ;
        case 0x2228:
            replaceValue = "&or;" ;
            break ;
        case 0x00aa:
            replaceValue = "&ordf;" ;
            break ;
        case 0x00ba:
            replaceValue = "&ordm;" ;
            break ;
        case 0x00f8:
            replaceValue = "&oslash;" ;
            break ;
        case 0x00f5:
            replaceValue = "&otilde;" ;
            break ;
        case 0x2297:
            replaceValue = "&otimes;" ;
            break ;
        case 0x00f6:
            replaceValue = "&ouml;" ;
            break ;
        case 0x00b6:
            replaceValue = "&para;" ;
            break ;
        case 0x2202:
            replaceValue = "&part;" ;
            break ;
        case 0x0025:
            replaceValue = "&percnt;" ;
            break ;
        case 0x2030:
            replaceValue = "&permil;" ;
            break ;
        case 0x22a5:
            replaceValue = "&perp;" ;
            break ;
        case 0x03c6:
            replaceValue = "&phi;" ;
            break ;
        case 0x03c0:
            replaceValue = "&pi;" ;
            break ;
        case 0x03d6:
            replaceValue = "&piv;" ;
            break ;
        case 0x00b1:
            replaceValue = "&plusmn;" ;
            break ;
        case 0x00a3:
            replaceValue = "&pound;" ;
            break ;
        case 0x2032:
            replaceValue = "&prime;" ;
            break ;
        case 0x220f:
            replaceValue = "&prod;" ;
            break ;
        case 0x221d:
            replaceValue = "&prop;" ;
            break ;
        case 0x03c8:
            replaceValue = "&psi;" ;
            break ;
        case 0x21d2:
            replaceValue = "&rArr;" ;
            break ;
        case 0x221a:
            replaceValue = "&radic;" ;
            break ;
        case 0x3009:
            replaceValue = "&rang;" ;
            break ;
        case 0x00bb:
            replaceValue = "&raquo;" ;
            break ;
        case 0x2192:
            replaceValue = "&rarr;" ;
            break ;
        case 0x2309:
            replaceValue = "&rceil;" ;
            break ;
        case 0x201d:
            replaceValue = "&rdquo;" ;
            break ;
        case 0x211c:
            replaceValue = "&real;" ;
            break ;
        case 0x230b:
            replaceValue = "&rfloor;" ;
            break ;
        case 0x03c1:
            replaceValue = "&rho;" ;
            break ;
        case 0x200f:
            replaceValue = "&rlm;" ;
            break ;
        case 0x203a:
            replaceValue = "&rsaquo;" ;
            break ;
        case 0x2019:
            replaceValue = "&rsquo;" ;
            break ;
        case 0x201a:
            replaceValue = "&sbquo;" ;
            break ;
        case 0x0161:
            replaceValue = "&scaron;" ;
            break ;
        case 0x22c5:
            replaceValue = "&sdot;" ;
            break ;
        case 0x00a7:
            replaceValue = "&sect;" ;
            break ;
        case 0x00ad:
            replaceValue = "&shy;" ;
            break ;
        case 0x03c3:
            replaceValue = "&sigma;" ;
            break ;
        case 0x03c2:
            replaceValue = "&sigmaf;" ;
            break ;
        case 0x223c:
            replaceValue = "&sim;" ;
            break ;
        case 0x2660:
            replaceValue = "&spades;" ;
            break ;
        case 0x2282:
            replaceValue = "&sub;" ;
            break ;
        case 0x2286:
            replaceValue = "&sube;" ;
            break ;
        case 0x2211:
            replaceValue = "&sum;" ;
            break ;
        case 0x2283:
            replaceValue = "&sup;" ;
            break ;
        case 0x00b9:
            replaceValue = "&sup1;" ;
            break ;
        case 0x00b2:
            replaceValue = "&sup2;" ;
            break ;
        case 0x00b3:
            replaceValue = "&sup3;" ;
            break ;
        case 0x2287:
            replaceValue = "&supe;" ;
            break ;
        case 0x00df:
            replaceValue = "&szlig;" ;
            break ;
        case 0x03c4:
            replaceValue = "&tau;" ;
            break ;
        case 0x2234:
            replaceValue = "&there4;" ;
            break ;
        case 0x03b8:
            replaceValue = "&theta;" ;
            break ;
        case 0x03d1:
            replaceValue = "&thetasym;" ;
            break ;
        case 0x2009:
            replaceValue = "&thinsp;" ;
            break ;
        case 0x00fe:
            replaceValue = "&thorn;" ;
            break ;
        case 0x02dc:
            replaceValue = "&tilde;" ;
            break ;
        case 0x00d7:
            replaceValue = "&times;" ;
            break ;
        case 0x2122:
            replaceValue = "&trade;" ;
            break ;
        case 0x21d1:
            replaceValue = "&uArr;" ;
            break ;
        case 0x00fa:
            replaceValue = "&uacute;" ;
            break ;
        case 0x2191:
            replaceValue = "&uarr;" ;
            break ;
        case 0x00fb:
            replaceValue = "&ucirc;" ;
            break ;
        case 0x00f9:
            replaceValue = "&ugrave;" ;
            break ;
        case 0x00a8:
            replaceValue = "&uml;" ;
            break ;
        case 0x03d2:
            replaceValue = "&upsih;" ;
            break ;
        case 0x03c5:
            replaceValue = "&upsilon;" ;
            break ;
        case 0x00fc:
            replaceValue = "&uuml;" ;
            break ;
        case 0x2118:
            replaceValue = "&weierp;" ;
            break ;
        case 0x03be:
            replaceValue = "&xi;" ;
            break ;
        case 0x00fd:
            replaceValue = "&yacute;" ;
            break ;
        case 0x00a5:
            replaceValue = "&yen;" ;
            break ;
        case 0x00ff:
            replaceValue = "&yuml;" ;
            break ;
        case 0x03b6:
            replaceValue = "&zeta;" ;
            break ;
        case 0x200d:
            replaceValue = "&zwj;" ;
            break ;
        case 0x200c:
            replaceValue = "&zwnj;" ;
            break ;
        default:
            isNoReplace = true ;
            break ;

        }
        if (isNoReplace)
            isNoReplace = false ;
        else    {
            tempString.replace(i, 1, replaceValue) ;
            size = tempString.length () ;
        }
        i ++ ;
    }
}

Text::Text(Document *doc, const String &_text)
    : CharacterData(doc, _text)
{
}

Text::Text(Document *doc)
    : CharacterData(doc)
{
}

Text::~Text()
{
}

PassRefPtr<Text> Text::splitText(unsigned offset, ExceptionCode& ec)
{
    ec = 0;

    // FIXME: This does not copy markers
    
    // INDEX_SIZE_ERR: Raised if the specified offset is negative or greater than
    // the number of 16-bit units in data.
    if (offset > m_data->length()) {
        ec = INDEX_SIZE_ERR;
        return 0;
    }

    // NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
    if (isReadOnlyNode()) {
        ec = NO_MODIFICATION_ALLOWED_ERR;
        return 0;
    }

    RefPtr<StringImpl> oldStr = m_data;
    RefPtr<Text> newText = createNew(oldStr->substring(offset));
    m_data = oldStr->substring(0, offset);

    dispatchModifiedEvent(oldStr.get());

    if (parentNode())
        parentNode()->insertBefore(newText.get(), nextSibling(), ec);
    if (ec)
        return 0;

    if (renderer())
        static_cast<RenderText*>(renderer())->setText(m_data);

    return newText.release();
}

static const Text* earliestLogicallyAdjacentTextNode(const Text* t)
{
    const Node* n = t;
    while ((n = n->previousSibling())) {
        Node::NodeType type = n->nodeType();
        if (type == Node::TEXT_NODE || type == Node::CDATA_SECTION_NODE) {
            t = static_cast<const Text*>(n);
            continue;
        }

        // We would need to visit EntityReference child text nodes if they existed
        ASSERT(type != Node::ENTITY_REFERENCE_NODE || !n->hasChildNodes());
        break;
    }
    return t;
}

static const Text* latestLogicallyAdjacentTextNode(const Text* t)
{
    const Node* n = t;
    while ((n = n->nextSibling())) {
        Node::NodeType type = n->nodeType();
        if (type == Node::TEXT_NODE || type == Node::CDATA_SECTION_NODE) {
            t = static_cast<const Text*>(n);
            continue;
        }

        // We would need to visit EntityReference child text nodes if they existed
        ASSERT(type != Node::ENTITY_REFERENCE_NODE || !n->hasChildNodes());
        break;
    }
    return t;
}

String Text::wholeText() const
{
    const Text* startText = earliestLogicallyAdjacentTextNode(this);
    const Text* endText = latestLogicallyAdjacentTextNode(this);

    Vector<UChar> result;
    Node* onePastEndText = endText->nextSibling();
    for (const Node* n = startText; n != onePastEndText; n = n->nextSibling()) {
        if (!n->isTextNode())
            continue;
        const Text* t = static_cast<const Text*>(n);
        const String& data = t->data();
        result.append(data.characters(), data.length());
    }

    return String::adopt(result);
}

PassRefPtr<Text> Text::replaceWholeText(const String& newText, ExceptionCode&)
{
    // We don't support "read-only" text nodes (no Entity node support)
    // Thus, we remove all adjacent text nodes, and replace the contents of this one.
    ASSERT(!isReadOnlyNode());
    // This method only raises exceptions when dealing with Entity nodes (which we don't support)

    // Protect startText and endText against mutation event handlers removing the last ref
    RefPtr<Text> startText = const_cast<Text*>(earliestLogicallyAdjacentTextNode(this));
    RefPtr<Text> endText = const_cast<Text*>(latestLogicallyAdjacentTextNode(this));

    RefPtr<Text> protectedThis(this); // Mutation event handlers could cause our last ref to go away
    Node* parent = parentNode(); // Protect against mutation handlers moving this node during traversal
    ExceptionCode ignored = 0;
    for (RefPtr<Node> n = startText; n && n != this && n->isTextNode() && n->parentNode() == parent;) {
        RefPtr<Node> nodeToRemove(n.release());
        n = nodeToRemove->nextSibling();
        parent->removeChild(nodeToRemove.get(), ignored);
    }

    if (this != endText) {
        Node* onePastEndText = endText->nextSibling();
        for (RefPtr<Node> n = nextSibling(); n && n != onePastEndText && n->isTextNode() && n->parentNode() == parent;) {
            RefPtr<Node> nodeToRemove(n.release());
            n = nodeToRemove->nextSibling();
            parent->removeChild(nodeToRemove.get(), ignored);
        }
    }

    if (newText.isEmpty()) {
        if (parent && parentNode() == parent)
            parent->removeChild(this, ignored);
        return 0;
    }

    setData(newText, ignored);
    return protectedThis.release();
}

String Text::nodeName() const
{
    return textAtom.domString();
}

Node::NodeType Text::nodeType() const
{
    return TEXT_NODE;
}

PassRefPtr<Node> Text::cloneNode(bool /*deep*/)
{
    return document()->createTextNode(m_data);
}

bool Text::rendererIsNeeded(RenderStyle *style)
{
    if (!CharacterData::rendererIsNeeded(style))
        return false;

    bool onlyWS = containsOnlyWhitespace();
    if (!onlyWS)
        return true;

    RenderObject *par = parentNode()->renderer();
    
    if (par->isTable() || par->isTableRow() || par->isTableSection() || par->isTableCol() || par->isFrameSet())
        return false;
    
    if (style->preserveNewline()) // pre/pre-wrap/pre-line always make renderers.
        return true;
    
    RenderObject *prev = previousRenderer();
    if (prev && prev->isBR()) // <span><br/> <br/></span>
        return false;
        
    if (par->isInlineFlow()) {
        // <span><div/> <div/></span>
        if (prev && !prev->isInline())
            return false;
    } else {
        if (par->isRenderBlock() && !par->childrenInline() && (!prev || !prev->isInline()))
            return false;
        
        RenderObject *first = par->firstChild();
        while (first && first->isFloatingOrPositioned())
            first = first->nextSibling();
        RenderObject *next = nextRenderer();
        if (!first || next == first)
            // Whitespace at the start of a block just goes away.  Don't even
            // make a render object for this text.
            return false;
    }
    
    return true;
}

RenderObject *Text::createRenderer(RenderArena *arena, RenderStyle *style)
{
#if ENABLE(SVG)
    if (parentNode()->isSVGElement())
        return new (arena) RenderSVGInlineText(this, m_data);
#endif // ENABLE(SVG)
    
#if defined(SHP_SUPPORT_WML)
	StringImpl *varName = 0 ;
	if (document()->htmlMode() == Document::Wml)	{
		varName = document()->frame()->getVariableReplacedString(m_data, WMLBrowserContext::ENoEscape)->impl() ;
	}
	if (varName)
		return new (arena) RenderText(this, varName);
#endif //SHP_SUPPORT-WML

    return new (arena) RenderText(this, m_data);
}

void Text::attach()
{
    createRendererIfNeeded();
    CharacterData::attach();
}

void Text::recalcStyle( StyleChange change )
{
    if (change != NoChange && parentNode())
        if (renderer())
            renderer()->setStyle(parentNode()->renderer()->style());
    if (changed() && renderer() && renderer()->isText())
        static_cast<RenderText*>(renderer())->setText(m_data);
    setChanged(NoStyleChange);
}

// DOM Section 1.1.1
bool Text::childTypeAllowed(NodeType)
{
    return false;
}

PassRefPtr<Text> Text::createNew(PassRefPtr<StringImpl> string)
{
    return new Text(document(), string);
}

String Text::toString() const
{
    // FIXME: substitute entity references as needed!
    return nodeValue();
}

#if PLATFORM(SHP)
bool Text::toString(HTMLSrcString &htmlSrc) const
{
    String tempString = nodeValue() ;
    substituteEntityReferences (tempString) ;
    htmlSrc.append (tempString) ;
	return true ;
}
#endif 

PassRefPtr<Text> Text::createWithLengthLimit(Document* doc, const String& text, unsigned& charsLeft, unsigned maxChars)
{
    if (charsLeft == text.length() && charsLeft <= maxChars) {
        charsLeft = 0;
        return new Text(doc, text);
    }
#if USE(SHP_UNICODE)    
	#undef min //Zafar. Here the min is defined somewhere else, so undefine it so that std::min works
#endif 
    unsigned start = text.length() - charsLeft;
# if USE(SHP_UNICODE)
    unsigned end = start + std::min(charsLeft, maxChars);
# else 
    unsigned end = start + min(charsLeft, maxChars);
# endif
    
    // check we are not on an unbreakable boundary
    TextBreakIterator* it = characterBreakIterator(text.characters(), text.length());
    if (end < text.length() && !isTextBreak(it, end))
        end = textBreakPreceding(it, end);
        
    // maxChars of unbreakable characters could lead to infinite loop
    if (end <= start)
        end = text.length();
    
    String nodeText = text.substring(start, end - start);
    charsLeft = text.length() - end;
        
    return new Text(doc, nodeText);
}

#ifndef NDEBUG
void Text::formatForDebugger(char *buffer, unsigned length) const
{
    String result;
    String s;
    
    s = nodeName();
    if (s.length() > 0) {
        result += s;
    }
          
    s = nodeValue();
    if (s.length() > 0) {
        if (result.length() > 0)
            result += "; ";
        result += "value=";
        result += s;
    }
          
    strncpy(buffer, result.deprecatedString().latin1(), length - 1);
}
#endif

} // namespace WebCore
