/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "HTMLImageElement.h"

#include "CSSHelper.h"
#include "CSSPropertyNames.h"
#include "CSSValueKeywords.h"
#include "EventNames.h"
#include "HTMLDocument.h"
#include "HTMLFormElement.h"
#include "HTMLNames.h"
#include "RenderImage.h"

#if defined (SHP_SUPPORT_WML)
#include "WMLDocument.h"
#include "WMLBrowserContext.h"
#endif //SHP_SUPPORT_WML
#include "chrome.h"
#include "renderImage.h"
#include "image.h"
#include "page.h"
using namespace std;

namespace WebCore {

using namespace EventNames;
using namespace HTMLNames;

HTMLImageElement::HTMLImageElement(Document* doc, HTMLFormElement* f)
    : HTMLElement(imgTag, doc)
    , m_imageLoader(this)
    , ismap(false)
    , m_form(f)
    , m_compositeOperator(CompositeSourceOver)
{
    if (f)
        f->registerImgElement(this);
}

HTMLImageElement::HTMLImageElement(const QualifiedName& tagName, Document* doc)
    : HTMLElement(tagName, doc)
    , m_imageLoader(this)
    , ismap(false)
    , m_form(0)
    , m_compositeOperator(CompositeSourceOver)
{
}

HTMLImageElement::~HTMLImageElement()
{
    if (m_form)
        m_form->removeImgElement(this);
}

bool HTMLImageElement::mapToEntry(const QualifiedName& attrName, MappedAttributeEntry& result) const
{
    if (attrName == widthAttr ||
        attrName == heightAttr ||
        attrName == vspaceAttr ||
        attrName == hspaceAttr ||
        attrName == valignAttr) {
        result = eUniversal;
        return false;
    }
    
    if (attrName == borderAttr || attrName == alignAttr) {
        result = eReplaced; // Shared with embed and iframe elements.
        return false;
    }

    return HTMLElement::mapToEntry(attrName, result);
}

#if PLATFORM(SHP)
bool HTMLImageElement::toString(HTMLSrcString &htmlSrc) const
{
	String tempString ;
    htmlSrc.append ("<") ;
    htmlSrc.append (Element::nodeName()) ;

	NamedAttrMap *attrMap = Element::attributes(true);

    if (attrMap) {
        unsigned numAttrs = attrMap->length();
        for (unsigned i = 0; i < numAttrs; i++) {
			htmlSrc.append (" ") ;
            Attribute *attribute = attrMap->attributeItem(i);
			htmlSrc.append (attribute->name().toString()) ;
            if (!attribute->value().isNull()) {
				htmlSrc.append ("=\"") ;
                // FIXME: substitute entities for any instances of " or '
				String hrefSrc = attribute->name().toString();
				if(hrefSrc == "src")	{
					String path = document()->getFilePath() ;

                    String fileName = document()->validateFilename (attribute->value(), path) ;
                    String absUrl = document()->completeURL(parseURL(attribute->value()));
                    bool isFileSaved = false ;
                    if (isFileSaved = document()->page()->isAddedToFileContext(absUrl))
    					fileName = document()->page()->getFileName (absUrl) ;
                    else
                        document()->page()->addSavePageList (absUrl, fileName) ;

					//now get the buffer from the renderer
					RenderImage *renderImage;
					renderImage = static_cast<RenderImage *>(renderer()) ;
					bool savedImage = false ;
                    const char* imageBuffer = 0 ;
                    int bufLength = 0 ;
					if(renderImage)	{
					    CachedImage* cachedImage;
						cachedImage = renderImage->cachedImage();
						if (cachedImage)	{
							Image* image = cachedImage->image();
                            SharedBuffer *buffer =0 ;
                            if (image)
                                buffer = image->data() ;
                            if (!buffer)    {
                                CachedResource *temp = static_cast<CachedResource*>(cachedImage) ;
                                buffer = temp->data() ;
                                //get the buffer from the cachedObject
                            }
                            if (buffer) {
                                imageBuffer = buffer->data() ;
                                bufLength = buffer->size() ;
                            }
							if (buffer)	{
								PlatformFileHandle filehandle = 0 ;
                                if (!isFileSaved)    {
							        savedImage = document()->page()->chrome()->savePageFileContext (filehandle, fileName, \
												    path, imageBuffer, bufLength, true) ;
                                }
								htmlSrc.append (fileName) ;
							}
						}
					}
					if (!savedImage)	{
						htmlSrc.append (" ") ;
					}
				}
				else	{
					tempString = attribute->value() ;
					htmlSrc.append (tempString) ;
				}
				htmlSrc.append ("\"") ;
            }
        }
		htmlSrc.append (">") ;
    }
	for (Node *child = firstChild(); child != NULL; child = child->nextSibling()) 
	{
		child->toString(htmlSrc);			
	}
	htmlSrc.append ("</") ;
	htmlSrc.append (Element::nodeName()) ;
	htmlSrc.append (">") ;
	return true ;
}
#endif // PLATFORM(SHP)

void HTMLImageElement::parseMappedAttribute(MappedAttribute* attr)
{
    const QualifiedName& attrName = attr->name();
    if (attrName == altAttr) {
        if (renderer() && renderer()->isImage())
            static_cast<RenderImage*>(renderer())->updateAltText();
    } else if (attrName == srcAttr)
        m_imageLoader.updateFromElement();
	else if (attrName == widthAttr){
#if PLATFORM(SHP)

        String attrValue = attr->value();
		//Dont allow width with negative values . eg : -100
        int percentIndex = attrValue.reverseFind('%');
		if ( -1 != percentIndex ) {
            addCSSLength(attr, CSS_PROP_WIDTH, attr->value());
		} else
        {
		    int width = attr->value().toInt();
			// SFB00003059 dont set image width to 0 if the value of the width attribute is -ve, let the image decide display width
            if (width >= 0)
                addCSSLength(attr, CSS_PROP_WIDTH, String::number(width));
        }
#else
        addCSSLength(attr, CSS_PROP_WIDTH, attr->value());
#endif
	} else if (attrName == heightAttr) {
#if PLATFORM(SHP)
        String attrValue = attr->value();
        int percentIndex = attrValue.reverseFind('%');
		if ( -1 != percentIndex ) {
		    addCSSLength(attr, CSS_PROP_HEIGHT, attr->value());
        }
        else
        {
		    //Dont allow height with negative values . eg : -100
		    int height = attr->value().toInt();
			// SFB00003059 dont set image height to 0 if the value of the height attribute is -ve, let the image decide display height
		    if (height >= 0)
			    addCSSLength(attr, CSS_PROP_HEIGHT, String::number(height));
        }
#else
		addCSSLength(attr, CSS_PROP_HEIGHT, attr->value());
#endif 
	} else if (attrName == borderAttr) {
        // border="noborder" -> border="0"
#if PLATFORM(SHP)
	 /* Please refer to bug# 20226 on https://bugs.webkit.org 
		As suggested by Dave Hyatt in his analysis for the above bug number, we trim the '%' sign 
		from the string we pass to the CSS parser.

		Border  attribute with  % unit ( e.g 10%) is not handled correctly by CSS parser and as a 
		result border width  properties (e.g left, top etc ) are not added.
	 */
		String borderValue = attr->value();
		String attrValue = attr->value();

		int percentIndex = attrValue.reverseFind('%');
		if( -1 != percentIndex )
		{		
			borderValue = attrValue.substring(0, percentIndex );
		}
			
		addCSSLength(attr, CSS_PROP_BORDER_WIDTH, attr->value().toInt() ? borderValue : "0");
		
#else
		addCSSLength(attr, CSS_PROP_BORDER_WIDTH, attr->value().toInt() ? attr->value() : "0");
#endif
       
        addCSSProperty(attr, CSS_PROP_BORDER_TOP_STYLE, CSS_VAL_SOLID);
        addCSSProperty(attr, CSS_PROP_BORDER_RIGHT_STYLE, CSS_VAL_SOLID);
        addCSSProperty(attr, CSS_PROP_BORDER_BOTTOM_STYLE, CSS_VAL_SOLID);
        addCSSProperty(attr, CSS_PROP_BORDER_LEFT_STYLE, CSS_VAL_SOLID);
    } else if (attrName == vspaceAttr) {
        addCSSLength(attr, CSS_PROP_MARGIN_TOP, attr->value());
        addCSSLength(attr, CSS_PROP_MARGIN_BOTTOM, attr->value());
    } else if (attrName == hspaceAttr) {
        addCSSLength(attr, CSS_PROP_MARGIN_LEFT, attr->value());
        addCSSLength(attr, CSS_PROP_MARGIN_RIGHT, attr->value());
    } else if (attrName == alignAttr)
        addHTMLAlignment(attr);
    else if (attrName == valignAttr)
        addCSSProperty(attr, CSS_PROP_VERTICAL_ALIGN, attr->value());
    else if (attrName == usemapAttr) {
        if (attr->value().domString()[0] == '#')
            usemap = attr->value();
        else
            usemap = document()->completeURL(parseURL(attr->value()));
        m_isLink = !attr->isNull();
    } else if (attrName == ismapAttr)
        ismap = true;
    else if (attrName == onabortAttr)
        setHTMLEventListener(abortEvent, attr);
    else if (attrName == onloadAttr)
        setHTMLEventListener(loadEvent, attr);
    else if (attrName == compositeAttr) {
        if (!parseCompositeOperator(attr->value(), m_compositeOperator))
            m_compositeOperator = CompositeSourceOver;
    } else if (attrName == nameAttr) {
        String newNameAttr = attr->value();
        if (inDocument() && document()->isHTMLDocument()) {
            HTMLDocument* doc = static_cast<HTMLDocument*>(document());
            doc->removeNamedItem(oldNameAttr);
            doc->addNamedItem(newNameAttr);
        }
        oldNameAttr = newNameAttr;
    } else if (attr->name() == idAttr) {
        String newIdAttr = attr->value();
        if (inDocument() && document()->isHTMLDocument()) {
            HTMLDocument *doc = static_cast<HTMLDocument *>(document());
            doc->removeDocExtraNamedItem(oldIdAttr);
            doc->addDocExtraNamedItem(newIdAttr);
        }
        oldIdAttr = newIdAttr;
        // also call superclass
        HTMLElement::parseMappedAttribute(attr);
    } else
        HTMLElement::parseMappedAttribute(attr);
}

String HTMLImageElement::altText() const
{
    // lets figure out the alt text.. magic stuff
    // http://www.w3.org/TR/1998/REC-html40-19980424/appendix/notes.html#altgen
    // also heavily discussed by Hixie on bugzilla
    String alt = getAttribute(altAttr);
    // fall back to title attribute
    if (alt.isNull())
        alt = getAttribute(titleAttr);
    return alt;
}

RenderObject* HTMLImageElement::createRenderer(RenderArena* arena, RenderStyle* style)
{
     if (style->contentData())
        return RenderObject::createObject(this, style);
     
     return new (arena) RenderImage(this);
}

void HTMLImageElement::attach()
{
    HTMLElement::attach();

    if (renderer() && renderer()->isImage()) {
        RenderImage* imageObj = static_cast<RenderImage*>(renderer());
        imageObj->setCachedImage(m_imageLoader.image());
        
        // If we have no image at all because we have no src attribute, set
        // image height and width for the alt text instead.
        if (!m_imageLoader.image() && !imageObj->cachedImage())
            imageObj->setImageSizeForAltText();
    }
}

void HTMLImageElement::insertedIntoDocument()
{
    if (document()->isHTMLDocument()) {
        HTMLDocument* doc = static_cast<HTMLDocument*>(document());

        doc->addNamedItem(oldNameAttr);
        doc->addDocExtraNamedItem(oldIdAttr);
    }

    HTMLElement::insertedIntoDocument();
}

void HTMLImageElement::removedFromDocument()
{
    if (document()->isHTMLDocument()) {
        HTMLDocument* doc = static_cast<HTMLDocument*>(document());

        doc->removeNamedItem(oldNameAttr);
        doc->removeDocExtraNamedItem(oldIdAttr);
    }

    HTMLElement::removedFromDocument();
}

int HTMLImageElement::width(bool ignorePendingStylesheets) const
{
    if (!renderer()) {
        // check the attribute first for an explicit pixel value
        bool ok;
        int width = getAttribute(widthAttr).toInt(&ok);
        if (ok)
            return width;
        
        // if the image is available, use its width
        if (m_imageLoader.image()) {
            float zoomFactor = document()->frame() ? document()->frame()->pageZoomFactor() : 1.0f;
            return m_imageLoader.image()->imageSize(zoomFactor).width();
        }
    }

    if (ignorePendingStylesheets)
        document()->updateLayoutIgnorePendingStylesheets();
    else
        document()->updateLayout();

    return renderer() ? renderer()->contentWidth() : 0;
}

int HTMLImageElement::height(bool ignorePendingStylesheets) const
{
    if (!renderer()) {
        // check the attribute first for an explicit pixel value
        bool ok;
        int height = getAttribute(heightAttr).toInt(&ok);
        if (ok)
            return height;
        
        // if the image is available, use its height
        if (m_imageLoader.image()) {
            float zoomFactor = document()->frame() ? document()->frame()->pageZoomFactor() : 1.0f;
            return m_imageLoader.image()->imageSize(zoomFactor).height();
        }   
    }

    if (ignorePendingStylesheets)
        document()->updateLayoutIgnorePendingStylesheets();
    else
        document()->updateLayout();

    return renderer() ? renderer()->contentHeight() : 0;
}

int HTMLImageElement::naturalWidth() const
{
    if (!m_imageLoader.image())
        return 0;

    return m_imageLoader.image()->imageSize(1.0f).width();
}

int HTMLImageElement::naturalHeight() const
{
    if (!m_imageLoader.image())
        return 0;
    
    return m_imageLoader.image()->imageSize(1.0f).height();
}
    
bool HTMLImageElement::isURLAttribute(Attribute* attr) const
{
    return attr->name() == srcAttr
        || attr->name() == lowsrcAttr
        || attr->name() == longdescAttr
        || (attr->name() == usemapAttr && attr->value().domString()[0] != '#');
}

String HTMLImageElement::name() const
{
    return getAttribute(nameAttr);
}

void HTMLImageElement::setName(const String& value)
{
    setAttribute(nameAttr, value);
}

String HTMLImageElement::align() const
{
    return getAttribute(alignAttr);
}

void HTMLImageElement::setAlign(const String& value)
{
    setAttribute(alignAttr, value);
}

String HTMLImageElement::alt() const
{
#if defined (SHP_SUPPORT_WML)
	if (document()->htmlMode() == Document::Wml)	{
		WMLDocument *doc = static_cast<WMLDocument *>(document()) ;
		String alt = getAttribute(altAttr) ;
		String temp ;
		String* domStr = document()->frame()->getVariableReplacedString (alt, WMLBrowserContext::ENoEscape);
		if (domStr)
			temp = domStr->copy();
		return temp ;
	}
	else
#endif //SHP_SUPPORT_WML
    return getAttribute(altAttr);
}

void HTMLImageElement::setAlt(const String& value)
{
    setAttribute(altAttr, value);
}

String HTMLImageElement::border() const
{
    return getAttribute(borderAttr);
}

void HTMLImageElement::setBorder(const String& value)
{
    setAttribute(borderAttr, value);
}

void HTMLImageElement::setHeight(int value)
{
    setAttribute(heightAttr, String::number(value));
}

int HTMLImageElement::hspace() const
{
    // ### return actual value
    return getAttribute(hspaceAttr).toInt();
}

void HTMLImageElement::setHspace(int value)
{
    setAttribute(hspaceAttr, String::number(value));
}

bool HTMLImageElement::isMap() const
{
    return !getAttribute(ismapAttr).isNull();
}

void HTMLImageElement::setIsMap(bool isMap)
{
    setAttribute(ismapAttr, isMap ? "" : 0);
}

String HTMLImageElement::longDesc() const
{
    return document()->completeURL(getAttribute(longdescAttr));
}

void HTMLImageElement::setLongDesc(const String& value)
{
    setAttribute(longdescAttr, value);
}

String HTMLImageElement::lowsrc() const
{
    return document()->completeURL(getAttribute(lowsrcAttr));
}

void HTMLImageElement::setLowsrc(const String& value)
{
    setAttribute(lowsrcAttr, value);
}

String HTMLImageElement::src() const
{
    return document()->completeURL(getAttribute(srcAttr));
}

void HTMLImageElement::setSrc(const String& value)
{
    setAttribute(srcAttr, value);
}

String HTMLImageElement::useMap() const
{
    return getAttribute(usemapAttr);
}

void HTMLImageElement::setUseMap(const String& value)
{
    setAttribute(usemapAttr, value);
}

int HTMLImageElement::vspace() const
{
    // ### return actual vspace
    return getAttribute(vspaceAttr).toInt();
}

void HTMLImageElement::setVspace(int value)
{
    setAttribute(vspaceAttr, String::number(value));
}

void HTMLImageElement::setWidth(int value)
{
    setAttribute(widthAttr, String::number(value));
}

int HTMLImageElement::x() const
{
    RenderObject* r = renderer();
    if (!r)
        return 0;
    int x, y;
    r->absolutePosition(x, y);
    return x;
}

int HTMLImageElement::y() const
{
    RenderObject* r = renderer();
    if (!r)
        return 0;
    int x, y;
    r->absolutePosition(x, y);
    return y;
}

bool HTMLImageElement::complete() const
{
    return m_imageLoader.imageComplete();
}

}
