/*
* ==============================================================================
*  Name         : FourWayNavigationShp.cpp
*  Part of      : WEBCORE / Page / SHP
*  Description  : Implementation of 4-Way Navigation
*  Date         : April 19, 2008
*  Author       : Prashant Nevase (n.prashant@siso.com)
* ==============================================================================
*/

#ifdef SUPPORT_FOUR_WAY_NAVIGATION //compile only if this flag is enabled

#include "config.h"
#include "Node.h"

#include "FrameView.h"
#include "EventHandler.h"
#include "Frame.h"

#include "HTMLAreaElement.h"
#include "HTMLNames.h"

#include "RenderFrame.h"

#include "FourWayNavigation.h"


//DEFINES
#define OFFSET_IN_BOUNDING_RECT		1
#define BAD_LOCATION				-1
#define DEFAULT_FRAME_WIDTH			250
#define DEFAULT_FRAME_HEIGHT		250
#define FWAY_MOD(x)	((x < 0)? -x : x)
#define FWAY_MIN(a, b) ((a) < (b) ? (a) : (b))

namespace WebCore {

//Member function definitions
void CFourWayNavigation::initFWNStatics()
{
}



void CFourWayNavigation::init()
{
}

void CFourWayNavigation::finalize()
{
}

CFourWayNavigation* CFourWayNavigation::createInstance()
{
	CFourWayNavigation* fourWayNavigation = new CFourWayNavigation;

	return fourWayNavigation;
}

void CFourWayNavigation::destroyInstance(CFourWayNavigation* fourWayNavigation)
{
	if(fourWayNavigation)
		delete fourWayNavigation;
}

CFourWayNavigation::CFourWayNavigation() :	m_frameVisibleBottom(DEFAULT_FRAME_HEIGHT),
											m_frameVisibleTop(0),
											m_frameVisibleRight(DEFAULT_FRAME_WIDTH),
											m_frameVisibleLeft(0),
											m_scrollStep(0),
	
											m_startHandlingKeyEvents(false)
{
	m_navigationFreezed = NAVIGATION_FREEZED_NONE;
	m_frameView = 0;


	m_currentFocusedNode.reset();
	m_nodeCollector.clear();
}

CFourWayNavigation::~CFourWayNavigation()
{
}

void CFourWayNavigation::setFrameView(FrameView* frameView)
{
	m_frameView = frameView;
	
	m_currentFocusedNode.reset();

	IntRect rect = m_frameView->frameGeometry();
	
	m_frameVisibleRight = rect.width();
	m_frameVisibleBottom = rect.height();
	m_frameVisibleTop = rect.y();
	m_frameVisibleLeft = rect.x();

	m_scrollStep = (FWAY_MIN(rect.height(), rect.width())) / 2;

	//release all the freezes
	m_navigationFreezed = NAVIGATION_FREEZED_NONE;

	doDefaultAction();
}

void CFourWayNavigation::resetPage()
{
	//pause handling key press events until page loading gets finished
	m_startHandlingKeyEvents = false;
	
	m_currentFocusedNode.reset();

	m_nodeCollector.clear();

	//release all the freezes
	m_navigationFreezed = NAVIGATION_FREEZED_NONE;
}

void CFourWayNavigation::pageLoadingFinished()
{
	doDefaultAction();
	//now start handling key press events
	m_startHandlingKeyEvents = true;
}

void CFourWayNavigation::doDefaultAction()
{
	if(this == CNavigation::m_navigation)
	{
		//set first default node
		setDefaultFocusedNode();
	}
}

void CFourWayNavigation::setDefaultFocusedNode()
{
	//collect the nodes in the new visible rect
	collectFocusableNodesInVisibleRect();

	//set the node which is closer to 0, 0 (i.e. left, top)
	if(m_nodeCollector.size() > 1)
	{
		int min = 0;

		for(int i = 1; i < m_nodeCollector.size(); ++i)
		{
			float iteratedNodeDist = getDistnace(FloatPoint(0, 0), FloatPoint(m_nodeCollector[i].leftEdge, m_nodeCollector[i].topEdge));

			float closerNodeDist = getDistnace(FloatPoint(0, 0), FloatPoint(m_nodeCollector[min].leftEdge, m_nodeCollector[min].topEdge));

			if(iteratedNodeDist < closerNodeDist)
			{
				min = i;
			}
		}

		m_currentFocusedNode = m_nodeCollector[min];
		
		focusCurrentNode();
	}
}

void CFourWayNavigation::focusCurrentNode()
{
	if(m_frameView && m_currentFocusedNode.node)
	{
		//m_currentFocusedNode.node->(focus(true);
		IntPoint wndPoint(m_currentFocusedNode.leftEdge, m_currentFocusedNode.topEdge);
		IntPoint contentsPoint = m_frameView->windowToContents(wndPoint);
		
		//Inform the webcore
		PlatformMouseEvent pme(wndPoint, contentsPoint, NoButton, MouseEventMoved,
			false, false, true, false, false, 0);

		m_frameView->frame()->eventHandler()->mouseMoved(pme);
	}
}

void CFourWayNavigation::activateCurrentNode()
{
	if(m_frameView && m_currentFocusedNode.node)
	{
		IntPoint wndPoint(m_currentFocusedNode.leftEdge, m_currentFocusedNode.topEdge);
		IntPoint contentsPoint = m_frameView->windowToContents(wndPoint);

		//send mouse pressed event
		PlatformMouseEvent pme(wndPoint, contentsPoint, LeftButton, MouseEventPressed,
							   true, false, false, false, false, 0);

		m_frameView->frame()->eventHandler()->handleMousePressEvent(pme);

		//send mouse released event
		PlatformMouseEvent pme1(wndPoint, contentsPoint, LeftButton, MouseEventReleased,
							   true, false, false, false, false, 0);

		m_frameView->frame()->eventHandler()->handleMouseReleaseEvent(pme1);
	}
}

void CFourWayNavigation::handleNavigationEvent(int keyCode, bool isKeyUp)
{
	if(m_startHandlingKeyEvents)
	{
		//skip navigation if freezed
		if(!isKeyUp)
		{
			//check for up/down/left/right navigations
			switch(keyCode)
			{
				case VK_HP_LEFT: //left
				{
					if(!(m_navigationFreezed & NAVIGATION_FREEZED_LEFT))
					{
						//release other freezes except left
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_RIGHT);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_UP);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_DOWN);

						if(!findNextFocusableNode(NAVIGATION_LEFT))
						{
							//freeze left navigation
							m_navigationFreezed |= NAVIGATION_FREEZED_LEFT;
						}
					}

					break;
				}
		        
				case VK_HP_RIGHT: //right
				{
					if(!(m_navigationFreezed & NAVIGATION_FREEZED_RIGHT))
					{
						//release other freezes except right
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_LEFT);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_UP);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_DOWN);
						
						if(!findNextFocusableNode(NAVIGATION_RIGHT))
						{
							//freeze right navigation
							m_navigationFreezed |= NAVIGATION_FREEZED_RIGHT;
						}
					}

					break;
				}
		        
				case VK_HP_UP: //up
				{
					if(!(m_navigationFreezed & NAVIGATION_FREEZED_UP))
					{
						//release other freezes except up
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_LEFT);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_RIGHT);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_DOWN);

						if(!findNextFocusableNode(NAVIGATION_UP))
						{
							//freeze up navigation
							m_navigationFreezed |= NAVIGATION_FREEZED_UP;
						}
					}

					break;
				}
		        
				case VK_HP_DOWN: //down
				{
					if(!(m_navigationFreezed & NAVIGATION_FREEZED_DOWN))
					{
						//release other freezes except down
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_LEFT);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_RIGHT);
						m_navigationFreezed &= ~(NAVIGATION_FREEZED_UP);


						if(!findNextFocusableNode(NAVIGATION_DOWN))
						{
							//freeze down navigation
							m_navigationFreezed |= NAVIGATION_FREEZED_DOWN;
						}
					}

					break;
				}
		        
				case VK_HP_OK: //enter
				{
					activateCurrentNode();
					break;
				}
			}
		}
	}
}

bool CFourWayNavigation::findNextFocusableNode(ENavigationType direction)
{
	bool result = true;
	//remove focus of current focused node
	m_currentFocusedNode.node = 0;

	//find the focusable node
	int location = getNextFocusableNode(direction);
	if(BAD_LOCATION != location)
	{
		m_currentFocusedNode = m_nodeCollector[location];
		focusCurrentNode();
	}
	else
	{
		if(scrollView(direction))
		{
			//collect the nodes in the new visible rect
			collectFocusableNodesInVisibleRect();

			location = getNextFocusableNode(direction);
			if(BAD_LOCATION != location)
			{
				m_currentFocusedNode = m_nodeCollector[location];
				focusCurrentNode();
			}
		}
		else
		{
			result = false;
		}
	}

	return result;
}

bool CFourWayNavigation::scrollView(ENavigationType direction)
{
	bool result = false;
	ScrollDirection scrollDir = ScrollDown;
	
	switch(direction)
	{
		case NAVIGATION_UP:
		{
			scrollDir = ScrollUp;
			break;
		}

		case NAVIGATION_DOWN:
		{
			scrollDir = ScrollDown;
			break;
		}

		case NAVIGATION_LEFT:
		{
			scrollDir = ScrollLeft;
			break;
		}

		case NAVIGATION_RIGHT:
		{
			scrollDir = ScrollRight;
			break;
		}
	}

	//scroll frame in the required direction
	IntSize oldOffset = m_frameView->scrollOffset();
	m_frameView->scroll(scrollDir, ScrollByLine, (m_scrollStep / LINE_STEP));
	IntSize newOffset = m_frameView->scrollOffset();

	//if there is difference in scolloffset, then we have scrolled, otherwise
	//we reached at the end edge.
	if(oldOffset != newOffset)
	{
		result = true;

		switch(direction)
		{
			case NAVIGATION_UP:
			{
				m_currentFocusedNode.topEdge += m_scrollStep;
				if(m_currentFocusedNode.topEdge > m_frameVisibleBottom)
					m_currentFocusedNode.topEdge = m_frameVisibleBottom;

				m_currentFocusedNode.bottomEdge = m_currentFocusedNode.topEdge;
				m_currentFocusedNode.vCenter = m_currentFocusedNode.topEdge;

				break;
			}

			case NAVIGATION_DOWN:
			{
				m_currentFocusedNode.bottomEdge -= m_scrollStep;
				if(m_currentFocusedNode.bottomEdge < m_frameVisibleTop)
					m_currentFocusedNode.bottomEdge = m_frameVisibleTop;

				m_currentFocusedNode.topEdge = m_currentFocusedNode.bottomEdge;
				m_currentFocusedNode.vCenter = m_currentFocusedNode.bottomEdge;

				break;
			}

			case NAVIGATION_LEFT:
			{
				m_currentFocusedNode.leftEdge += m_scrollStep;
				if(m_currentFocusedNode.leftEdge > m_frameVisibleRight)
					m_currentFocusedNode.leftEdge = m_frameVisibleRight;

				m_currentFocusedNode.rightEdge = m_currentFocusedNode.leftEdge;
				m_currentFocusedNode.hCenter = m_currentFocusedNode.leftEdge;

				break;
			}

			case NAVIGATION_RIGHT:
			{
				m_currentFocusedNode.rightEdge -= m_scrollStep;
				if(m_currentFocusedNode.rightEdge < m_frameVisibleLeft)
					m_currentFocusedNode.rightEdge = m_frameVisibleLeft;

				m_currentFocusedNode.leftEdge = m_currentFocusedNode.rightEdge;
				m_currentFocusedNode.hCenter = m_currentFocusedNode.rightEdge;

				break;
			}
		}
	}

	return result;
}

int CFourWayNavigation::getNextFocusableNode(ENavigationType direction)
{
	int location = BAD_LOCATION;
	NodeWithEdges proposedNode;

	switch(direction)
	{
		case NAVIGATION_UP:
		{
			for(int i = 0; i < m_nodeCollector.size(); ++i)
			{
				if(m_nodeCollector[i].bottomEdge <= m_currentFocusedNode.topEdge)
				{
					if(BAD_LOCATION == location)
					{
						//we are entering first time
						proposedNode = m_nodeCollector[i];
						location = i;
					}
					else
					{
						float iteratedNodeDist = getRelativeVerticalDistance(m_nodeCollector[i].leftEdge, m_nodeCollector[i].rightEdge, m_nodeCollector[i].bottomEdge,
																			m_currentFocusedNode.leftEdge, m_currentFocusedNode.rightEdge, m_currentFocusedNode.topEdge);

						float proposedNodeDist = getRelativeVerticalDistance(proposedNode.leftEdge, proposedNode.rightEdge, proposedNode.bottomEdge,
																			m_currentFocusedNode.leftEdge, m_currentFocusedNode.rightEdge, m_currentFocusedNode.topEdge);

						if(iteratedNodeDist < proposedNodeDist)
						{
							//change proposed node to current iteration
							proposedNode = m_nodeCollector[i];
							location = i;
						}
					}
				}
			}

			break;
		}

		case NAVIGATION_DOWN:
		{
			for(int i = 0; i < m_nodeCollector.size(); ++i)
			{
				if(m_nodeCollector[i].topEdge >= m_currentFocusedNode.bottomEdge)
				{
					if(BAD_LOCATION == location)
					{
						//we are entering first time
						proposedNode = m_nodeCollector[i];
						location = i;
					}
					else
					{
						float iteratedNodeDist = getRelativeVerticalDistance(m_nodeCollector[i].leftEdge, m_nodeCollector[i].rightEdge, m_nodeCollector[i].topEdge,
																			m_currentFocusedNode.leftEdge, m_currentFocusedNode.rightEdge, m_currentFocusedNode.bottomEdge);

						float proposedNodeDist = getRelativeVerticalDistance(proposedNode.leftEdge, proposedNode.rightEdge, proposedNode.topEdge,
																			m_currentFocusedNode.leftEdge, m_currentFocusedNode.rightEdge, m_currentFocusedNode.bottomEdge);

						if(iteratedNodeDist < proposedNodeDist)
						{
							//change proposed node to current iteration
							proposedNode = m_nodeCollector[i];
							location = i;
						}
					}
				}
			}

			break;
		}

		case NAVIGATION_LEFT:
		{
			for(int i = 0; i < m_nodeCollector.size(); ++i)
			{
				if(m_nodeCollector[i].rightEdge <= m_currentFocusedNode.leftEdge)
				{
					if(BAD_LOCATION == location)
					{
						//we are entering first time
						proposedNode = m_nodeCollector[i];
						location = i;
					}
					else
					{
						float iteratedNodeDist = getRelativeHorizontalDistance(m_nodeCollector[i].topEdge, m_nodeCollector[i].bottomEdge, m_nodeCollector[i].rightEdge,
																			m_currentFocusedNode.topEdge, m_currentFocusedNode.bottomEdge, m_currentFocusedNode.leftEdge);

						float proposedNodeDist = getRelativeHorizontalDistance(proposedNode.topEdge, proposedNode.bottomEdge, proposedNode.rightEdge,
																			m_currentFocusedNode.topEdge, m_currentFocusedNode.bottomEdge, m_currentFocusedNode.leftEdge);

						if(iteratedNodeDist < proposedNodeDist)
						{
							//change proposed node to current iteration
							proposedNode = m_nodeCollector[i];
							location = i;
						}
					}
				}
			}

			break;
		}

		case NAVIGATION_RIGHT:
		{
			for(int i = 0; i < m_nodeCollector.size(); ++i)
			{
				if(m_nodeCollector[i].leftEdge >= m_currentFocusedNode.rightEdge)
				{
					if(BAD_LOCATION == location)
					{
						//we are entering first time
						proposedNode = m_nodeCollector[i];
						location = i;
					}
					else
					{
						float iteratedNodeDist = getRelativeHorizontalDistance(m_nodeCollector[i].topEdge, m_nodeCollector[i].bottomEdge, m_nodeCollector[i].leftEdge,
																			m_currentFocusedNode.topEdge, m_currentFocusedNode.bottomEdge, m_currentFocusedNode.rightEdge);

						float proposedNodeDist = getRelativeHorizontalDistance(proposedNode.topEdge, proposedNode.bottomEdge, proposedNode.leftEdge,
																			m_currentFocusedNode.topEdge, m_currentFocusedNode.bottomEdge, m_currentFocusedNode.rightEdge);

						if(iteratedNodeDist < proposedNodeDist)
						{
							//change proposed node to current iteration
							proposedNode = m_nodeCollector[i];
							location = i;
						}
					}
				}
			}

			break;
		}
	}

	return location;
}

float CFourWayNavigation::getRelativeVerticalDistance(int left, int right, int v, int leftRef, int rightRef, int vRef)
{
	if(right <= leftRef)
		return getDistnace(FloatPoint(right, v), FloatPoint(leftRef, vRef));

	if(left >= rightRef)
		return getDistnace(FloatPoint(left, v), FloatPoint(rightRef, vRef));

	//handle overlapping condition
	return FWAY_MOD(v - vRef);
}

float CFourWayNavigation::getRelativeHorizontalDistance(int top, int bottom, int h, int topRef, int bottomRef, int hRef)
{
	if(bottom <= topRef)
		return getDistnace(FloatPoint(h, bottom), FloatPoint(hRef, topRef));

	if(top >= bottomRef)
		return getDistnace(FloatPoint(h, top), FloatPoint(hRef, bottomRef));

	//handle overlapping condition
	return FWAY_MOD(h - hRef);
}

void CFourWayNavigation::collectFocusableNodesInVisibleRect()
{
	PassRefPtr<HTMLCollection> allNodes = m_frameView->frame()->document()->all();
	Node* node;

	int contentsx = m_frameView->contentsX();
	int contentsy = m_frameView->contentsY();
	int visibleW = m_frameView->visibleWidth();
	int visibleH = m_frameView->visibleHeight();

	//remove previously collected nodes
	m_nodeCollector.clear();

	IntRect viewRect(contentsx, contentsy, visibleW, visibleH);

	for(int i = 0; i < allNodes->length(); i++)
    {
		node = allNodes->item(i);
		RenderObject* renderer = node->renderer();
		if(renderer)
		{
			IntRect rect = renderer->absoluteBoundingBoxRect();
			if(rect.intersects(viewRect) && canNodeHaveFocus(node))
			{
				NodeWithEdges nodeWithEdges;
				nodeWithEdges.node = node;

				//set the scanlines for node
				IntRect boundingRect = nodeWithEdges.node->renderer()->absoluteBoundingBoxRect();
				boundingRect.intersect(viewRect);
				IntRect wndRect = m_frameView->contentsToWindow(boundingRect);
				
				nodeWithEdges.leftEdge = wndRect.x();
				nodeWithEdges.rightEdge = wndRect.x() + wndRect.width();
				nodeWithEdges.topEdge = wndRect.y();
				nodeWithEdges.bottomEdge = wndRect.y() + wndRect.height();

				nodeWithEdges.hCenter = nodeWithEdges.leftEdge + 
								((nodeWithEdges.rightEdge - nodeWithEdges.leftEdge) / 2);

				nodeWithEdges.vCenter = nodeWithEdges.topEdge + 
								((nodeWithEdges.bottomEdge - nodeWithEdges.topEdge) / 2);

				m_nodeCollector.append(nodeWithEdges);
			}
		}
	}
}

} //namespace WebCore

#endif //SUPPORT_FOUR_WAY_NAVIGATION