/*
* ==============================================================================
*  Name         : ScrollNavigationShp.cpp
*  Part of      : WEBCORE / Page / SHP
*  Description  : Interface of Navigation
*  Date         : April 29, 2008
*  Authors      : Vineet Chaudhary (vineet.c@siso.com),
*                 Prashant Nevase (n.prashant@siso.com)
* ==============================================================================
*/

#include "Platform.h"

#include "Page.h"
#include "EventHandler.h"
#include "Frame.h"
#include "FrameView.h"
#include "PlatformKeyboardEvent.h"
#include "HTMLNames.h"

#include "RenderFrame.h"
#include "PlatformScrollBar.h"
#include "Node.h"

#include "MouseEventWithHitTestResults.h"
#include "HitTestRequest.h"

#include "ScrollNavigation.h"


#define SCROLL_NAVIGATION_DEFAULT_STEP			(LINE_STEP * 3)
#define SCROLL_OFFSET_MAX_VALUE_THRESHOLD		(100)
#define SCROLL_NAVIGATION_ACCELERATION_STEP		(LINE_STEP)

namespace WebCore {

void CScrollNavigation::initSNStatics()
{
}


CScrollNavigation::CScrollNavigation()
{
	m_scrollOffset = SCROLL_NAVIGATION_DEFAULT_STEP;
	
	m_fastScrolling.scrollOffsetMaxValueThreshold = SCROLL_OFFSET_MAX_VALUE_THRESHOLD;
	m_fastScrolling.scrollOffsetMultiplier = 1;
	m_fastScrolling.scrollOffsetLastMultiplier = 0;
	m_fastScrolling.isFastScrolling = false;
}

CScrollNavigation::~CScrollNavigation()
{
}

void CScrollNavigation::init()
{
	m_defaultScrollStep = SCROLL_NAVIGATION_DEFAULT_STEP;
	m_defaultAccelerationStep = SCROLL_NAVIGATION_ACCELERATION_STEP;
}

void CScrollNavigation::finalize()
{
}

CScrollNavigation* CScrollNavigation::createInstance()
{
	CScrollNavigation* scrollNavigation = new CScrollNavigation;

	return scrollNavigation;
}

void CScrollNavigation::destroyInstance(CScrollNavigation* scrollNavigation)
{
	if(scrollNavigation)
		delete scrollNavigation;
}

void CScrollNavigation::setFrameView(FrameView* frameView)
{
	if(frameView)
	{
		m_frameView = frameView;
	}
}

void CScrollNavigation::resetPage()
{
	if(m_mainFrameView)
	{
		setFrameView(m_mainFrameView);
		m_hoveredNode = 0;
	}
}

void CScrollNavigation::pageLoadingFinished()
{
}

void CScrollNavigation::handleNavigationEvent(int keyCode, bool isKeyUp)
{
	if(m_frameView)
	{
		if(isKeyUp)
			stopFastScrolling();
		else
			startFastScrolling();

		if(!isKeyUp)
		{
			switch(keyCode)
			{
				case VK_HP_UP:
				{
					m_frameView->scroll(ScrollUp, ScrollByLine, (m_scrollOffset / LINE_STEP));
					break;
				}

				case VK_HP_DOWN: 
				{
					m_frameView->scroll(ScrollDown, ScrollByLine, (m_scrollOffset / LINE_STEP));
					break;
				}

				case VK_HP_LEFT:
				{
					m_frameView->scroll(ScrollLeft, ScrollByLine, (m_scrollOffset / LINE_STEP));
					break;
				}

				case VK_HP_RIGHT:
				{
					m_frameView->scroll(ScrollRight, ScrollByLine, (m_scrollOffset / LINE_STEP));
					break;
				}

				case VK_HP_OK:
				{
					break;
				}

				default:
				{
					m_frameView->frame()->eventHandler()->keyEvent(PlatformKeyboardEvent(keyCode, isKeyUp));
					break;
				}
			}
		}
	}
}

#ifdef SUPPORT_TOUCH_SCREEN
void CScrollNavigation::touchPressed(IntPoint wndPoint)
{
	if(m_frameView)
	{
		IntPoint contentsPoint = m_frameView->windowToContents(wndPoint);
		m_pointerX = contentsPoint.x();
		m_pointerY = contentsPoint.y();
	}
}

void CScrollNavigation::touchMoved(IntPoint wndPoint)
{
	if(m_frameView)
	{
		IntPoint contentsPoint = m_frameView->windowToContents(wndPoint);
		m_pointerX = contentsPoint.x();
		m_pointerY = contentsPoint.y();

		FrameView* view = frameViewAtWindowPoint(wndPoint);
		
		if(view != m_frameView)
		{
			setFrameView(view);
		}
	}
}

void CScrollNavigation::touchReleased(IntPoint wndPoint)
{
	if(m_frameView)
	{
		IntPoint documentPoint = m_frameView->windowToContents(wndPoint);
		m_pointerX = documentPoint.x();
		m_pointerY = documentPoint.y();

        Document* doc = m_frameView->frame()->document();

		Node* node = doc->elementFromPoint(documentPoint.x(), documentPoint.y());
        if(node && node->hasTagName(HTMLNames::imgTag))
        {
            if(node->parent() && node->parent()->hasTagName(HTMLNames::aTag))
            {
                node = node->parent();
            }
        }

        bool isNodeHidden = false;
        if(node && node->isElementNode() )
        {
            Element * element = static_cast<Element*>(node);
            isNodeHidden = element->isInputTypeHidden();
        }

        if(false == isNodeHidden)
        {
            doc->setFocusedNode(node);
        }
	}
}

Node* CScrollNavigation::getHoveredNode()
{
	if(m_frameView)
	{
		IntPoint documentPoint(m_pointerX, m_pointerY);
		IntPoint wndPoint = m_frameView->contentsToWindow(wndPoint);

		const HitTestRequest hitTest(false, true);
		PlatformMouseEvent pme(wndPoint, documentPoint, NoButton, MouseEventMoved,
			false, false, true, false, false, 0);
		m_hoveredNode = m_frameView->frame()->document()->prepareMouseEvent(hitTest, documentPoint, pme).targetNode(); 
		return m_hoveredNode;
	}
}
#endif //SUPPORT_TOUCH_SCREEN

}
//namespace WebCore