/*
 * Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
 * Copyright (C) 2006, 2007 Apple Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "Cursor.h"

#include "Image.h"
#include "IntPoint.h"
#include "NotImplemented.h"

//plaftorm graphics
#include "gwes.h"
#include "RsrcMgr.h" 
#include "SFBrowserRsrc.h"

WebCore::CPlatformCursorHelper gpch;
#ifdef SFB_GLOBAL_STATIC_HANDLING
void initCursorShpGlobals()
{
    gpch.init();
}
#endif


#define ALPHA_CURSORS
#define SFB_RSRC_FILE_NAME "SFBrowserRsrc.rbm"
namespace WebCore {


Cursor::Cursor(const Cursor& other)
    : m_impl(other.m_impl)
{
}

Cursor::Cursor(Image* img, const IntPoint& hotspot) 
{ 
	notImplemented();
}

Cursor::Cursor(PlatformCursor impl)
{
	m_impl = impl;
}

Cursor::~Cursor()
{
	notImplemented();
}

Cursor& Cursor::operator=(const Cursor& other)
{
    m_impl = other.m_impl;
    return *this;
}


const Cursor& pointerCursor()
{
    static const Cursor c(POINTER_CURSOR);
	return c;
}

const Cursor& crossCursor()
{
    static const Cursor c(CROSS_CURSOR);
    return c;
}

const Cursor& handCursor()
{
    static const Cursor c(HAND_CURSOR);
    return c;
}

const Cursor& iBeamCursor()
{
    static const Cursor c(IBEAM_CURSOR);
    return c;
}

const Cursor& waitCursor()
{
    static const Cursor c(WAIT_CURSOR);
    return c;
}

const Cursor& helpCursor()
{
    static const Cursor c(HELP_CURSOR);
    return c;
}

const Cursor& eastResizeCursor()
{
    static const Cursor c(EASTRESIZE_CURSOR);
    return c;
}

const Cursor& northResizeCursor()
{
    static const Cursor c(NORTHRESIZE_CURSOR);
    return c;
}

const Cursor& northEastResizeCursor()
{
    static const Cursor c(NORTHEASTRESIZE_CURSOR);
    return c;
}

const Cursor& northWestResizeCursor()
{
    static const Cursor c(NORTHWESTRESIZE_CURSOR);
	return c;
}

const Cursor& southResizeCursor()
{
    static const Cursor c(SOUTHRESIZE_CURSOR);
	return c;
}

const Cursor& southEastResizeCursor()
{
    static const Cursor c(SOUTHEASTRESIZE_CURSOR);
	return c;
}

const Cursor& southWestResizeCursor()
{
    static const Cursor c(SOUTHWESTRESIZE_CURSOR);
	return c;
}

const Cursor& westResizeCursor()
{
    static const Cursor c(WESTRESIZE_CURSOR);
	return c;
}

const Cursor& northSouthResizeCursor()
{
    static const Cursor c(NORTHSOUTHRESIZE_CURSOR);
	return c;
}

const Cursor& eastWestResizeCursor()
{
    static const Cursor c(EASTWESTRESIZE_CURSOR);
	return c;
}

const Cursor& northEastSouthWestResizeCursor()
{
    static const Cursor c(NORTHEASTSOUTHWESTRESIZE_CURSOR);
    return c;
}

const Cursor& northWestSouthEastResizeCursor()
{
    static const Cursor c(NORTHWESTSOUTHEASTRESIZE_CURSOR);
    return c;
}

const Cursor& columnResizeCursor()
{
    static const Cursor c(COLUMNRESIZE_CURSOR);
    return c;
}

const Cursor& rowResizeCursor()
{
    static const Cursor c(ROWRESIZE_CURSOR);
    return c;
}

const Cursor& moveCursor() 
{
    static const Cursor c(MOVE_CURSOR);
    return c;
}

const Cursor& verticalTextCursor()
{
    static const Cursor c(VERTICALTEXT_CURSOR);
    return c;
}

const Cursor& cellCursor()
{
    return pointerCursor();
}

const Cursor& contextMenuCursor()
{
    return pointerCursor();
}

const Cursor& aliasCursor()
{
    return pointerCursor();
}

const Cursor& progressCursor()
{
    static const Cursor c(PROGRESS_CURSOR);
	return c;
}

const Cursor& noDropCursor()
{
    static const Cursor c(NODROP_CURSOR);
    return c;
}

const Cursor& copyCursor()
{
    return pointerCursor();
}

const Cursor& mailCursor()
{
    static const Cursor c(MAIL_CURSOR);
    return c;
}

const Cursor& phoneCursor()
{
	static const Cursor c(PHONE_CURSOR);
    return c;
}

const Cursor& noneCursor()
{
    return pointerCursor();
}

const Cursor& notAllowedCursor()
{
    static const Cursor c(NOTALLOWED_CURSOR);
	return c;
}

// FIXME: <rdar://problem/5224996> Add zoom in and zoom out cursors
const Cursor& zoomInCursor()
{
    return pointerCursor();
}

const Cursor& zoomOutCursor()
{
    return pointerCursor();
}

//platform specific helper for drawing actual cursors
#define CURSOR_WIDTH						(25)
#define CURSOR_HEIGHT						(25)
#define CURSOR_X_OFFSET						(5)
#define CURSOR_Y_OFFSET						(0)
#define OUT_OF_SCREEN_CURSOR_LOCATION_X		(5000)
#define OUT_OF_SCREEN_CURSOR_LOCATION_Y		(500)

void CPlatformCursorHelper::Initialize(int hWin)
{
	if(!gpch.m_hCursorWin)
	{
		if(hWin)
		{
			gpch.m_hCursorWin = WsCreateSimple(hWin, 0, 0, CURSOR_WIDTH, CURSOR_HEIGHT, 0);
			WmSetWindowTransparentMode(gpch.m_hCursorWin, true);
			WmSetWindowBgColor(gpch.m_hCursorWin, WHITE);
			WmSetWindowPos(gpch.m_hCursorWin, 
				OUT_OF_SCREEN_CURSOR_LOCATION_X, 
				OUT_OF_SCREEN_CURSOR_LOCATION_Y);
		}
	}
}

void CPlatformCursorHelper::Finalize()
{
	if(gpch.m_hCursorWin)
	{
		//vamsi - check reqd due to F480 UI flow.
		if(WmIsWindowValid(gpch.m_hCursorWin))
			WmDestroyWindow(gpch.m_hCursorWin);
		gpch.m_hCursorWin = 0;
	}
}

void CPlatformCursorHelper::SetCursor(const Cursor& cursor)
{
	gpch.m_currentCursor = cursor;
}

void CPlatformCursorHelper::DrawCursor(int hGC, int pointerX, int pointerY)
{
	WmBitmap*		pBitmap = NULL;

	if(gpch.m_hCursorWin)
	{
		int locX = pointerX - CURSOR_X_OFFSET;
		int locY = pointerY - CURSOR_Y_OFFSET;
		int width = CURSOR_WIDTH;
		int height = CURSOR_HEIGHT;

		int copyToX = 0;
		int copyToY = 0;

		WmSetWindowPos(gpch.m_hCursorWin, locX, locY);

		if(locX < 0)
		{
			width += locX;
			copyToX -= locX;
			locX = 0;
		}

		if(locY < 0)
		{
			height += locY;
			copyToY -= locY;
			locY = 0;
		}

		HGC hCursorGC = WmCreateWindowMemGC(gpch.m_hCursorWin);
		if(hCursorGC > 0)
		{
			WmCopyGC(hCursorGC, copyToX, copyToY, hGC, locX, locY, width, height);
			CursorType ct = gpch.m_currentCursor.impl();

			switch(ct)
			{
				case POINTER_CURSOR:
				{		
					pBitmap = RsrcLoadBitmap(SFB_RSRC_FILE_NAME, IDB_SFB_CURSOR_ARROW_WHITE);//sneha
					if(pBitmap)
						WmDrawBitmap(hCursorGC, 0, 0, pBitmap);
					
					break;
				}

				case IBEAM_CURSOR:
				{
					pBitmap = RsrcLoadBitmap(SFB_RSRC_FILE_NAME, IDB_SFB_IBEAM_CURSOR);//sneha
					if(pBitmap)
						WmDrawBitmap(hCursorGC, 0, 0, pBitmap);
					
					break;
				}
				
				case HAND_CURSOR:
				{

					pBitmap = RsrcLoadBitmap(SFB_RSRC_FILE_NAME, IDB_SFB_CURSOR_HAND);//sneha
					if(pBitmap)
						WmDrawBitmap(hCursorGC, 0, 0, pBitmap);
					
					break;
				}

				case MAIL_CURSOR:
				{
					int rectHeight = 10;
					int rectWidth = 10;
					WmDrawFilledRect(hCursorGC, GREEN, 0, 0, rectHeight, rectWidth);
					
					break;
				}

				case PHONE_CURSOR:
				{
					int rectHeight = 10;
					int rectWidth = 10;
					WmDrawFilledRect(hCursorGC, RED, 0, 0, rectHeight, rectWidth);
					
					break;
				}

				default:
				{
					pBitmap = RsrcLoadBitmap(SFB_RSRC_FILE_NAME, IDB_SFB_CURSOR_ARROW);//sneha
					if(pBitmap)
						WmDrawBitmap(hCursorGC, 0, 0, pBitmap);

					break;
				}

			}

			WmShowGC(hCursorGC);
			WmDeleteGC(hCursorGC);
		}

		//Once the cursor is drawn, move the window outside screen.
		WmSetWindowPos(gpch.m_hCursorWin, 
			OUT_OF_SCREEN_CURSOR_LOCATION_X, 
			OUT_OF_SCREEN_CURSOR_LOCATION_Y);
	}

	if(pBitmap)
		RsrcFreeBitmap(pBitmap);
}

int CPlatformCursorHelper::GetCursorHeight()
{
	return CURSOR_HEIGHT;
}

int CPlatformCursorHelper::GetCursorWidth()
{
	return CURSOR_WIDTH;
}

int CPlatformCursorHelper::GetCursorXOffset()
{
	return CURSOR_X_OFFSET;
}

int CPlatformCursorHelper::GetCursorYOffset()
{
	return CURSOR_Y_OFFSET;
}

}
