/*
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2008 Collabora, Ltd.
 * Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "FileSystem.h"

#include "NotImplemented.h"
#include "PlatformString.h"
#include "CString.h"

#include "ShpJasmine.h"

#define TEMP_FILE_PATH "/User/Br/Temp/"
#define HOME_DIR_PATH "/User/Br/"
#define TEMP_FILE_NAME "_tmp.dat"

namespace WebCore {

bool fileExists(const String& path)
{
    return SFBalIsFileExist(path.latin1().data());
}

bool deleteFile(const String& path)
{
    return SFBalRemoveFile(path.latin1().data());
}

bool deleteEmptyDirectory(const String& path)
{
    return SFBalRemoveDir(path.latin1().data());
}

bool getFileSize(const String& path, long long& resultSize)
{
    resultSize = SFBalGetFileSize(path.latin1().data());
    return SFB_INVALID_HOBJ == resultSize ? false : true;
}

bool getFileModificationTime(const String&, time_t&)
{
    // Not yet implemented by SHP :(
    notImplemented();
    return false;
}

String pathByAppendingComponent(const String& path, const String& component)
{
    notImplemented();
    return String::fromUTF8("");
}

bool makeAllDirectories(const String& path)
{
    // return if the path already exists
    if(SFBalFmIsExistDir(path.latin1().data()))
    {
        return true;
    }

    String tempPath;
    int pos = 0;

    // Loop to find existing directories
    do
    {
        pos = path.find(BAL_SLASH_CHAR, pos);

        if(-1 == pos)
        {
            tempPath = path;
            break;
        }
        tempPath = path.left(++pos);
    }while(SFBalFmIsExistDir(tempPath.latin1().data()));

    bool result = true;
    
    // Loop to create remaining directories
    while( tempPath != path )
    {
        result = SFBalCreateDir(tempPath.latin1().data());
        pos = path.find(BAL_SLASH_CHAR, pos);
        if(-1 == pos)
        {
            tempPath = path;
        }
        else
        {
            tempPath = path.left(++pos);
        }
    }

    // If the path doesn't end with '/'
    if(!SFBalFmIsExistDir(path.latin1().data()))
    {
        result = SFBalCreateDir(path.latin1().data());
    }

    return result;
}

String homeDirectoryPath()
{
    return HOME_DIR_PATH;
}

String pathGetFileName(const String& pathName)
{
    return pathName.substring(pathName.reverseFind('/') + 1);
}

CString openTemporaryFile(const char* prefix, PlatformFileHandle& handle)
{
    char tempPath[] = TEMP_FILE_PATH;

    if(!SFBalFmIsExistDir(tempPath))
    {
        if(!makeAllDirectories(tempPath))
        {
            return "";
        }
}

    char tempFile[BAL_FILEPATH_MAX_LEN];
    strcpy(tempFile, tempPath);
    strcat(tempFile, prefix);
    strcat(tempFile, TEMP_FILE_NAME);

    PlatformFileHandle tempHandle = SFBalOpenFile(tempFile, BAL_FILE_CREATE | 
                                                        BAL_FILE_READ |
                                                        BAL_FILE_WRITE);

    if (isHandleValid(tempHandle)) {
        handle = tempHandle;
        return tempFile;
    }

    return "";
}

/*
 * @desc    openFile : function called open a file
 * @param   path     : File path
 * @param   fileName : Name of the file
 * @param   mode     : File open mode. For e.g. FILE_MODE_CREATE, FILE_MODE_READ, 
					   FILE_MODE_CREATE | FILE_MODE_READ | FILE_MODE_WRITE, etc.
 * @param   handle   : File Handle
 * @return  File path
 */
CString openFile(const char* path,const char* fileName, UINT32 mode, PlatformFileHandle& handle)
{
	char tempFile[BAL_FILEPATH_MAX_LEN];
	
	if(!SFBalFmIsExistDir(path))
    {
        if(!makeAllDirectories(path))
        {
            return "";
        }
	}
    
	strcpy(tempFile, path);
    strcat(tempFile, fileName);

	PlatformFileHandle tempHandle = SFBalOpenFile(tempFile, mode);

    handle = tempHandle;	
	return tempFile;
}

/*
 * @desc    readFile  : function called read data from a file
 * @param   handle    : File Handle
 * @param   buffer    : Buffer to read
 * @param   bufferLen : Length to be read
 * @param   bytesRead : Total length read 
 * @return  True/False
 */
bool readFile(PlatformFileHandle handle, char* buffer, long long bufferLen, unsigned int* bytesRead)
{
	return SFBalReadFile(handle , buffer , bufferLen , (UINT32* ) bytesRead);
}

/*
 * @desc    readFileWithSeek  : Read data from a file with a seek position
 * @param   handle            : File Handle
 * @param   buffer            : Buffer to read
 * @param   bufferLen         : Length to be read
 * @param   bytesRead         : Total length read
 * @param   offset            : Seek position
 * @param   mode              : File seek mode For e.g FILE_SEEK_SET, FILE_SEEK_CUR, FILE_FM_SEEK_END.
                                The default mode is FILE_SEEK_CUR
 * @return  True/False
 */
bool readFileWithSeek(PlatformFileHandle handle, char* buffer, long long bufferLen, unsigned int* bytesRead, unsigned int offset, UINT32 mode)
{
	bool retVal = false;

	retVal = SFBalSeekFile(handle, mode , offset);
	retVal = SFBalReadFile(handle , buffer , bufferLen , (UINT32* ) bytesRead);

	return retVal;
}

/*
 * @desc    closeFile : function called close a file
 * @param   handle    : File Handle
 * @return  
 */
void closeFile(PlatformFileHandle& handle)
{
     SFBalCloseFile(handle);
}

/*
 * @desc    writeToFile : function to write data to a file
 * @param   handle   : File Handle
 * @param   data     : Data to write into the file
 * @param   length   : Length of data to write
 * @return  
 */
int writeToFile(PlatformFileHandle handle, const char* data, int length)
{
    int retVal = 0;

    if(SFBalWriteFile(handle, (void*)data, length))
    {
        retVal = length;
    }
    return retVal;
}

bool unloadModule(PlatformModule module)
{
    notImplemented();
    return false;
}

#if TEST_FILE_APIS
void testFileAPIs()
{
    // Make sure /User/Br/Temp.txt exists
    long long retVal = fileExists("/User/Br/Temp.txt");
    retVal = deleteFile("/User/Br/Temp.txt");
    
    retVal = fileExists("/User/Br/Temp.txt");
    retVal = deleteFile("/User/Br/Temp.txt");

    // Temp is an empty directory
    // Temp1 is a non-empty directory
    // Temp2 directory does not exists
    retVal = deleteEmptyDirectory("/User/Br/Temp");
    retVal = deleteEmptyDirectory("/User/Br/Temp1");
    retVal = deleteEmptyDirectory("/User/Br/Temp2");

    // Temp.txt does not exists since it is deleted above
    getFileSize("/User/Br/Temp1.txt", retVal);
    getFileSize("/User/Br/Temp.txt", retVal);

    String home = homeDirectoryPath();
    String path = pathGetFileName("/User/Br/Temp1.txt");

    // This will create /User/Br/Temp/Test_tmp.dat file
    PlatformFileHandle handle;
    openTemporaryFile("Test", handle);
    retVal = writeToFile(handle, "This is a test", 14);
    closeFile(handle);

    // Temp1 directory already exists
    retVal = makeAllDirectories("/User/Br/Temp1");
    retVal = makeAllDirectories("/User/Br/Temp1/Temp2/Temp3/Temp4/Temp5/");
    retVal = makeAllDirectories("/User/Br/Temp1/Temp2/Temp3/Temp4/Temp5/Temp6");
    retVal = makeAllDirectories("/User/Br/Temp1/Temp2/Temp3/Temp4/Temp5/Temp6");
}
#endif // TEST_FILE_APIS

}
