/*
 * Copyright (C) 2007 Apple Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "PlatformScrollBar.h"

#include "EventHandler.h"
#include "FrameView.h"
#include "Frame.h"
#include "GraphicsContext.h"
#include "IntRect.h"
#include "PlatformMouseEvent.h"
#include "NotImplemented.h"

#include "gwes.h"
#include "wsscroll.h"

// Enable/disable the platform scroll bar support

#ifdef USE_PLATFORM_SCROLLBARS

#define VERTICAL_SCROLLBAR_WIDTH		7
#define HORIZONTAL_SCROLLBAR_HEIGHT		7

#else

#define VERTICAL_SCROLLBAR_WIDTH        6
#define HORIZONTAL_SCROLLBAR_HEIGHT     6

#define WHITE_COLOR						255, 255, 255
#define TRACK_SHADOW_COLOR				132, 197, 255
#define TRACK_COLOR						15, 65, 205
#define THUMB_BORDER_COLOR				5, 46, 141
#define THUMB_COLOR						70, 121, 255
#define THUMB_LIGHT_COLOR				132, 197, 255

// minimum height and width of the vertical and hortizontal scroll bars respectively
#define MIN_THUMB_SIZE					6

#endif // USE_PLATFORM_SCROLLBARS


// FIXME: There are repainting problems due to Aqua scroll bar buttons' visual overflow.

using namespace std;

namespace WebCore {

PlatformScrollbar::PlatformScrollbar(ScrollbarClient* client, ScrollbarOrientation orientation, ScrollbarControlSize size)
    : Scrollbar(client, orientation, size), m_hoveredPart(NoPart), m_pressedPart(NoPart), m_pressedPos(0),
      m_scrollTimer(this, &PlatformScrollbar::autoscrollTimerFired),
      m_overlapsResizer(false),
#ifndef USE_PLATFORM_SCROLLBARS
	  m_thumbRect(0, 0, 0, 0),
#endif
	  m_parent(NULL)
{
	int style;
	IntRect scrollRect;
	IntRect clipRect = client->windowClipRect();
	if(HorizontalScrollbar == orientation)
	{
		style = SCROLL_STYLE_HORIZONTAL;
		scrollRect.setX(clipRect.x());
		scrollRect.setY(clipRect.y() + clipRect.height() - HORIZONTAL_SCROLLBAR_HEIGHT);
		scrollRect.setWidth(clipRect.width());
		scrollRect.setHeight(HORIZONTAL_SCROLLBAR_HEIGHT);
	}
	else
	{
		style = SCROLL_STYLE_VERTICAL;
		scrollRect.setX(clipRect.x() + clipRect.width() - VERTICAL_SCROLLBAR_WIDTH);
		scrollRect.setY(clipRect.y());
		scrollRect.setWidth(VERTICAL_SCROLLBAR_WIDTH);
		scrollRect.setHeight(clipRect.height());
	}
	setFrameGeometry(scrollRect);

#ifdef USE_PLATFORM_SCROLLBARS
	if (scrollRect.height() > 0 && scrollRect.width() > 0)
	{
		
		HWIN hWin = WsCreateScroll(gHWin, style, 
										scrollRect.x(), scrollRect.y(), 
										scrollRect.width(), scrollRect.height(), 
										NULL);

		setShpWidgetHandle(hWin);

        if (WmIsWindowValid(shpWidgetHandle()))
        {
            //The last param to WsSetScrollRange should be non zero to avoid divide by zero
		    WsSetScrollRange(shpWidgetHandle(), 1, 1, 1, 1);
		    WmDrawWindow(shpWidgetHandle(), TRUE);
        }
	}
#endif // USE_PLATFORM_SCROLLBARS
}

#ifdef USE_PLATFORM_SCROLLBARS
//this constructor is added for iframeviews
PlatformScrollbar::PlatformScrollbar(ScrollbarClient* client, ScrollbarOrientation orientation, IntRect& rect)
	: Scrollbar(client, orientation, RegularScrollbar), 
	  m_hoveredPart(NoPart), m_pressedPart(NoPart), m_pressedPos(0),
      m_scrollTimer(this, &PlatformScrollbar::autoscrollTimerFired),
      m_overlapsResizer(false),
	  m_parent(NULL)
{
	int style;

	if(HorizontalScrollbar == orientation)
	{
		style = SCROLL_STYLE_HORIZONTAL;
		rect.setHeight(PlatformScrollbar::horizontalScrollbarHeight());
		rect.setWidth(rect.width() - PlatformScrollbar::verticalScrollbarWidth());
	}
	else
	{
		style = SCROLL_STYLE_VERTICAL;
		rect.setWidth(PlatformScrollbar::verticalScrollbarWidth());
		rect.setHeight(rect.height() - PlatformScrollbar::horizontalScrollbarHeight());
	}

	setFrameGeometry(rect);

	if (rect.height() > 0 && rect.width() > 0)
	{
		
		HWIN hWin = WsCreateScroll(gHWin, style, 
											0, 0, 
											rect.width(), rect.height(), 
											NULL);
		m_hGC = WmCreateWindowGC(hWin);

		setShpWidgetHandle(hWin);
		setGCHandle(m_hGC);
	
		WmSetWindowPos(hWin, rect.x(), rect.y());
	}
}
#endif // USE_PLATFORM_SCROLLBARS

PlatformScrollbar::~PlatformScrollbar()
{
	//vamsi - check reqd due to F480 UI flow.
#ifdef USE_PLATFORM_SCROLLBARS
	if(WmIsWindowValid(shpWidgetHandle()) )
		WmDestroyWindow(shpWidgetHandle());
#endif // USE_PLATFORM_SCROLLBARS
}

void PlatformScrollbar::updateThumbPosition()
{
#ifdef USE_PLATFORM_SCROLLBARS
	if(m_totalSize && WmIsWindowValid(shpWidgetHandle()))
	{
		WsSetScrollRange(shpWidgetHandle(), 0, m_totalSize-m_visibleSize, m_visibleSize, m_totalSize);
		WsSetScrollPos(shpWidgetHandle(), m_currentPos);
	}
#else
    m_thumbRect = thumbRect();
#endif // USE_PLATFORM_SCROLLBARS
}

void PlatformScrollbar::updateThumbProportion()
{
#ifdef USE_PLATFORM_SCROLLBARS
	notImplemented();
#else
	m_thumbRect = thumbRect();
#endif
}

void PlatformScrollbar::invalidateTrack()
{
	notImplemented();
}

void PlatformScrollbar::invalidatePart(ScrollbarPart part)
{
	notImplemented();
}

int PlatformScrollbar::width() const
{
    return Widget::width();
}

int PlatformScrollbar::height() const
{
    return Widget::height();
}

void PlatformScrollbar::setRect(const IntRect& rect)
{
	IntRect oldRect = frameGeometry();
	IntRect givenRect = rect;

	if(HorizontalScrollbar == m_orientation )
	{
		//prevent zero height of horizontal scrollbar
		//changed the check to less than equal to zero so that we do not get negative values for height- CQ bug 1741 : Hema
		if(givenRect.height() <= 0)
			givenRect.setHeight(PlatformScrollbar::horizontalScrollbarHeight());
	}
	else 
	{
		//prevent zero width of vertical scrollbar
		//changed the check to less than equal to zero so that we do not get negative values for width- CQ bug 1741 : Hema
		if(givenRect.width() <= 0)
			givenRect.setWidth(PlatformScrollbar::verticalScrollbarWidth());
	}

	if(oldRect != givenRect)
	{
		setFrameGeometry(givenRect);
		geometryChanged();
	}
}

void PlatformScrollbar::geometryChanged()
{
#ifdef USE_PLATFORM_SCROLLBARS
	HWIN hWin = shpWidgetHandle();
    
	if(WmIsWindowValid(hWin))
	{
		IntRect rect = frameGeometry();
		WmRect wndRect;
		wndRect.x = rect.x();
		wndRect.y = rect.y();
		wndRect.w = rect.width();
		wndRect.h = rect.height();
		
		// Check included for CQ--1948 Thariq
		if(wndRect.w < 0)
			wndRect.w = 0;
		if(wndRect.h < 0)
			wndRect.h = 0;

		WmSetWindowPos(hWin, wndRect.x, wndRect.y);
		WmSetWindowRect(hWin, &wndRect);
	}
#else
    m_thumbRect = thumbRect();
#endif // USE_PLATFORM_SCROLLBARS
}

void PlatformScrollbar::setParent(ScrollView* parentView)
{
	m_parent = parentView;
}

void PlatformScrollbar::removeFromParent()
{
	if(m_parent)
	m_parent->removeChild(this);
}

// 2008.07.25 modified for scroll touch. 
bool PlatformScrollbar::isEnabled() const
{
    return TRUE;
}

void PlatformScrollbar::setEnabled(bool enabled)
{
	notImplemented();
}

void PlatformScrollbar::paint(GraphicsContext* graphicsContext, const IntRect& damageRect)
{
#ifdef USE_PLATFORM_SCROLLBARS
   if (graphicsContext->paintingDisabled() || !WmIsWindowValid(shpWidgetHandle()))
        return;
	updateThumbPosition();
	WmDrawWindow(shpWidgetHandle(), TRUE);
#else
    if(graphicsContext->paintingDisabled())
		return;

	paintTrack(graphicsContext, frameGeometry(), true, damageRect);
    paintThumb(graphicsContext, m_thumbRect, damageRect);
#endif // USE_PLATFORM_SCROLLBARS
}

IntRect PlatformScrollbar::backButtonRect() const
{
	notImplemented();
    return IntRect(0, 0, 0, 0);
}

IntRect PlatformScrollbar::forwardButtonRect() const
{
	notImplemented();
    return IntRect(0, 0, 0, 0);
}
// 2008.07.30  bs0425.lee modified for touch scrollbar. 
IntRect PlatformScrollbar::trackRect() const
{
 	if (m_orientation == HorizontalScrollbar) 
 	{
 	//if (!hasButtons())
            return IntRect(x(), y(), width(), HORIZONTAL_SCROLLBAR_HEIGHT);
       /* return IntRect(x() + cButtonLength[controlSize()], y(), width() - 2 * cButtonLength[controlSize()], cHorizontalHeight[controlSize()]);*/
}

  //if (!hasButtons())
    return IntRect(x(), y(), VERTICAL_SCROLLBAR_WIDTH,height());
   /* return IntRect(x(), y() + cButtonLength[controlSize()], cVerticalWidth[controlSize()], height() - 2 * cButtonLength[controlSize()]);*/
}


// 2008.07.30   bs0425.lee modified for touch scrollbar. 

IntRect PlatformScrollbar::thumbRect() const
{
#ifdef USE_PLATFORM_SCROLLBARS
	notImplemented();
    return IntRect(0, 0, 0, 0);
#else
    // Set the rect for thumb as the scrollbar track rect
    // We only need to change the thumb length(width/height) and position(x/y)
    if(!m_totalSize)
	{
		return IntRect(0, 0, 0, 0);
	}

	IntRect rect(frameGeometry());

    if(HorizontalScrollbar == m_orientation)
    {
        // Calculate the thumb width relative to the total page width and the visible width
        rect.setWidth((m_visibleSize * rect.width()) / m_totalSize);

        if(rect.width() < MIN_THUMB_SIZE)
        {
            // we restrict the thumb size to be not less than MIN_THUMB_SIZE
            // hence we need to subtract an offset from the thumb position in this case
            int thumbOffset = ((MIN_THUMB_SIZE - rect.width()) * m_currentPos) / m_totalSize;
            rect.setWidth(MIN_THUMB_SIZE);
            rect.setX(((trackLength() * m_currentPos) / m_totalSize) - thumbOffset);
        }
        else
        {
            rect.setX((trackLength() * m_currentPos) / m_totalSize);
        }
    }
    else
    {
        // Calculate the thumb height relative to the total page height and the visible height
        rect.setHeight((m_visibleSize * rect.height()) / m_totalSize);

        if(rect.height() < MIN_THUMB_SIZE)
        {
            // we restrict the thumb size to be not less than MIN_THUMB_SIZE
            // hence we need to subtract an offset from the thumb position in this case
            int thumbOffset = ((MIN_THUMB_SIZE - rect.height()) * m_currentPos) / m_totalSize;
            rect.setHeight(MIN_THUMB_SIZE);
            rect.setY(((trackLength() * m_currentPos) / m_totalSize) - thumbOffset);
        }
        else
        {
            rect.setY((trackLength() * m_currentPos) / m_totalSize);
        }
    }

    return rect;
#endif
}


// 2008.07.28 bs0425.lee modified for scroll touch.
void PlatformScrollbar::splitTrack(const IntRect& trackRect, IntRect& beforeThumbRect, IntRect& thumbRect, IntRect& afterThumbRect) const
{
	  // This function won't even get called unless we're big enough to have some combination of these three rects where at least
    // one of them is non-empty.
   // int thumbPos = thumbPosition();
  if (m_orientation == HorizontalScrollbar) {
        thumbRect = m_thumbRect;
        beforeThumbRect = IntRect(trackRect.x(), trackRect.y(), m_thumbRect.x(), trackRect.height());
        afterThumbRect = IntRect(m_thumbRect.x() + m_thumbRect.width(), trackRect.y(), trackRect.width()-m_thumbRect.x()- m_thumbRect.width(), trackRect.height());
    } else {
        thumbRect =m_thumbRect;
        beforeThumbRect = IntRect(trackRect.x(), trackRect.y(), trackRect.width(), m_thumbRect.y());
        afterThumbRect = IntRect(trackRect.x(), m_thumbRect.y() + m_thumbRect.height(), trackRect.width(), trackRect.bottom() - m_thumbRect.bottom());
    }
}

int PlatformScrollbar::thumbPosition() const
{
	notImplemented();
    return 0;
}

int PlatformScrollbar::thumbLength() const
{
	notImplemented();
	return 0;
}

int PlatformScrollbar::trackLength() const
{
#ifdef USE_PLATFORM_SCROLLBARS
	notImplemented();
	return 0;
#else
    // track length would be height of the vertical scroll bar
    // and widht of the horizontal scroll bar
    int length = frameGeometry().height();
    if(HorizontalScrollbar == m_orientation)
    {
        length = frameGeometry().width();
    }

    return length;
#endif // USE_PLATFORM_SCROLLBARS
}

void PlatformScrollbar::paintButton(GraphicsContext* context, const IntRect& rect, bool start, const IntRect& damageRect) const
{
	notImplemented();
}

void PlatformScrollbar::paintTrack(GraphicsContext* context, const IntRect& rect, bool start, const IntRect& damageRect) const
{
#ifdef USE_PLATFORM_SCROLLBARS
	notImplemented();
#else
    context->fillRect(rect, Color(WHITE_COLOR));
	
	IntRect trackRect(rect);
	
	if(HorizontalScrollbar == m_orientation)
	{
		trackRect.setHeight(1);
		
		//1 paint upper shadow
		trackRect.setY(trackRect.y() + 2);
		context->fillRect(trackRect, Color(TRACK_SHADOW_COLOR));

		//2 paint track
		trackRect.setY(trackRect.y() + 1);
		context->fillRect(trackRect, Color(TRACK_COLOR));

		//3 paint lower shadow
		trackRect.setY(trackRect.y() + 1);
		context->fillRect(trackRect, Color(TRACK_SHADOW_COLOR));
	}
	else
	{
		trackRect.setWidth(1);
		
		//1 paint left shadow
		trackRect.setX(trackRect.x() + 2);
		context->fillRect(trackRect, Color(TRACK_SHADOW_COLOR));

		//2 paint track
		trackRect.setX(trackRect.x() + 1);
		context->fillRect(trackRect, Color(TRACK_COLOR));

		//3 paint right shadow
		trackRect.setX(trackRect.x() + 1);
		context->fillRect(trackRect, Color(TRACK_SHADOW_COLOR));
	}
#endif // USE_PLATFORM_SCROLLBARS
}

void PlatformScrollbar::paintThumb(GraphicsContext* context, const IntRect& rect, const IntRect& damageRect) const
{
#ifdef USE_PLATFORM_SCROLLBARS
	notImplemented();
#else
	IntRect thumbRect(rect);

	if(HorizontalScrollbar == m_orientation)
	{
		thumbRect.setY(thumbRect.y() + 1);
		context->fillRect(thumbRect, Color(THUMB_BORDER_COLOR));

		thumbRect.setHeight(1);

		thumbRect.setX(thumbRect.x() + 1);
		thumbRect.setWidth(thumbRect.width() - 2);
		
		//1 paint thumb
		thumbRect.setY(thumbRect.y() + 1);
		context->fillRect(thumbRect, Color(THUMB_COLOR));

		//2 paint thumb gradient
		thumbRect.setY(thumbRect.y() + 1);
		context->fillRect(thumbRect, Color(THUMB_LIGHT_COLOR));

		//3 paint thumb
		thumbRect.setY(thumbRect.y() + 1);
		context->fillRect(thumbRect, Color(THUMB_COLOR));
	}
	else
	{
		thumbRect.setX(thumbRect.x() + 1);
		context->fillRect(thumbRect, Color(THUMB_BORDER_COLOR));

		thumbRect.setWidth(1);

		thumbRect.setY(thumbRect.y() + 1);
		thumbRect.setHeight(thumbRect.height() - 2);
		
		//1 paint thumb
		thumbRect.setX(thumbRect.x() + 1);
		context->fillRect(thumbRect, Color(THUMB_COLOR));

		//2 paint thumb gradient
		thumbRect.setX(thumbRect.x() + 1);
		context->fillRect(thumbRect, Color(THUMB_LIGHT_COLOR));

		//3 paint thumb
		thumbRect.setX(thumbRect.x() + 1);
		context->fillRect(thumbRect, Color(THUMB_COLOR));
	}
#endif // USE_PLATFORM_SCROLLBARS
}

ScrollbarPart PlatformScrollbar::hitTest(const PlatformMouseEvent& evt)
{
    ScrollbarPart result = NoPart;
	notImplemented();
    return result;
}

bool PlatformScrollbar::handleMouseMoveEvent(const PlatformMouseEvent& evt)
{
	notImplemented();
	return false;
}


bool PlatformScrollbar::handleMouseOutEvent(const PlatformMouseEvent& evt)
{
	notImplemented();
	return false;
}

bool PlatformScrollbar::handleMousePressEvent(const PlatformMouseEvent& evt)
{
	notImplemented();
	return false;
}

bool PlatformScrollbar::handleMouseReleaseEvent(const PlatformMouseEvent& evt)
{
	notImplemented();
	return false;
}

void PlatformScrollbar::startTimerIfNeeded(double delay)
{
	notImplemented();
}

void PlatformScrollbar::stopTimerIfNeeded()
{
	notImplemented();
}

void PlatformScrollbar::autoscrollPressedPart(double delay)
{
	notImplemented();
}

void PlatformScrollbar::autoscrollTimerFired(Timer<PlatformScrollbar>*)
{
	notImplemented();
}

ScrollDirection PlatformScrollbar::pressedPartScrollDirection()
{
	notImplemented();
    return ScrollDown;
}

ScrollGranularity PlatformScrollbar::pressedPartScrollGranularity()
{
	notImplemented();
    return ScrollByLine;
}

bool PlatformScrollbar::thumbUnderMouse()
{
	notImplemented();
	return false;
}

int PlatformScrollbar::horizontalScrollbarHeight(ScrollbarControlSize controlSize)
{
	return HORIZONTAL_SCROLLBAR_HEIGHT;
}

int PlatformScrollbar::verticalScrollbarWidth(ScrollbarControlSize controlSize)
{
	return VERTICAL_SCROLLBAR_WIDTH;
}

int PlatformScrollbar::minThumbSize()
{
#ifdef USE_PLATFORM_SCROLLBARS
	return 0;
#else
	return MIN_THUMB_SIZE;
#endif
}

IntRect PlatformScrollbar::windowClipRect() const
{
    IntRect clipRect(0, 0, width(), height());
	notImplemented();
    return clipRect;
}

void PlatformScrollbar::paintGripper(HDC hdc, const IntRect& rect) const
{
	notImplemented();
}

IntRect PlatformScrollbar::gripperRect(const IntRect& thumbRect) const
{
    return IntRect();
}

void PlatformScrollbar::themeChanged()
{
}

}

