/*
 * Copyright (C) 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2008 Collabora, Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
#ifdef SUPPORT_NP_PLUGIN

#include "config.h"
#include "ShpJasmine.h"
#include "PluginStream.h"

#include "CString.h"
#include "DocumentLoader.h"
#include "Frame.h"
#include "FrameLoader.h"
#include "PluginDebug.h"
#include "SharedBuffer.h"
#include "SubresourceLoader.h"

#if PLATFORM(SHP)
#include "PluginView.h"
#include "PluginDatabase.h"
#define TEMP_FILE_PATH "/User/Br/Plugin/"
#endif // PLATFORM(SHP)

// We use -2 here because some plugins like to return -1 to indicate error
// and this way we won't clash with them.
static const int WebReasonNone = -2;

using std::max;
using std::min;

namespace WebCore {

typedef HashMap<NPStream*, NPP> StreamMap;

static StreamMap staticStreams;

#ifdef SFB_GLOBAL_STATIC_HANDLING

void deInitStreamMap()
{
    staticStreams.clear();
}

#endif


static StreamMap& streams()
{

    return staticStreams;
}

PluginStream::PluginStream(PluginStreamClient* client, Frame* frame, const ResourceRequest& resourceRequest, bool sendNotification, void* notifyData, const NPPluginFuncs* pluginFuncs, NPP instance, const PluginQuirkSet& quirks)
    : RefCounted<PluginStream>(0)
    , m_resourceRequest(resourceRequest)
    , m_client(client)
    , m_frame(frame)
    , m_notifyData(notifyData)
    , m_sendNotification(sendNotification)
    , m_streamState(StreamBeforeStarted)
    , m_loadManually(false)
    , m_delayDeliveryTimer(this, &PluginStream::delayDeliveryTimerFired)
    , m_deliveryData(0)
    , m_tempFileHandle(invalidPlatformFileHandle)
    , m_pluginFuncs(pluginFuncs)
    , m_instance(instance)
    , m_quirks(quirks)
#if PLATFORM(SHP)
	, m_length(0)
#endif
{
    ASSERT(m_instance);
	PRINT("PG : PluginStream::PluginStream ENTERED \n");

    m_stream.url = 0;
    m_stream.ndata = 0;
    m_stream.pdata = 0;
    m_stream.end = 0;
    m_stream.notifyData = 0;
    m_stream.lastmodified = 0;

    streams().add(&m_stream, m_instance);
	PRINT("PG : PluginStream::PluginStream EXITS \n");
}

PluginStream::~PluginStream()
{
	PRINT("PG : PluginStream::~PluginStream ENTERED \n");
 
	ASSERT(m_streamState != StreamStarted);
    ASSERT(!m_loader);
	SysDebug((MID_BROWSER, "PG : PluginStream::~PluginStream destroyed stream pointer 0x%08x \n", this ));

	if(m_stream.url)
	    free((char*)m_stream.url);
	// BUG SFB00001336, SFB00001369 
	if(m_loader)
	{
		stop();
	}
	streams().remove(&m_stream);
	PRINT("PG : PluginStream::~PluginStream EXITS \n");

}

void PluginStream::start()
{
    ASSERT(!m_loadManually);
	PRINT("PG : PluginStream::start ENTERED \n");

	SysDebug((MID_BROWSER, "PG : PluginStream::start NetscapePlugInStreamLoader created for plugin stream pointer 0x%08x \n", this ));


    m_loader = NetscapePlugInStreamLoader::create(m_frame, this);

    m_loader->setShouldBufferData(false);
    m_loader->documentLoader()->addPlugInStreamLoader(m_loader.get());
    m_loader->load(m_resourceRequest);
	PRINT("PG : PluginStream::start EXITS \n");

}

void PluginStream::stop()
{
	PRINT("PG : PluginStream::stop ENTERED \n");

#if PLATFORM(SHP)&& USE(LOW_BANDWIDTH_DISPLAY)
    if(m_frame && m_frame->loader())
    {
        m_frame->loader()->loadDone();
    }
#endif

    m_streamState = StreamStopped;

    if (m_loadManually) {
        ASSERT(!m_loader);

        DocumentLoader* documentLoader = m_frame->loader()->activeDocumentLoader();
        ASSERT(documentLoader);

        if (documentLoader->isLoadingMainResource())
            documentLoader->cancelMainResourceLoad(m_frame->loader()->cancelledError(m_resourceRequest));

        return;
    }

    if (m_loader) {
        m_loader->cancel();
        m_loader = 0;
    }
	PRINT("PG : PluginStream::stop EXITS \n");
}

void PluginStream::startStream()
{
	PRINT("PG : PluginStream::startStream ENTERED \n");
	ASSERT(m_streamState == StreamBeforeStarted);

    const KURL& responseURL = m_resourceResponse.url();

    // Some plugins (Flash) expect that javascript URLs are passed back decoded as this is the
    // format used when requesting the URL.
    if (responseURL.string().startsWith("javascript:", false))
#if PLATFORM(SHP)   // strdup is not supported in RVCT
	{
		m_stream.url = (char* ) malloc(AcStrlen(responseURL.decode_string(responseURL.deprecatedString()).utf8()) + 1);
		AcStrcpy( (char*) m_stream.url, responseURL.decode_string(responseURL.deprecatedString()).utf8());
	}
#else
        m_stream.url = strdup(responseURL.decode_string(responseURL.deprecatedString()).utf8());
#endif
    else
#if PLATFORM(SHP)
	{
		m_stream.url = (char* ) malloc(AcStrlen(responseURL.decode_string(responseURL.deprecatedString()).utf8()) + 1);
		AcStrcpy( (char*) m_stream.url, responseURL.decode_string(responseURL.deprecatedString()).utf8());
	}
#else
        m_stream.url = strdup(responseURL.deprecatedString().utf8());
#endif
    CString mimeTypeStr = m_resourceResponse.mimeType().utf8();
    
    long long expectedContentLength = m_resourceResponse.expectedContentLength();

    if (m_resourceResponse.isHTTP()) {
        Vector<UChar> stringBuilder;
        String separator(": ");

        HTTPHeaderMap::const_iterator end = m_resourceResponse.httpHeaderFields().end();
        for (HTTPHeaderMap::const_iterator it = m_resourceResponse.httpHeaderFields().begin(); it != end; ++it) {
            stringBuilder.append(it->first.characters(), it->first.length());
            stringBuilder.append(separator.characters(), separator.length());
            stringBuilder.append(it->second.characters(), it->second.length());
            stringBuilder.append((UChar)'\n');
        }

        m_headers = String::adopt(stringBuilder).utf8();

        // If the content is encoded (most likely compressed), then don't send its length to the plugin,
        // which is only interested in the decoded length, not yet known at the moment.
        // <rdar://problem/4470599> tracks a request for -[NSURLResponse expectedContentLength] to incorporate this logic.
        String contentEncoding = m_resourceResponse.httpHeaderField("Content-Encoding");
        if (!contentEncoding.isNull() && contentEncoding != "identity")
            expectedContentLength = -1;
    }

    m_stream.headers = m_headers.data();
    m_stream.pdata = 0;
    m_stream.ndata = this;
    m_stream.end = max(expectedContentLength, 0LL);
    m_stream.lastmodified = m_resourceResponse.lastModifiedDate();
    m_stream.notifyData = m_notifyData;

	m_transferMode = NP_NORMAL;
    m_offset = 0;
    m_reason = WebReasonNone;

    // Protect the stream if destroystream is called from within the newstream handler
    RefPtr<PluginStream> protect(this);

    NPError npErr = m_pluginFuncs->newstream(m_instance, (NPMIMEType)mimeTypeStr.data(), &m_stream, false, &m_transferMode);
    
    // If the stream was destroyed in the call to newstream we return
    if (m_reason != WebReasonNone)
        return;
        
    m_streamState = StreamStarted;

    if (npErr != NPERR_NO_ERROR)
        cancelAndDestroyStream(npErr);

    if (m_transferMode == NP_NORMAL)
        return;

#if PLATFORM(SHP) 
	//<TODO>::Need to give the multiple file save support.
	/* PG : if the File already exists, delete the file */
	char tempFile[256];
	strcpy(tempFile, TEMP_FILE_PATH);
    strcat(tempFile, "Play.dat");

	if(fileExists(tempFile))
	{
		deleteFile(tempFile);
	}
	m_path = openFile(TEMP_FILE_PATH,"Play.dat", FILE_MODE_CREATE |
												FILE_MODE_READ |
												FILE_MODE_WRITE, m_tempFileHandle);
	m_stream.FilePath = m_path.data();
#else
	m_path = openTemporaryFile("WKP", m_tempFileHandle);
#endif

    // Something went wrong, cancel loading the stream
    if (!isHandleValid(m_tempFileHandle))
        cancelAndDestroyStream(NPRES_NETWORK_ERR);
	PRINT("PG : PluginStream::startStream EXITS \n");
}

NPP PluginStream::ownerForStream(NPStream* stream)
{
	PRINT("PG : PluginStream::ownerForStream ENTERED \n");
	PRINT("PG : PluginStream::ownerForStream EXITS \n");
    return streams().get(stream);
}

void PluginStream::cancelAndDestroyStream(NPReason reason)
{
    RefPtr<PluginStream> protect(this);
	PRINT("PG : PluginStream::cancelAndDestroyStream ENTERED \n");

    destroyStream(reason);
    stop();
	PRINT("PG : PluginStream::cancelAndDestroyStream EXITS \n");
}

void PluginStream::destroyStream(NPReason reason)
{
	PRINT("PG : PluginStream::destroyStream Overloaded (reason) ENTERED \n");
	m_reason = reason;
	
#if PLATFORM(SHP)	
	if (m_reason != NPRES_DONE) {
        // Stop any pending data from being streamed
        m_deliveryData = NULL;
		m_length = 0;
    } else if (m_deliveryData && m_length > 0) {
        // There is more data to be streamed, don't destroy the stream now.
		return;
    }
#else
	if (m_reason != NPRES_DONE) {
        // Stop any pending data from being streamed
        if (m_deliveryData)
            m_deliveryData->resize(0);
    } else if (m_deliveryData && m_deliveryData->size() > 0) {
        // There is more data to be streamed, don't destroy the stream now.
        return;
    }
#endif//PLATFORM(SHP)
    
    destroyStream();
	PRINT("PG : PluginStream::destroyStream Overloaded (reason) EXITS \n");
}

void PluginStream::destroyStream()
{
	PRINT("PG : PluginStream::destroyStream ENTERED \n");
	if (m_streamState == StreamStopped)
        return;

    ASSERT (m_reason != WebReasonNone);

#if PLATFORM(SHP)
	ASSERT (!m_deliveryData || m_length == 0);
#else
	ASSERT (!m_deliveryData || m_deliveryData->size() == 0);
#endif//PLATFORM(SHP)

#if !PLATFORM(SHP)
	/* This check is to fix a Crash on target : trying to destroy 
	 * invalid FILE handle */
	if(m_tempFileHandle)
	{
		closeFile(m_tempFileHandle);
	}
	else
	{
		PRINT("PG : PluginStream::destroyStream  trying to destroy invalid FILE Handle \n");
	}
#endif
    bool newStreamCalled = m_stream.ndata;

    if (newStreamCalled) {
#if !PLATFORM(SHP)
			if (m_reason == NPRES_DONE && (m_transferMode == NP_ASFILE || m_transferMode == NP_ASFILEONLY)) {
		        ASSERT(!m_path.isNull());

	            m_pluginFuncs->asfile(m_instance, &m_stream, m_path.data());
	      }
#endif
        NPError npErr = m_pluginFuncs->destroystream(m_instance, &m_stream, m_reason);
        LOG_NPERROR(npErr);

        m_stream.ndata = 0;
    }

    if (m_sendNotification) {
        // Flash 9 can dereference null if we call NPP_URLNotify without first calling NPP_NewStream
        // for requests made with NPN_PostURLNotify; see <rdar://5588807>
        if (!newStreamCalled && m_quirks.contains(PluginQuirkFlashURLNotifyBug) &&
            equalIgnoringCase(m_resourceRequest.httpMethod(), "POST")) {
            // Protect the stream if NPN_DestroyStream is called from NPP_NewStream
            RefPtr<PluginStream> protect(this);

            m_transferMode = NP_NORMAL;
            m_stream.url = "";
            m_stream.notifyData = m_notifyData;

            m_pluginFuncs->newstream(m_instance, "", &m_stream, false, &m_transferMode);
            m_pluginFuncs->destroystream(m_instance, &m_stream, m_reason);

            // in successful requests, the URL is dynamically allocated and freed in our
            // destructor, so reset it to 0
            m_stream.url = 0;
        }
        m_pluginFuncs->urlnotify(m_instance, m_resourceRequest.url().deprecatedString().utf8(), m_reason, m_notifyData);
    }

    m_streamState = StreamStopped;

    // streamDidFinishLoading can cause us to be deleted.
    RefPtr<PluginStream> protect(this);
    if (!m_loadManually)
        m_client->streamDidFinishLoading(this);

    if (!m_path.isNull()) {
//        String tempFilePath = String::fromUTF8(m_path.data());
		String tempFilePath = String::fromUTF8(m_stream.FilePath);
		deleteFile(tempFilePath);
    }
	PRINT("PG : PluginStream::destroyStream EXITS \n");
}

void PluginStream::delayDeliveryTimerFired(Timer<PluginStream>* timer)
{
	PRINT("PG : PluginStream::delayDeliveryTimerFired ENTERED \n");

    ASSERT(timer == &m_delayDeliveryTimer);

    deliverData();
	PRINT("PG : PluginStream::delayDeliveryTimerFired EXITS \n");

}

void PluginStream::deliverData()
{
	PRINT("PG : PluginStream::deliverData ENTERED \n");
    ASSERT(m_deliveryData);
    
    if (m_streamState == StreamStopped)
        // FIXME: We should cancel our job in the SubresourceLoader on error so we don't reach this case
        return;

    ASSERT(m_streamState != StreamBeforeStarted);

#if PLATFORM(SHP)    
	if (!m_stream.ndata || m_length == 0)
#else
	if (!m_stream.ndata || m_deliveryData->size() == 0)
#endif//PLATFORM(SHP)
        return;

#if PLATFORM(SHP)
	int32 totalBytes = m_length;
#else
	int32 totalBytes = m_deliveryData->size();
#endif//PLATFORM(SHP)

    int32 totalBytesDelivered = 0;

    while (totalBytesDelivered < totalBytes) {
        int32 deliveryBytes = m_pluginFuncs->writeready(m_instance, &m_stream);

        if (deliveryBytes <= 0) {
            m_delayDeliveryTimer.startOneShot(0);
            break;
        } else {
            deliveryBytes = min(deliveryBytes, totalBytes - totalBytesDelivered);
            int32 dataLength = deliveryBytes;

#if PLATFORM(SHP)
            const char* data = m_deliveryData + totalBytesDelivered;
#else
			char* data = m_deliveryData->data() + totalBytesDelivered;
#endif//PLATFORM(SHP)

            // Write the data
            deliveryBytes = m_pluginFuncs->write(m_instance, &m_stream, m_offset, dataLength, (void*)data);
            if (deliveryBytes < 0) {
                LOG_PLUGIN_NET_ERROR();
                cancelAndDestroyStream(NPRES_NETWORK_ERR);
                return;
            }
            deliveryBytes = min(deliveryBytes, dataLength);
            m_offset += deliveryBytes;
            totalBytesDelivered += deliveryBytes;
        }
    }

    if (totalBytesDelivered > 0) {
        if (totalBytesDelivered < totalBytes) {
            int remainingBytes = totalBytes - totalBytesDelivered;
#if !PLATFORM(SHP) 
            memmove(m_deliveryData->data(), m_deliveryData->data() + totalBytesDelivered, remainingBytes);
            m_deliveryData->resize(remainingBytes);
#endif//#if !PLATFORM(SHP)
        } else {
#if PLATFORM(SHP)
		m_deliveryData = NULL;//Not owned by Webcore so no need to free assiging it to NULL.
		m_length = 0;
#else
		m_deliveryData->resize(0);
#endif//!PLATFORM(SHP)
            if (m_reason != WebReasonNone)
                destroyStream();
        }
    } 
	PRINT("PG : PluginStream::deliverData EXITS \n");
}

void PluginStream::sendJavaScriptStream(const KURL& requestURL, const CString& resultString)
{
	PRINT("PG : PluginStream::sendJavaScriptStream ENTERED \n");

    didReceiveResponse(0, ResourceResponse(requestURL, "text/plain", resultString.length(), "", ""));

    if (m_streamState == StreamStopped)
        return;

    if (!resultString.isNull()) {
        didReceiveData(0, resultString.data(), resultString.length());
        if (m_streamState == StreamStopped)
            return;
    }

    m_loader = 0;

    destroyStream(resultString.isNull() ? NPRES_NETWORK_ERR : NPRES_DONE);
	PRINT("PG : PluginStream::sendJavaScriptStream EXITS \n");

}

void PluginStream::didReceiveResponse(NetscapePlugInStreamLoader* loader, const ResourceResponse& response)
{
	PRINT("PG :  PluginStream::didReceiveResponse ENTERS \n");
	ASSERT(loader == m_loader);
    ASSERT(m_streamState == StreamBeforeStarted);


    m_resourceResponse = response;

#if PLATFORM(SHP)
	// get the platform plugin depending on the MIME type
	/* To fix crash : when the Phone is in Silent mode, Plugin module
	 * returns ERROR */
	NPError err = initPlugin();
#endif
	if (err == NPERR_NO_ERROR)
	{
	startStream();
	}
	PRINT("PG :  PluginStream::didReceiveResponse EXITS \n");
}

void PluginStream::didReceiveData(NetscapePlugInStreamLoader* loader, const char* data, int length)
{
	PRINT("PG :  PluginStream::didReceiveData ENTERED \n");
    ASSERT(loader == m_loader);
    ASSERT(length > 0);
    ASSERT(m_streamState == StreamStarted);

    // If the plug-in cancels the stream in deliverData it could be deleted, 
    // so protect it here.
    RefPtr<PluginStream> protect(this);
	
#if PLATFORM(SHP)    
	

	if (m_transferMode != NP_ASFILEONLY) {
		m_length = length;
		
		m_deliveryData = data;
		
		deliverData();
	}
#else	
	if (m_transferMode != NP_ASFILEONLY) {
        if (!m_deliveryData)
            m_deliveryData.set(new Vector<char>);

        int oldSize = m_deliveryData->size();
        m_deliveryData->resize(oldSize + length);
        memcpy(m_deliveryData->data() + oldSize, data, length);

        deliverData();
    }
#endif//PLATFORM(SHP)

    if (m_streamState != StreamStopped && isHandleValid(m_tempFileHandle)) {
        int bytesWritten = writeToFile(m_tempFileHandle, data, length);
        if (bytesWritten != length)
            cancelAndDestroyStream(NPRES_NETWORK_ERR);
    }
	PRINT("PG :  PluginStream::didReceiveData EXITS \n");
}

void PluginStream::didFail(NetscapePlugInStreamLoader* loader, const ResourceError&)
{
#if PLATFORM(SHP)
	//Plugin stream Object gets destroyed in destroyStream(NPRES_DONE)
	//please be careful while using member variables
	Frame *frame = m_frame ;
	RefPtr<Frame> protector(frame);
#endif //PLATFORM(SHP)
	PRINT("PG :  PluginStream::didFail ENTERED \n");

    ASSERT(loader == m_loader);

    m_loader = 0;

    LOG_PLUGIN_NET_ERROR();
    destroyStream(NPRES_NETWORK_ERR);

#if PLATFORM(SHP)&& USE(LOW_BANDWIDTH_DISPLAY)
    if(frame && frame->loader())
    {
        frame->loader()->loadDone();
    }
#endif

	PRINT("PG :  PluginStream::didFail EXITS \n");

}

void PluginStream::didFinishLoading(NetscapePlugInStreamLoader* loader)
{
	PRINT("PG :  PluginStream::didFinishLoading ENTERED \n");

    ASSERT(loader == m_loader);
    ASSERT(m_streamState == StreamStarted);

#if PLATFORM(SHP)
	//Plugin stream Object gets destroyed in destroyStream(NPRES_DONE)
	//please be careful while using member variables
	Frame *frame = m_frame ;
	RefPtr<Frame> protector(frame);
#endif //PLATFORM(SHP)
    m_loader = 0;

#if PLATFORM(SHP)
	if(m_transferMode == NP_ASFILEONLY)
	{
	if(m_tempFileHandle)
	{
		closeFile(m_tempFileHandle);
	}
	else
	{
		PRINT("PG : PluginStream::destroyStream  trying to destroy invalid FILE Handle \n");
	}
    bool newStreamCalled = m_stream.ndata;
	m_reason = NPRES_DONE;
    if (newStreamCalled) {
        if (m_reason == NPRES_DONE && (m_transferMode == NP_ASFILE || m_transferMode == NP_ASFILEONLY)) {
            ASSERT(!m_path.isNull());

            m_pluginFuncs->asfile(m_instance, &m_stream, m_path.data());
        }
	}
	}
	else
	{
#endif
	 destroyStream(NPRES_DONE);
#if PLATFORM(SHP)
	}
#endif

#if PLATFORM(SHP)&& USE(LOW_BANDWIDTH_DISPLAY)
    if(frame && frame->loader())
    {
        frame->loader()->loadDone();
    }
#endif
	PRINT("PG :  PluginStream::didFinishLoading EXITS \n");
}

#if PLATFORM(SHP)
NPError PluginStream::initPlugin()
{
	PRINT("PG :  PluginStream::initPlugin ENTERS \n");
	PluginView* view = (PluginView*)m_client;
	
	if(!view->plugin()){
		PRINT("PG :  PluginStream::initPlugin find MIME type matching plugin\n");
		String mimeTypeCopy(m_resourceResponse.mimeType().latin1().mutableData());
		PluginPackage* plugin = PluginDatabase::installedPlugins()->findPlugin(m_stream.url, mimeTypeCopy);			
		view->setPlugin(plugin);
	}

    if (!view->plugin()) {
        ASSERT(view->status() == PluginStatusCanNotFindPlugin);
        return NPERR_MIME_NOT_SUPPORTED;
    }

    if (!view->plugin()->load()) {
        view->setPlugin(0);
        view->setStatus(PluginStatusCanNotLoadPlugin);
        return NPERR_MODULE_LOAD_FAILED_ERROR;
    }

	m_pluginFuncs= view->plugin()->pluginFuncs(); //setting the m_pluginFuncs

    if (!view->start()) {
        view->setStatus(PluginStatusCanNotLoadPlugin);
        return NPERR_MODULE_LOAD_FAILED_ERROR;
    }
	
	view->setStatus(PluginStatusLoadedSuccessfully);
	PRINT("PG :  PluginStream::initPlugin EXITS \n");
	return NPERR_NO_ERROR;
}
#endif // PLATFORM(SHP)

}
#endif //SUPPORT_NP_PLUGIN