/*
 * Copyright (C) 2006, 2007 Apple Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
#ifdef SUPPORT_NP_PLUGIN

#include "config.h"
#include "PluginPackage.h"
#include "NotImplemented.h"

#include "Timer.h"
#include "DeprecatedString.h"
#include "npruntime_impl.h"
#include "PluginDebug.h"

#include "NPGluePluginDLL.h"

namespace WebCore {

PluginPackage::~PluginPackage()
{
	PRINT("PG : PluginPackage::~PluginPackage ENTERED \n");
    ASSERT(!m_isLoaded);
	if(m_pluginDLLInfo->szDescription)
	{
		free(m_pluginDLLInfo->szDescription);
	}
	if(m_pluginDLLInfo->szExtentions)
	{
		free(m_pluginDLLInfo->szExtentions);
	}
	if(m_pluginDLLInfo->szMimeTypes)
	{
		free(m_pluginDLLInfo->szMimeTypes);
	}
	if(m_pluginDLLInfo->szName)
	{
		free(m_pluginDLLInfo->szName);
	}
	fastFree(m_pluginDLLInfo);
	PRINT("PG : PluginPackage::~PluginPackage EXITS \n");
}

static Vector<String> splitString(const String& str, char delimiter, int padTo)
{
     int pos = 0;
    int newPos;
    Vector<String> result;
    DeprecatedString ds = str.deprecatedString();
    String s;
    do {

        newPos = ds.find(delimiter, pos);

        if (newPos == -1)
            s = ds.mid(pos);
        else
            s = ds.mid(pos, newPos - pos);

        if (!s.isEmpty())
            result.append(s);

        pos = newPos + 1;
    } while (newPos != -1);

    while (padTo != -1 && static_cast<int>(result.size()) < padTo)
        result.append("");

    return result;
}

PluginPackage::PluginPackage(NPGluePluginDLLInfo* pluginDLLInfo)
    : RefCounted<PluginPackage>(0)
    , m_isLoaded(false)
    , m_loadCount(0)
	, m_pluginDLLInfo(pluginDLLInfo)
{
	PRINT("PG : PluginPackage::PluginPackage ENTERED \n");
	PRINT("PG : PluginPackage::PluginPackage EXITS \n");
}

bool PluginPackage::fetchInfo()
{
	PRINT("PG : PluginPackage::fetchInfo ENTERED \n");
	Vector<String> mimeTypes = splitString(m_pluginDLLInfo->szMimeTypes, '|', -1);
	Vector<String> fileExtents = splitString(m_pluginDLLInfo->szExtentions, '|', mimeTypes.size());

	m_description = m_pluginDLLInfo->szDescription;
	// BUG SFB00000923 java script checks for the name and version of Plugin
	m_name = m_pluginDLLInfo->szName;
	// TODO : should we return if name and description is NULL
	if (m_name.isNull() || m_description.isNull())
		return false;

	for(int j=0; j < mimeTypes.size(); j++) {
		mimeTypes[j] = mimeTypes[j].lower();
		m_mimeToExtensions.add(mimeTypes[j], splitString(fileExtents[j], ',', -1));
		m_mimeToDescriptions.add(mimeTypes[j], *m_pluginDLLInfo);
	}
	PRINT("PG : PluginPackage::fetchInfo EXITS \n");
    return true;
}

bool PluginPackage::load()
{
	PRINT("PG : PluginPackage::load ENTERED \n");

    m_isLoaded = true;

    NP_GetEntryPointsProcPtr NP_GetEntryPoints = 0;
    NP_InitializeProcPtr NP_Initialize = 0;
    NPError npErr;

    NP_Initialize = m_pluginDLLInfo->fpNP_Initialize;
    NP_GetEntryPoints = m_pluginDLLInfo->fpNP_GetEntryPoints;
    m_NPP_Shutdown = m_pluginDLLInfo->fpNP_Shutdown;

    if (!NP_Initialize || !NP_GetEntryPoints || !m_NPP_Shutdown)
        goto abort;

    memset(&m_pluginFuncs, 0, sizeof(m_pluginFuncs));
    m_pluginFuncs.size = sizeof(m_pluginFuncs);

    npErr = NP_GetEntryPoints(&m_pluginFuncs);
    LOG_NPERROR(npErr);
    if (npErr != NPERR_NO_ERROR)
        goto abort;

    memset(&m_browserFuncs, 0, sizeof(m_browserFuncs));
    m_browserFuncs.size = sizeof (m_browserFuncs);
    m_browserFuncs.version = NP_VERSION_MINOR;
    m_browserFuncs.geturl = NPN_GetURL;
    m_browserFuncs.posturl = NPN_PostURL;
    m_browserFuncs.requestread = NPN_RequestRead;
    m_browserFuncs.newstream = NPN_NewStream;
    m_browserFuncs.write = NPN_Write;
    m_browserFuncs.destroystream = NPN_DestroyStream;
    m_browserFuncs.status = NPN_Status;
    m_browserFuncs.uagent = NPN_UserAgent;
    m_browserFuncs.memalloc = NPN_MemAlloc;
    m_browserFuncs.memfree = NPN_MemFree;
    m_browserFuncs.memflush = NPN_MemFlush;
    m_browserFuncs.reloadplugins = NPN_ReloadPlugins;
    m_browserFuncs.geturlnotify = NPN_GetURLNotify;
    m_browserFuncs.posturlnotify = NPN_PostURLNotify;
    m_browserFuncs.getvalue = NPN_GetValue;
    m_browserFuncs.setvalue = NPN_SetValue;
    m_browserFuncs.invalidaterect = NPN_InvalidateRect;
    m_browserFuncs.invalidateregion = NPN_InvalidateRegion;
    m_browserFuncs.forceredraw = NPN_ForceRedraw;
    m_browserFuncs.getJavaEnv = NPN_GetJavaEnv;
    m_browserFuncs.getJavaPeer = NPN_GetJavaPeer;
    m_browserFuncs.pushpopupsenabledstate = NPN_PushPopupsEnabledState;
    m_browserFuncs.poppopupsenabledstate = NPN_PopPopupsEnabledState;

    m_browserFuncs.releasevariantvalue = _NPN_ReleaseVariantValue;
    m_browserFuncs.getstringidentifier = _NPN_GetStringIdentifier;
    m_browserFuncs.getstringidentifiers = _NPN_GetStringIdentifiers;
    m_browserFuncs.getintidentifier = _NPN_GetIntIdentifier;
    m_browserFuncs.identifierisstring = _NPN_IdentifierIsString;
    m_browserFuncs.utf8fromidentifier = _NPN_UTF8FromIdentifier;
    m_browserFuncs.intfromidentifier = _NPN_IntFromIdentifier;
    m_browserFuncs.createobject = _NPN_CreateObject;
    m_browserFuncs.retainobject = _NPN_RetainObject;
    m_browserFuncs.releaseobject = _NPN_ReleaseObject;
    m_browserFuncs.invoke = _NPN_Invoke;
    m_browserFuncs.invokeDefault = _NPN_InvokeDefault;
    m_browserFuncs.evaluate = _NPN_Evaluate;
    m_browserFuncs.getproperty = _NPN_GetProperty;
    m_browserFuncs.setproperty = _NPN_SetProperty;
    m_browserFuncs.removeproperty = _NPN_RemoveProperty;
    m_browserFuncs.hasproperty = _NPN_HasProperty;
    m_browserFuncs.hasmethod = _NPN_HasMethod;
    m_browserFuncs.setexception = _NPN_SetException;
	m_browserFuncs.enumerate = _NPN_Enumerate;

    npErr = NP_Initialize(&m_browserFuncs);
    LOG_NPERROR(npErr);

    if (npErr != NPERR_NO_ERROR)
        goto abort;

    m_loadCount++;
    return true;
abort:
    unloadWithoutShutdown();
	PRINT("PG : PluginPackage::load EXITS \n");
    return false;
}

void PluginPackage::unload()
{
	PRINT("PG : PluginPackage::unload ENTERED \n");
    if (!m_isLoaded)
        return;

    if (--m_loadCount > 0)
        return;

    m_NPP_Shutdown();

    unloadWithoutShutdown();
	PRINT("PG : PluginPackage::unload EXITS \n");

}

void PluginPackage::unloadWithoutShutdown()
{
	PRINT("PG : PluginPackage::unloadWithoutShutdown ENTERED \n");

    if (!m_isLoaded)
        return;

    ASSERT(m_loadCount == 0);

    // <rdar://5530519>: Crash when closing tab with pdf file (Reader 7 only)
    // If the plugin has subclassed its parent window, as with Reader 7, we may have
    // gotten here by way of the plugin's internal window proc forwarding a message to our
    // original window proc. If we free the plugin library from here, we will jump back
    // to code we just freed when we return, so delay calling FreeLibrary at least until
    // the next message loop
    // <TODO> : do some unloading maybe destroyPlugin

    m_isLoaded = false;
	PRINT("PG : PluginPackage::unloadWithoutShutdown EXITS \n");

}

PluginPackage* PluginPackage::createPackage(NPGluePluginDLLInfo* pluginDLLInfo)
{
	PRINT("PG : PluginPackage::createPackage ENTERED \n");
    PluginPackage* package = new PluginPackage(pluginDLLInfo);

    if (!package->fetchInfo()) {
        delete package;
        return 0;
    }
	PRINT("PG : PluginPackage::createPackage EXITS \n");
    return package;
}

unsigned PluginPackage::hash() const
{
	PRINT("PG : PluginPackage::hash ENTERED \n");

    const unsigned hashCodes[] = {
    m_name.impl()->hash(),
    m_description.impl()->hash(),
    m_mimeToExtensions.size()
	};
	PRINT("PG : PluginPackage::hash EXITS \n");
    return StringImpl::computeHash(reinterpret_cast<const UChar*>(hashCodes), sizeof(hashCodes) / sizeof(UChar));
}

bool PluginPackage::equal(const PluginPackage& a, const PluginPackage& b)
{
	PRINT("PG : PluginPackage::equal ENTERED \n");
    if (a.m_name != b.m_name)
        return false;

    if (a.m_description != b.m_description)
        return false;

    if (a.m_mimeToExtensions.size() != b.m_mimeToExtensions.size())
        return false;

    MIMEToExtensionsMap::const_iterator::Keys end = a.m_mimeToExtensions.end().keys();
    for (MIMEToExtensionsMap::const_iterator::Keys it = a.m_mimeToExtensions.begin().keys(); it != end; ++it) {
        if (!b.m_mimeToExtensions.contains(*it))
            return false;
    }

	PRINT("PG : PluginPackage::equal EXITS \n");
    return true;
}

}


#endif //SUPPORT_NP_PLUGIN
