/*
* ==============================================================================
*  Name        : LoadListeners.cpp
*  Part of     : WebKit / BrowserView
*  Description : 
*/


// INCLUDE FILES
#include "ShpJasmine.h"

#include "types.h"
#include "LoadListeners.h"
#include "ContentListener.h"
#include "UrlLoader.h"
#include "UrlResponseInfo.h"
#include "UrlResponseHeaderInfo.h"
#include "ContentDispatcher.h"
#include "WebKitLoader.h"
#include "WebKitBridge.h"
#include "WebCoreBridge.h"
#include "CString.h"

#include "ResourceError.h"
#include "ResourceLoader.h"
#include "ResourceResponse.h"
#include "HTTPParsers.h"
#include "MIMETypeRegistry.h"

#include "DeprecatedString.h"
#include "DocumentLoader.h"
#include "FrameLoader.h"
#include "Frame.h"
#include "Document.h"
#include "WebControl.h"

// EXTERNAL DATA STRUCTURES

#include "WebDefs.h" // WebControl Error Code


// EXTERNAL FUNCTION PROTOTYPES

// CONSTANTS

// MACROS
#define DEFAULT_HTML_CONTENT_TYPE ("text/html")

// LOCAL CONSTANTS AND MACROS

// MODULE DATA STRUCTURES

// LOCAL FUNCTION PROTOTYPES
int __ConvertStatusToError( int status );
int __ConvertToWebControlError( int error );

// FORWARD DECLARATIONS

// ============================= LOCAL FUNCTIONS ===============================

// ============================ MEMBER FUNCTIONS ===============================

// -----------------------------------------------------------------------------
// PageLoadListener::PageLoadListener
// C++ default constructor 
// -----------------------------------------------------------------------------
//
PageLoadListener::PageLoadListener(
								   WebKitLoader* _webKitLoader, char* requestUrl )
								   : m_webKitLoader( _webKitLoader ),
#ifdef SUPPORT_CONTENT_DOWNLOADER
								   m_resourceLoader( NULL ),
								   m_contentDownloader( NULL )
#else // SUPPORT_CONTENT_DOWNLOADER
								   m_resourceLoader( NULL )
#endif // SUPPORT_CONTENT_DOWNLOADER
{
	m_totalBodyLen = 0;
	m_requestUrl = requestUrl;
}


// Destructor
PageLoadListener::~PageLoadListener()
{
#ifdef SUPPORT_CONTENT_DOWNLOADER
	delete m_contentDownloader;
#endif // SUPPORT_CONTENT_DOWNLOADER
}

// -----------------------------------------------------------------------------
// PageLoadListener::Headers
//
//
// -----------------------------------------------------------------------------
//
eErrorCode PageLoadListener::Headers(
									 int _transactionId,
									 UrlResponseInfo* _response )
{
	UrlResponseHeaderInfo* headerInfo = _response->HeaderInfo();
	String contentType( extractMIMETypeFromMediaType( headerInfo->ContentType() ) );

	// header is NULL for file:// requests
	//if( headerInfo && !headerInfo->ContentType())
	//vamsi - above line supposed to be or case ?
	if( headerInfo && !contentType.isEmpty())
	{
#ifdef SUPPORT_CONTENT_DOWNLOADER

		// Give the data to content downloader if it is not to be
		// handled by page
		// <TODO> : Check for other contentTypes for a page
		if( !MIMETypeRegistry::isSupportedNonImageMIMEType( contentType ) && !GetResourceLoader()->isSupportedImageSrc(contentType) )
		{
			// Get the type of the download from content type
			eContentDownloadType cdType = GetDownloadType( contentType );

			// Create the content downloader since its not a html data
			m_contentDownloader = new ContentDownloadListener( m_webKitLoader,
				_transactionId, 
				cdType );

			// Send the headers to the content downloader
			m_contentDownloader->Headers( _transactionId, _response );

			// This will cleanup the webcore resource loader
			WebCore::ResourceError r;
			m_resourceLoader->didFail(r);
		}
		else
#endif // SUPPORT_CONTENT_DOWNLOADER
		{
			//// Check if WebCore can render this content type
			//if( CanShowMIMEType(contentType))
			//    {
			// Check for redirection
			if( headerInfo->GetRedirectedUrl() )
			{
				// change the actual request
				KURL redirectedUrl( headerInfo->GetRedirectedUrl() );
				m_resourceLoader->documentLoader()->actualRequest().setURL( redirectedUrl );
				// update the base url
				m_resourceLoader->frameLoader()->frame()->document()->setBaseURL( redirectedUrl.string().deprecatedString() );

				// Temporarily setting all redirected requests to GET.
				m_resourceLoader->documentLoader()->request().setHTTPMethod("GET");	

			}

			String cD("Content-Disposition");
			KURL url = headerInfo->ResponseUrl() ;
			if (headerInfo->GetIsRedirected())
				url = headerInfo->GetRedirectedUrl() ;
			ResourceResponse resResponse( url,
				extractMIMETypeFromMediaType( WebCore::String( headerInfo->ContentType() ) ),
				headerInfo->ContentLength(),
				extractCharsetFromMediaType( WebCore::String( headerInfo->ContentType() ) ),
				filenameFromHTTPContentDisposition( WebCore::String( headerInfo->GetResponseHeaderField(cD) ) ) );
			String time("Last-Modified");
			//vprajan: RVCT fix
			String lastModified( headerInfo->GetResponseHeaderField(time) );

			// if we dont receive Last-Modified element then we check it wheater there is any element for "Date"
			if( lastModified.length() == 0)
			{
				time = String("Date");
				lastModified = headerInfo->GetResponseHeaderField(time);
			}

			resResponse.setLastModifiedTime( lastModified );
			resResponse.setHTTPStatusCode( headerInfo->HttpStatus());

			//         /* gnana -modified for Http refresh header */
			//if(headerInfo->GetRefresh())
			//{
			//resResponse.setHTTPHeaderField("Refresh",headerInfo->GetRefresh());
			//}
			//          /*end of modification gnana*/

			// set all headers
			// this is required to redirect it to download engine.
			HTTPHeaderMap headersMap =  headerInfo->ResponseHeaders() ;
			HTTPHeaderMap::iterator iter = headersMap.begin();
			HTTPHeaderMap::iterator endIter = headersMap.end();

			for( ; iter != endIter; ++iter )
				resResponse.setHTTPHeaderField( iter->first, iter->second );


			// Send the notification to WebCore. This
			m_resourceLoader->didReceiveResponse(resResponse);
			//           }
			//           else // WebCore cannot render this MIME type, so treat this as a plugin
			//           {
			//               String pluginHtml = CreatePluginHTMLSource( headerInfo->RequestUrl(), 
			//                                                           extractMIMETypeFromMediaType( WebCore::String( headerInfo->ContentType() ) )
			//                                                           );

			//               // Create the response for the plugin HTML source
			//	        ResourceResponse resResponse( KURL(headerInfo->ResponseUrl()),
			//                                               DEFAULT_HTML_CONTENT_TYPE,
			//                                               pluginHtml.length(),
			//                                               "",
			//                                               "");
			//               resResponse.setHTTPStatusCode( HTTP_OK );
			//               // send the response to WebCore
			//               m_resourceLoader->didReceiveResponse(resResponse);

			////            /* gnana -modified for Http refresh header */
			//		//if(headerInfo->GetRefresh())
			//		//{
			//		//resResponse.setHTTPHeaderField("Refresh",headerInfo->GetRefresh());
			//		//}
			//  //           /*end of modification gnana*/

			//	// set all headers
			//	// this is required to redirect it to download engine.
			//	HTTPHeaderMap headersMap =  headerInfo->ResponseHeaders() ;
			//	HTTPHeaderMap::iterator iter = headersMap.begin();
			//	HTTPHeaderMap::iterator endIter = headersMap.end();
			//
			//             	for( ; iter != endIter; ++iter )
			//		resResponse.setHTTPHeaderField( iter->first, iter->second );

			//               // Send the data and complete notification to WebCore
			//               m_resourceLoader->didReceiveData(pluginHtml.latin1().data(), pluginHtml.length(), 0, false);
			//               m_resourceLoader->didFinishLoading();

			//               // Finally cancel the transaction since we will download this data through a plugin
			//               m_webKitLoader->CancelTransaction( _transactionId, ERR_CANCEL );
			//          }
		}
	}

	return ERR_OK;
}

// -----------------------------------------------------------------------------
// PageLoadListener::Response
//
//
// -----------------------------------------------------------------------------
//
void PageLoadListener::Response(
								int _transactionId,
								UrlResponseInfo* _response )
{
	// fake 200
	UrlResponseHeaderInfo* header = ((UrlResponseHeaderInfo*)_response->HeaderInfo());

#ifdef SUPPORT_CONTENT_DOWNLOADER
	if( m_contentDownloader )
	{
		// Pass the data to content downloader
		m_contentDownloader->Response( _transactionId, _response );
	}
	else
#endif // SUPPORT_CONTENT_DOWNLOADER
	{
		// Send the data to the webcore resource loader    
		m_resourceLoader->didReceiveData((const char*)_response->Body(), _response->BodyLength(), 0, false);
		m_totalBodyLen += _response->BodyLength();
	}
}

// -----------------------------------------------------------------------------
// PageLoadListener::Complete
//
//
// -----------------------------------------------------------------------------
//
void PageLoadListener::Complete(
								int _transactionId,
								UrlResponseInfo* _response,
								eErrorCode _error )
{
#ifdef SUPPORT_CONTENT_DOWNLOADER
	if( m_contentDownloader )
	{
		// Notify the content downloader
		m_contentDownloader->Complete( _transactionId, _response, _error );
	}
	else
#endif // SUPPORT_CONTENT_DOWNLOADER
	{
		if( m_totalBodyLen == 0 && _error != ERR_CANCEL && _response->HeaderInfo() != NULL )
		{
			// shkim
			// In case total body length is zero, 
			// http status code should be passed to UI layer as error 
			int errorCode = __ConvertStatusToError( _response->HeaderInfo()->HttpStatus() );
			WebCore::ResourceError r( "", errorCode, m_requestUrl, "" );
			m_resourceLoader->didFail(r);
		}
		else if (ERR_CANCEL != _error)
		{	// Notify the webcore resource loader
			// only if the transaction is not canceled
			m_resourceLoader->didFinishLoading();
		}
	}
}

// -----------------------------------------------------------------------------
// PageLoadListener::HandleError
//
//
// -----------------------------------------------------------------------------
//
void PageLoadListener::HandleError(
								   int _transactionId,
								   int _error )
{
#ifdef SUPPORT_CONTENT_DOWNLOADER
	if( m_contentDownloader )
	{
		// Propogate the error to the content downloader
		m_contentDownloader->HandleError( _transactionId, _error );
	}
	else
#endif // SUPPORT_CONTENT_DOWNLOADER
	{
		int webControlError = __ConvertToWebControlError( _error );
		WebCore::ResourceError r( "", webControlError, m_requestUrl, "" );
		m_resourceLoader->didFail(r);      
	}
}

#ifdef SUPPORT_CONTENT_DOWNLOADER
// -----------------------------------------------------------------------------
// PageLoadListener::GetDownloadType
//
//
// -----------------------------------------------------------------------------
//
eContentDownloadType PageLoadListener::GetDownloadType(WebCore::String contentType)
{
	// Default download type is CD_DIRECT_DOWNLOAD
	eContentDownloadType retCDType = CD_DIRECT_DOWNLOAD;

	// Check if the content type is jad
	if( MIMETypeRegistry::isSupportMidpDownloadMIMEType( contentType ) )
	{
		retCDType = CD_MIDP_DOWNLOAD;
	}
	// Is the content type OMA?
	else if( MIMETypeRegistry::isSupportOmaDownloadMIMEType( contentType ) )
	{
		retCDType = CD_OMA_DOWNLOAD;
	}

	return retCDType;
}
#endif // SUPPORT_CONTENT_DOWNLOADER

// -----------------------------------------------------------------------------
// PageLoadListener::CanShowMIMEType
//
//
// -----------------------------------------------------------------------------
//
bool PageLoadListener::CanShowMIMEType(const String& _mimeType) const
{
	if (MIMETypeRegistry::isSupportedImageMIMEType(_mimeType))
		return true;

	if (MIMETypeRegistry::isSupportedNonImageMIMEType(_mimeType))
		return true;

	return false;
}


// -----------------------------------------------------------------------------
// PageLoadListener::CreatePluginHTMLSource
//
//
// -----------------------------------------------------------------------------
//
WebCore::String PageLoadListener::CreatePluginHTMLSource( 
	const WebCore::String& _requestUrl,
	const WebCore::String& _MIMEType )
{
	WebCore::String pluginHTML( "<html><body><embed width=\"100%\" height=\"100%\" name=\"plugin\" src=\"" );

	pluginHTML.append( _requestUrl );
	pluginHTML.append( "\" type=\"" );
	pluginHTML.append( _MIMEType );
	pluginHTML.append("\"/> </body></html>");

	return pluginHTML.characters();;
}

// -----------------------------------------------------------------------------
// ResourceLoadListener::ResourceLoadListener
// C++ default constructor can NOT contain any code, that
// might leave.
// -----------------------------------------------------------------------------
//
ResourceLoadListener::ResourceLoadListener(
	WebKitLoader* _webKitLoader )
	: m_webKitLoader( _webKitLoader ),
	m_totalLength( 0 ),
	m_resourceLoader( NULL )
{
}


// Destructor
ResourceLoadListener::~ResourceLoadListener()
{
}

// -----------------------------------------------------------------------------
// ResourceLoadListener::Headers
//
//
// -----------------------------------------------------------------------------
//
eErrorCode ResourceLoadListener::Headers(
	int _transactionId,
	UrlResponseInfo* _response )
{
	UrlResponseHeaderInfo* headerInfo = _response->HeaderInfo();

	// header is NULL for file:// requests
	if( headerInfo )
	{
		String cD("Content-Disposition");
		KURL url = headerInfo->ResponseUrl() ;
		if (headerInfo->GetIsRedirected())
			url = headerInfo->GetRedirectedUrl() ;
		ResourceResponse resResponse( url,
			extractMIMETypeFromMediaType( WebCore::String( headerInfo->ContentType() ) ),
			headerInfo->ContentLength(),
			extractCharsetFromMediaType( WebCore::String( headerInfo->ContentType() ) ),
			filenameFromHTTPContentDisposition( WebCore::String( headerInfo->GetResponseHeaderField(cD) ) ) );
		resResponse.setHTTPStatusCode( headerInfo->HttpStatus() );

		//Rashmi:Added this as part of fix for SFB00002916  
		String time("Last-Modified");
		//vprajan: RVCT fix
		String lastModified( headerInfo->GetResponseHeaderField(time) );

		// if we dont receive Last-Modified element then we check it wheater there is any element for "Date"
		if( lastModified.length() == 0)
		{
			time = String("Date");
			lastModified = headerInfo->GetResponseHeaderField(time);
		}

		resResponse.setLastModifiedTime( lastModified );
		resResponse.setHTTPStatusCode( headerInfo->HttpStatus());

		//         /* gnana -modified for Http refresh header */
		//if(headerInfo->GetRefresh())
		//{
		//resResponse.setHTTPHeaderField("Refresh",headerInfo->GetRefresh());
		//}
		//          /*end of modification gnana*/

		// set all headers
		// this is required to redirect it to download engine.
		HTTPHeaderMap headersMap =  headerInfo->ResponseHeaders() ;
		HTTPHeaderMap::iterator iter = headersMap.begin();
		HTTPHeaderMap::iterator endIter = headersMap.end();

		for( ; iter != endIter; ++iter )
			resResponse.setHTTPHeaderField( iter->first, iter->second );

		// Send the notification to WebCore. This
		m_resourceLoader->didReceiveResponse(resResponse);
	}

	return ERR_OK;
}

// -----------------------------------------------------------------------------
// ResourceLoadListener::Response
//
//
// -----------------------------------------------------------------------------
//
void ResourceLoadListener::Response(
									int _transactionId,
									UrlResponseInfo* _response )
{

	if(m_resourceLoader->shouldBufferData())
	{
        if(0 < _response->BodyLength())
        {
		    // collate the data
		    char* newData = (char*) MemoryManager::Alloc(_response->BodyLength() * sizeof(char));
		    AcMemcpy(newData, _response->Body(), _response->BodyLength() * sizeof(char));

		    m_data.append(newData);
		    m_dataLength.append(_response->BodyLength());
		    m_totalLength += _response->BodyLength();
        }
	}
	else  
	{
		// Send the data to webcore resource loader
		m_resourceLoader->didReceiveData((const char*)_response->Body(), _response->BodyLength(), 0, false);
	}
}

// -----------------------------------------------------------------------------
// ResourceLoadListener::Complete
//
//
// -----------------------------------------------------------------------------
//
void ResourceLoadListener::Complete(
									int _transactionId,
									UrlResponseInfo* _response,
									eErrorCode _error )
{
	// Notify webcore resource loader about transaction complete
	// only if the transaction is not canceled
	if( ERR_CANCEL != _error )
	{
        bool allocFailed = false;

		if(m_resourceLoader->shouldBufferData())
		{
			char* body = NULL;
			if( MemoryManager::PreCheck( m_totalLength * sizeof(char), true ) )
            {
                if( 0 < m_totalLength )
                {
			        body = (char*) MemoryManager::Alloc( m_totalLength * sizeof(char) );

			        for(unsigned i = 0, offset = 0; i < m_data.size(); i++)
			        {
				        int currentLength = m_dataLength[i];
				        char* currentData = m_data[i];
				        AcMemcpy((body + offset), m_data[i], (currentLength * sizeof(char)));
				        offset += currentLength;
				        MemoryManager::Free(currentData);
			        }
			        m_data.clear();
			        m_dataLength.clear();
                }
			    m_resourceLoader->didReceiveData(body, m_totalLength, m_totalLength, true);
            }
            else
            {
                allocFailed = true;
                HandleError( _transactionId, ERR_OUT_OF_MEMORY );
            }
		}

        if( false == allocFailed )
        {
		    m_resourceLoader->didFinishLoading();
        }
	}
}

// -----------------------------------------------------------------------------
// ResourceLoadListener::HandleError
//
//
// -----------------------------------------------------------------------------
//
void ResourceLoadListener::HandleError(
									   int /*aTransactionId*/,
									   int _error )
{
	for(unsigned i = 0; i < m_data.size(); i++)
	{
		MemoryManager::Free(m_data[i]);
	}
	m_data.clear();
	m_dataLength.clear();

	WebCore::ResourceError r;
	m_resourceLoader->didFail(r);
}


int __ConvertStatusToError( int status )
{
	int error = WebControl::ERROR_NOCONTENT_DEFAULT;

	switch( status )
	{
	case 400:
		error = WebControl::ERROR_NOCONTENT_HTTP400;
		break;

	case 401:
		error = WebControl::ERROR_NOCONTENT_HTTP401;
		break;

	case 402:
		error = WebControl::ERROR_NOCONTENT_HTTP402;
		break;

	case 403:
		error = WebControl::ERROR_NOCONTENT_HTTP403;
		break;

	case 404:
		error = WebControl::ERROR_NOCONTENT_HTTP404;
		break;

	case 405:
		error = WebControl::ERROR_NOCONTENT_HTTP405;
		break;

	case 406:
		error = WebControl::ERROR_NOCONTENT_HTTP406;
		break;

	case 407:
		error = WebControl::ERROR_NOCONTENT_HTTP407;
		break;

	case 408:
		error = WebControl::ERROR_NOCONTENT_HTTP408;
		break;

	case 409:
		error = WebControl::ERROR_NOCONTENT_HTTP409;
		break;

	case 410:
		error = WebControl::ERROR_NOCONTENT_HTTP410;
		break;

	case 411:
		error = WebControl::ERROR_NOCONTENT_HTTP411;
		break;

	case 412:
		error = WebControl::ERROR_NOCONTENT_HTTP412;
		break;

	case 413:
		error = WebControl::ERROR_NOCONTENT_HTTP413;
		break;

	case 414:
		error = WebControl::ERROR_NOCONTENT_HTTP414;
		break;

	case 415:
		error = WebControl::ERROR_NOCONTENT_HTTP415;
		break;

	case 500:
		error = WebControl::ERROR_NOCONTENT_HTTP500;
		break;

	case 501:
		error = WebControl::ERROR_NOCONTENT_HTTP501;
		break;

	case 502:
		error = WebControl::ERROR_NOCONTENT_HTTP502;
		break;

	case 503:
		error = WebControl::ERROR_NOCONTENT_HTTP503;
		break;

	case 504:
		error = WebControl::ERROR_NOCONTENT_HTTP504;
		break;

	case 505:
		error = WebControl::ERROR_NOCONTENT_HTTP505;
		break;

	default:
		error = WebControl::ERROR_NOCONTENT_DEFAULT;
		break;
	}

	return error;
}

//gnana
int __ConvertToWebControlError( int error )
{
	// [todo]
	// need to covert to webcontrol error
	// refer to error codes of WebControl in WebDefs.h
	int errorType = WebControl::ERROR_UNKNOWN;


	switch(error)
	{
	case  SFB_PROTO_ERROR_NETWORK_FAILURE:
		errorType = WebControl::ERROR_PROTO_NETWORK_FAILURE;
		break;
	case  SFB_PROTO_ERROR_WRONG_APN:
		errorType = WebControl::ERROR_PROTO_INVALID_APN;
		break;
	case  SFB_PROTO_ERROR_AUTH_FAIL:
		errorType = WebControl::ERROR_PROTO_AUTH_FAIL;
		break;
	case  SFB_PROTO_ERROR_INVALID_PARAM:
		errorType = WebControl::ERROR_PROTO_INVALID_PARAM;
		break;
	case  SFB_PROTO_ERROR_TIMEOUT:
		errorType = WebControl::ERROR_PROTO_TIME_OUT;
		break;		
	case  SFB_PROTO_ERROR_FDN:
		errorType = WebControl::ERROR_PROTO_FDN_NOT_ALLOWED;
		break;
	case SFB_PROTO_ERROR_NOT_SUBSCRIBED:
		errorType = WebControl::ERROR_PROTO_NOT_SUBSCRIBED;
		break;
	case SFB_PROTO_ERROR_NO_SERVICE:
		errorType = WebControl::ERROR_PROTO_NO_SERVICE;
		break;
	case SFB_PROTO_ERROR_MAX_CONNECTIONS:
		errorType = WebControl::ERROR_PROTO_MAX_CONNECTED;
		break;
	case SFB_PROTO_ERROR_NOT_SUPPORT_3GCSD:
		errorType = WebControl::ERROR_PROTO_NOT_SUPPORT_3GCSD;
		break;
	case SFB_PROTO_ERROR_MODEM_IN_USE:
		errorType = WebControl::ERROR_PROTO_NETWORK_USED;
		break;
	case SFB_PROTO_ERROR_UNKNOWN:
		errorType = WebControl::ERROR_PROTO_UNKNOWN;
		break;
	case SFB_HTTPERR_INVALID_PARAM:
		errorType = WebControl::ERROR_HTTPERR_INVALID_PARAM;
		break;
	case SFB_HTTPERR_UNKNOWN:
		errorType = WebControl::ERROR_HTTPERR_UNKNOWN;
		break;			
	case SFB_HTTPERR_INVALID_PROXY:
		errorType = WebControl::ERROR_HTTPERR_INVALID_PROXY;
		break;
	case SFB_HTTPERR_OUT_OF_MEMORY:
		errorType = WebControl::ERROR_HTTPERR_OUT_OF_MEMORY;
		break;		
	case SFB_HTTPERR_NO_RESPONSE:
		errorType = WebControl::ERROR_HTTPERR_NO_RESPONSE;
		break;			
	case SFB_HTTPERR_INIT:		
		errorType = WebControl::ERROR_HTTPERR_INIT;
		break;			
	case SFB_HTTPERR_NETDOWN:		
		errorType = WebControl::ERROR_HTTPERR_NETDOWN;
		break;			
	case SFB_HTTPERR_TIMEOUT:		
		errorType = WebControl::ERROR_HTTPERR_TIMEOUT;
		break;		
	case SFB_HTTPERR_HOST_UNREACH:
		errorType = WebControl::ERROR_HTTPERR_HOST_UNREACH;
		break;
	case SFB_HTTPERR_CONN_RESET:		
		errorType = WebControl::ERROR_HTTPERR_CONN_RESET;
		break;
	case SFB_HTTPERR_INTERNAL:		
		errorType = WebControl::ERROR_HTTPERR_INTERNAL;
		break;		
	case SFB_HTTPERR_CHUNKEDTR:	
		errorType = WebControl::ERROR_HTTPERR_CHUNKEDTR;
		break;			
	case SFB_HTTPERR_USER:	
		errorType = WebControl::ERROR_HTTPERR_USER;
		break;			
	case SFB_HTTPERR_TOO_BIG:
		errorType = WebControl::ERROR_HTTPERR_TOO_BIG;
		break;
	case SFB_HTTPERR_NOT_SUPPORTED_SVC:
		errorType = WebControl::ERROR_HTTPERR_NOT_SUPPORTED_SVC;
		break;
	case SFB_HTTPERR_NO_CONTENT:		
		errorType = WebControl::ERROR_HTTPERR_NO_CONTENT;
		break;
	case SFB_HTTPERR_WRITEFAIL:	
		errorType = WebControl::ERROR_HTTPERR_WRITEFAIL;
		break;			
	case SFB_HTTPERR_AUTHFAIL:
		errorType = WebControl::ERROR_HTTPERR_AUTHFAIL;
		break;
	case SFB_HTTPERR_DNSFAIL:		
		errorType = WebControl::ERROR_HTTPERR_DNSFAIL;
		break;		
	case SFB_HTTPERR_UAGENT_NOT_ALLOWED:
		errorType = WebControl::ERROR_HTTPERR_UAGENT_NOT_ALLOWED;
		break;
	case SFB_AUTH_INVALID_FORMAT:
		errorType = WebControl::ERROR_AUTH_INVALID_FORMAT;
		break;
		//rohit
	case SFB_BROWSER_ERROR_BADURL:
		errorType = WebControl::ERROR_INVALID_URL;
		break;

		//Anshul Aug 20
	case SFB_NETWORKERR_NO_SERVICE:
		errorType = WebControl::ERROR_PROTO_NO_SERVICE;
		break;

	case SFB_NETWORKERR_WRONG_APN:
		errorType = WebControl::ERROR_PROTO_INVALID_APN;
		break;

	case SFB_NETWORKERR_TIMEOUT:
		errorType = WebControl::ERROR_PROTO_TIME_OUT;
		break;

	case SFB_NETWORKERR_AUTH_FAIL:
		errorType = WebControl::ERROR_PROTO_AUTH_FAIL;
		break;

	case SFB_NETWORKERR_INVALID_PARAM:
		errorType = WebControl::ERROR_PROTO_INVALID_PARAM;
		break;

	case SFB_NETWORKERR_FDN:
		errorType = WebControl::ERROR_PROTO_FDN_NOT_ALLOWED;
		break;

	case SFB_NETWORKERR_UNKNOWN:
		errorType = WebControl::ERROR_PROTO_UNKNOWN;
		break;

	case SFB_MEMORY_FAIL:
		errorType = WebControl::ERROR_OUTOF_MEMORY;
		break;

	case SFB_STATUS_OUT_OF_MEMORY:
		errorType = WebControl::ERROR_OUTOF_MEMORY;
		break;

	case SFB_STATUS_INVALID_PARAM:
		errorType = WebControl::ERROR_PROTO_INVALID_PARAM;
		break;

	case SFB_STATUS_PROTO_NETWORK_UNCONNECTED:
		errorType = WebControl::ERROR_PROTO_NETWORK_FAILURE;
		break;

	case SFB_STATUS_PROTO_NETWORK_BUSY:
		errorType = WebControl::ERROR_PROTO_NETWORK_FAILURE;
		break;

	case  SFB_STATUS_PROTO_ERROR_WSP_CONN_FAILED:
		errorType = WebControl::ERROR_PROTO_NETWORK_FAILURE;
		break;
		//Anshul changes end
		//Bug 2924
	case  SFB_PROTO_NO_RESOURCE:
		errorType = WebControl::ERROR_PROTO_NO_RESOURCE;
		break;
	case  SFB_BROWSER_FILE_ERROR:
		errorType = WebControl::ERROR_FILE_ERROR;
		break;
	case  SFB_BROWSER_ERROR_PROFILE_SETTING_ERROR:
		errorType = WebControl::ERROR_PROFILE_SETTING_ERROR;
		break;
	case  SFB_PROTO_ERROR_MODEM_IN_USE_CANCEL:
		errorType = WebControl::ERROR_PROTO_MODEM_IN_USE_CANCEL;
		break;
		
	default :
		errorType = WebControl::ERROR_UNKNOWN;
		break;

	}

	return errorType;
}

//  End of File

