#ifndef BASETRANSACTION_H
#define BASETRANSACTION_H

//  INCLUDES
#include "UrlResponseInfo.h"

#if PLATFORM(SHP)
#include "CMMDefs.h"
#endif // PLATFORM(SHP)


// CONSTANTS

// MACROS

// DATA TYPES
// #define ASYNC_CONTENT_SENDING

// FUNCTION PROTOTYPES

// FORWARD DECLARATIONS
class UrlLoader;
class UrlResponseHeaderInfo;
class BaseTransaction;

// CLASS DECLARATION

/**
*
*  @lib resLoader.lib
*/
class ITransactionObserver
    {
    public: // New functions

        /**
        * This function is called when the header information is available
        * @param _transaction Observed transaction.
        * @param _headers Header information
        * @return ErrNotSupported if it is not a native browser content
        */
        virtual eErrorCode Headers( BaseTransaction* _transaction, UrlResponseInfo* _headers ) = 0;

        /**
        * This function is called when a new partial body is available
        * @param _transaction Observed transaction.
        * @param _response Response information
        * @return void
        */
        virtual void Response( BaseTransaction* _transaction, UrlResponseInfo* _response ) = 0;

        /**
        * This function is called when the transaction is complete
        * @param _transaction Observed transaction.
        * @param _error HTTP error (404) / cancel error
        * @return
        */
        virtual void Complete( BaseTransaction* _transaction, UrlResponseInfo* _response, eErrorCode _error ) = 0;

        /**
        * This function is called when an error happens to the transaction.
        * @param _transaction Observed transaction.
        * @param _error HTTP error
        * @return
        */
        virtual bool HandleError( BaseTransaction* _transaction, int _error ) = 0;

        /**
        * Get the url loader so that the transaction can have the session.
        * @param
        * @return Url loader object
        */
        virtual UrlLoader* GetUrlLoader() = 0;
    };

/**
*  This class is the base class for all type of transactions.
*  It supports incremental data loading.
*  @lib resLoader.lib
*/
class BaseTransaction
    CMM_MODIFIED
    {
    public:  // Constructors
        
    	/**
        * Construct.
        * @param _urlRequestInfo Url request info created by
        * resource loader client
        * @param _transactionObserver Transaction observer -It is a url loader mostly.
        * @param _transactionPriority Transaction priority (active object priority)
        * @return
        */
        BaseTransaction( UrlRequestInfo* _urlRequestInfo,
            ITransactionObserver* _transactionObserver);

    protected: // destructor

        /**
        * Destructor.
        * We are making the destructor protected since this is ref counted
        */
        virtual ~BaseTransaction();

    public: // new functions

        /**
        * lock
        */
        void Ref() { ++m_refCount; }

        /**
        * unlock
        */
        void Deref() { if( --m_refCount <= 0 ) delete this; }

        /**
        * Get the transaction id
        * @param
        * @return Transaction id
        */
        int TransactionId() const { return m_urlRequestInfo->TransId(); }

        /**
        * Sumbmit transaction
        * submit must fail asynchronously
        * @param
        * @return OS wide error code
        */
        virtual eErrorCode SubmitTransaction();

        virtual TransportHandle GetTransportHandle() = 0;

        virtual void SetTransportHandle(TransportHandle ) = 0;

        /**
        * Cancel this transaction.
        * @param _error This error code is propagated up to
        * the resource loader client through Complete()
        * @return void
        */
        virtual void CancelTransaction( eErrorCode _error = ERR_OK );

        /**
        * This function is called when the header info is available.
        * @return void
        */
        virtual void HandleGotResponseHeaders() = 0;

        /**
        * This functions is called when a new chunk is available.
        * @param _body
        * @oaram _len Length of the body data
        * @return void
        */
        virtual void HandleResponseBodyData(void* _body, int _len) = 0;

        /**
        * Get the partial response body
        * @param
        * @return Pointer to the partial response body
        */
        char* ResponseBody();

        /**
        * Get the response header
        * @param
        * @return Pointer to the response header
        */
        UrlResponseHeaderInfo* ResponseHeaders() { return m_responseInfo->HeaderInfo(); }

        /**
        * Destroy the chunk buffer
        * @param
        * @return void
        */
        void PurgeData();

        /**
        * Get the url request info created by the resource loader client
        * @param
        * @return Url request info
        */
        const UrlRequestInfo* GetUrlRequestInfo() const { return m_urlRequestInfo; }

        /**
        * Get the associated download id
        * @since 3.1
        * @param
        * @return -1 if not download is performed on this trans
        */
        eTransactionEvents Status() { return m_status; }

        /**
        * Set the status 
        * @since 3.1
        * @param status 
        * @return none
        */
        void SetStatus(eTransactionEvents _status) { m_status = _status; }

        void Run(eErrorCode _error = ERR_OK );

        /**
        * It is called when the transaction is complete
        * @param
        * @return void
        */
        virtual void TransactionComplete( eErrorCode aError = ERR_OK );

        /**
        * It handles transaction errors. Mostly it passes them back to the
        * observer
        * @param _error Transaction error.
        * @return void
        */
        virtual bool HandleError( int _error );

        /**
        * Create header information object
        * @param
        * @return Header information object
        */
        virtual void CreateHeaderResponse() = 0;

		/**
        * Passes the transaction status info to the caller
        * @param
        * @return transaction status
        */
		bool isTransCancelled() { return m_isTransCancelled; }

#if HTTP_POST_UPLOAD_SUPPORT
		typedef struct
		{
			int elementNo;
			int seekPosition;
		}postDataInfo;

		postDataInfo getPostDataInfo(){ return m_postDataInfo;}
		void setPostDataInfo(postDataInfo info){ m_postDataInfo = info;}

#endif

#if PLATFORM(SHP)
		/**
        * Stop the timer.
        * @return void
        */
		void StopTransTimer();
#endif

    protected:

        /**
        * Async call of the submit. SubmitTransaction calls
        * this function asynchronously
        * @param
        * @return void
        */
        virtual eErrorCode DoSubmit() = 0;

        /**
        * Send partial response to the transaction observer
        * @param
        * @return void
        */
        virtual void SendPartialResponse();

        /**
        * Add new package to the partial body buffer
        * @param _chunkPtr New chunk pointer.
        * @return void
        */
        void AddBodyData( const void* _chunkPtr, int _len );

        /**
        * Create response information object
        * @param
        * @return Response information object
        */
        virtual void SetBodyResponse();               

    protected: // Data

      // preserve the transaction status info
      eTransactionEvents            m_status;

      // url request info -created by the resloader client
      UrlRequestInfo*               m_urlRequestInfo;           // owned

      // transaction observer
      ITransactionObserver*         m_transactionObserver;

      // pointer to the partial body buffer
      char*                         m_responseBuffer;

      // length of the partial body
      int                           m_responseBufferLen;

      // header info
      UrlResponseInfo*              m_responseInfo;          // owned

      // url fragment (if it exists)
      char*                         m_fragment; // owned
      //
      int                           m_refCount;

    private:

      bool                          m_isTransCancelled;

#if HTTP_POST_UPLOAD_SUPPORT
	  postDataInfo m_postDataInfo;
#endif

  };

#endif      // CBASETRANSACTION_H

// End of File

