#ifndef CURLRESPONSEHEADERINFO_H
#define CURLRESPONSEHEADERINFO_H

//  INCLUDES
#include "BridgeTypes.h"
#include "WebCoreUrlResponseInfo.h"

#include "HTTPHeaderMap.h"
#include "PlatformString.h"
#include "CMMDefs.h"

// CONSTANTS

// MACROS

// DATA TYPES

// FUNCTION PROTOTYPES

// FORWARD DECLARATIONS
class WebCoreKeyValuePair;
// CLASS DECLARATION
using namespace WebCore;
/**
*  This class holds the response header information
*  @lib resLoader.lib
*  @since 3.1
*/
class UrlResponseHeaderInfo : public IWebCoreUrlResponseHeaderInfo CMM_MODIFIED_ADD
    {

    public:  // Constructors and destructor        
        
        /**
        * Constructor
        */
        UrlResponseHeaderInfo();
        
        /**
        * Destructor.
        */
        virtual ~UrlResponseHeaderInfo();

    public: // New functions

        /**
        * Set transaction id
        * @param _transId Transaction id
        * @return 
        */
        inline void SetTransId( int _transId ) { m_transId = _transId; }

        /**
        * Set top level info
        * @param aTopLevel True this transaction is a top level one
        * @return void
        */
        inline void SetTopLevel( bool _topLevel ) { m_topLevel = _topLevel; }

        /**
        * Set true if the request is initiated by a child frame
        * @param _childFrame child frame flag
        * @return void
        */
        inline void SetChildFrameLoad( bool _childFrame ) { m_childFrame = _childFrame; }

        /**
        * Get child frame flag
        * @return Child frame flag
        */
        inline bool ChildFrameLoad() const { return m_childFrame; }

        /**
        * Set request url. Note that it could be
        * different to that response url.
        * @param _url Request url
        * @return void
        */
        void SetRequestUrl( const char* _requestUrl );

        /**
        * Set response url. Note that it could be
        * different to that request url.
        * @param _url Response url
        * @return void
        */
        void SetResponseUrl( const char* _responseUrl );

        /**
        * Set HTTP status code ( 200, 404 )
        * @param eHttpStatus HTTP status code
        * @return void
        */
        inline void SetHttpStatus( eHttpStatus _httpStatus ) { m_httpStatus = _httpStatus; }

        /**
        * Set response method ( get, head, post )
        * @param _method Response method
        * @return void
        */
        inline void SetMethod( int _method ) { m_method = _method; }

        /**
        * Set response content type
        * @param _contentType Content type
        * @return void
        */
        void SetContentType( const char* _contentType );

        /**
        * Set response charset
        * @param _charset Charset info
        * @return void
        */
        void SetCharset( const char* _charset );

        /**
        * Set refresh header. 
        * @param _refresh refresh header
        * @return void
        */
        void SetRefresh( const char* _refresh );

        /**
        * Set expires header. 
        * @param _expires 
        * @return void
        */
        inline void SetExpires( time_t _expires ) { m_expires = _expires; }

        /**
        * Set boundary on the multipart response
        * @param _boundary Boundary string
        * @return void
        */
        void SetBoundary( const char* _boundary );

        /**
        * Set content location header on the 
        * multipart response
        * @param _contentLocation Content location header 
        * string
        * @return void
        */
        void SetContentLocation( const char* _contentLocation );
    
        /**
        * Set response content length
        * @param _contentLength Response content length
        * @return 
        */
        inline void SetContentLength( double _contentLength ) { m_contentLength = _contentLength; }

        /**
        * Set response headers in name-value pair format
        * @param _name of the header field e.g. "content type"
        * @param _value of the header field e.g. "text/html"
        * @return void
        */
        void SetResponseHeaders( const HTTPHeaderMap _fieldList )
            {
                m_responseHeaders = _fieldList;
            }

        /**
        * Set response header field in name-value pair format
        * @param _name of the header field e.g. "content type"
        * @param _value of the header field e.g. "text/html"
        * @return void
        */
        void SetResponseHeaderField( String _name, String _value );

        /**
        * Get response header field in name-value pair format
        * @param _name of the header field e.g. "content type"
        * @return _value corresponding to the _name in the map
        */
        String GetResponseHeaderField( String& _name ) const;

        /**
        * Indicate whether response is multipart
        * @param 
        * @return void
        */
        inline void SetIsMultipart( bool _isMultipart ) { m_isMultipart = _isMultipart; }

        /**
        * Indicates whether response is multipart
        * @param 
        * @return 
        */
        inline bool IsMultipart() { return m_isMultipart; }

        /**
        * Set if the data is inside the request (for examp: dataurl
        * @param _dataInside data is available right away
        * @return void
        */
        void SetDataInside( bool _dataInside ) { m_dataInside = _dataInside; }

        /**
        * Get if the data is in the url
        * @return True if data is available right away
        */
        bool DataInside( ) const { return m_dataInside; }

    
        /**
        * Set if the url is redirected
        * @param _isRedirected
        * @return void
        */
        inline void SetIsRedirected( bool _isRedirected ) { m_isRedirected = _isRedirected; }

        /**
        * Set the redirected url
        * @param _redirectedUrl
        * @return void
        */
        void SetRedirectedUrl( const char* _redirectedUrl );

        /**
        * Get the redirected url
        * @param void
        * @return redirected Url
        */
		inline char* GetRedirectedUrl( ) { return m_redirectedUrl; }

        /* Gnana modification for Http refresh header */
	/**
        * Get the refresh header value
        * @param void
        * @return refresh header valu
        */
		inline char* GetRefresh() { return m_refresh; }
         /* end of modification Gnana*/

		/**
        * Set to true if the connection is through a proxy
        * @param
        * @return void
        */
        void SetUsingProxy(bool _proxy) { m_usingProxy = _proxy; }

        /**
        * Set to true if the connection is through a proxy
        * @param
        * @return
        */
        bool UsingProxy() const { return m_usingProxy; }

    public: // from IWebCoreUrlResponseHeaderInfo

        /**
        * Get transaction id
        * @param 
        * @return Transaction id 
        */
        inline int TransId() const { return m_transId; }

        /**
        * Get top level info
        * @param 
        * @return True this transaction is a top level one 
        */
        inline bool TopLevel() const { return m_topLevel; }

        /**
        * Get request url. Note that it could be
        * different to that response url.
        * @param 
        * @return Request url
        */
        inline char* RequestUrl() const { return m_requestUrl; }

        /**
        * Get response url. Note that it could be
        * different to that request url.
        * @since 3.1
        * @param 
        * @return Response url 
        */
        inline char* ResponseUrl() const { return m_responseUrl; }

        /**
        * Get HTTP status code ( 200, 404 )
        * @param 
        * @return HTTP status code 
        */
        inline eHttpStatus HttpStatus() const { return m_httpStatus; }

        /**
        * Get response method ( get, head, post )
        * @param 
        * @return Response method 
        */
        inline unsigned int Method() const { return m_method; }

        /**
        * Get response content type
        * @param 
        * @return Content type 
        */
        inline char* ContentType() const { return m_contentTypeString; }

        /**
        * Get response charset
        * @param 
        * @return Charset info 
        */
        inline char* Charset() const { return m_charset; }

        /**
        * Get expire date
        * @param 
        * @return expire date
        */
        inline time_t Expires() const { return m_expires; }

        /**
        * Get boundary on the multipart response
        * @param 
        * @return Boundary string 
        */
        inline char* Boundary() const { return m_boundary; }

        /**
        * Get content location header on the 
        * multipart response
        * @param 
        * @return Content location header string 
        */
        inline char* ContentLocation() const { return m_contentLocation; }

        /**
        * Get certification info
        * @param 
        * @return Certificatio info 
        */
        inline double ContentLength() const { return m_contentLength; }


        /**
        * Get response header in name-value pair format
        * @param 
        * @return Response header in name-value pair format
        */
       const HTTPHeaderMap& ResponseHeaders() const;

        /**
        * Get response header in string format
        * @param 
        * @return Response header in string format
        */
        WebCore::String ResponseHeadersString() const;

        /**
        * Appends to the response header string
        * @param _heade - header string
        * @return 
        */
        void AddToResponseHeadersString(WebCore::String _header);

        inline bool GetIsRedirected() const {return m_isRedirected;}

    private: // Data
    
        // The request URL. Note that it could be different to that response url. 
        char*          m_requestUrl;                 
        // The response URL. Note that it could be different to that request url. 
        char*          m_responseUrl;                 
        // The string representation of content type
        char*          m_contentTypeString; 
        // The string representation of Content Base header
        char*          m_contentLocation; 
        // The string representation of boundary (multipart)
        char*          m_boundary;          
        // response's character set
        char*          m_charset;            
        // refresh header
        char*          m_refresh;            
        // Response status as received from server
        eHttpStatus         m_httpStatus;         
        // Method used for the final request
        unsigned int          m_method;             
        // The transaction id of the response
        int            m_transId;            
        // ETrue if the transaction is a top level one
        bool           m_topLevel;
        // ETrue if the request is initiated by a child frame
        bool                 m_childFrame;
        // The content length of the response. Used only for HEAD requests
        double         m_contentLength;      
        // The response headers
        HTTPHeaderMap m_responseHeaders;         
        //
        WebCore::String      m_httpHeaderString;
        // 
        time_t       m_expires;
        // True if the top-level content is multipart
        bool           m_isMultipart;
        // data is in the url
        bool m_dataInside;

        bool m_isRedirected;
		//
		char* m_redirectedUrl;
        // Set to true if the connection is through a proxy
        bool m_usingProxy;
    };

#endif      // URLRESPONSEHEADERINFO_H
            
// End of File 
