/*
     File:       CFNetwork/CFURLRequestPriv.h
     Copyright:  (c) 2007 by Apple Inc., all rights reserved
*/
#ifndef __CFURLREQUESTPRIV__
#define __CFURLREQUESTPRIV__

#ifndef __CFNETWORKDEFS__
#include <CFNetwork/CFNetworkDefs.h>
#endif

#ifndef __COREFOUNDATION__
#include <CoreFoundation/CoreFoundation.h>
#endif

#ifndef __CFHTTPMESSAGE__
#include <CFNetwork/CFHTTPMessage.h>
#endif

#ifndef __CFHTTPCOOKIESPRIV__
#include <CFNetwork/CFHTTPCookiesPriv.h>
#endif



#include <AvailabilityMacros.h>

#if PRAGMA_ONCE
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

#if PRAGMA_ENUM_ALWAYSINT
    #pragma enumsalwaysint on
#endif


/*
 *  CFURLRequestCachePolicy
 *  
 *  Discussion:
 *    The caching policy to be used when processing the request
 */
enum CFURLRequestCachePolicy {

  /*
   * Allow the underlying protocol (like HTTP) to choose the most
   */
  kCFURLRequestCachePolicyProtocolDefault = 0,

  /*
   * Ignore any cached contents, requiring that the content come from
   * the origin server
   */
  kCFURLRequestCachePolicyReloadIgnoringCache = 1,

  /*
   * Return the contents of the cache (if any), otherwise load from the
   * origin server
   */
  kCFURLRequestCachePolicyReturnCacheDataElseLoad = 2,

  /*
   * Return the contents of the cache (if any), otherwise, return
   * nothing
   */
  kCFURLRequestCachePolicyReturnCacheDataDontLoad = 3
};
typedef enum CFURLRequestCachePolicy CFURLRequestCachePolicy;


/*
 *  CFURLRequestRef
 *  
 *  Discussion:
 *    The type for a request for a URL
 */
typedef const struct _CFURLRequest*     CFURLRequestRef;
/*
 *  CFURLRequestGetTypeID()
 *  
 *  Discussion:
 *    Returns the type ID for the CFType CFURLRequest
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API CFTypeID 
CFURLRequestGetTypeID(void)                                   AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCreate()
 *  
 *  Discussion:
 *    Create a new request
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Parameters:
 *    
 *    alloc:
 *      The allocator to use to create the new object
 *    
 *    URL:
 *      The URL being requested; must not be NULL
 *    
 *    cachePolicy:
 *      The cache policy to use when retrieving the URL
 *    
 *    timeout:
 *      How long to attempt to retrieve the URL before timing out
 *    
 *    mainDocumentURL:
 *      The URL of the document associated with the requested URL
 *  
 *  Result:
 *    The newly created CFURLRequest object
 *  
 */
APPLE_WIN_API CFURLRequestRef 
CFURLRequestCreate(
  CFAllocatorRef            alloc,
  CFURLRef                  URL,
  CFURLRequestCachePolicy   cachePolicy,
  CFTimeInterval            timeout,
  CFURLRef                  mainDocumentURL)                  AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCreateCopy()
 *  
 */
APPLE_WIN_API CFURLRequestRef 
CFURLRequestCreateCopy(
  CFAllocatorRef    alloc,
  CFURLRequestRef   originalRequest)                          AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestGetURL()
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Result:
 *    Returns the URL for the request
 *  
 */
APPLE_WIN_API CFURLRef 
CFURLRequestGetURL(CFURLRequestRef request)                   AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestGetCachePolicy()
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Result:
 *    Returns the cache policy for the request
 *  
 */
APPLE_WIN_API CFURLRequestCachePolicy 
CFURLRequestGetCachePolicy(CFURLRequestRef request)           AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestGetTimeoutInterval()
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Result:
 *    Returns the timeout interval for the request
 *  
 */
APPLE_WIN_API CFTimeInterval 
CFURLRequestGetTimeoutInterval(CFURLRequestRef request)       AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestGetMainDocumentURL()
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Result:
 *    Returns the URL for the main document associated with the request
 *    (if any)
 *  
 */
APPLE_WIN_API CFURLRef 
CFURLRequestGetMainDocumentURL(CFURLRequestRef request)       AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestGetProxySettings()
 *  
 *  Discussion:
 *    The setting dictionary matches the SystemConfiguration proxy
 *    dictionary
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API CFDictionaryRef 
CFURLRequestGetProxySettings(CFURLRequestRef request)         AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestGetSSLProperties()
 *  
 *  Discussion:
 *    See CFNetwork/CFSocketStream.h for a full description of the
 *    expected keys and values in the returned dictionary
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API CFDictionaryRef 
CFURLRequestGetSSLProperties(CFURLRequestRef request)         AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/* Immutable HTTP extensions; you will need a basic understanding of the structure of HTTP requests to understand these functions */
/*
 *  CFURLRequestCreateHTTPRequest()
 *  
 *  Discussion:
 *    Creates a request for an HTTP URL; use this in preference to
 *    CFURLRequestCreate in order to configure HTTP-specific pieces of
 *    the request like the HTTP headers
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Parameters:
 *    
 *    alloc:
 *      The allocator for allocating the new request
 *    
 *    httpRequest:
 *      The HTTP request to be issued.  Note that headers in the
 *      request are not considered complete; when the request is
 *      issued, the protocol may choose to add additional headers or
 *      modify the existing headers if neecded to properly process the
 *      request.  The body of this request will be used unless
 *      bodyStream, below, is set; the body stream takes precedence.
 *    
 *    bodyStream:
 *      A read stream containing the body of the HTTP request.  The
 *      stream should be unopened, unscheduled, and have no client. 
 *      When the URL request is processed, the stream will be opened
 *      and read from to provide the body fo the request. Note that
 *      this stream will take precedence over any body data set on
 *      httpRequest, above.
 *    
 *    cachePolicy:
 *      The cache policy to use when retrieving the URL
 *    
 *    timeout:
 *      How long to attempt to retrieve the URL before timing out
 *    
 *    mainDocumentURL:
 *      The URL of the document associated with the requested URL
 *  
 *  Result:
 *    The newly created CFURLRequest object
 *  
 */
APPLE_WIN_API CFURLRequestRef 
CFURLRequestCreateHTTPRequest(
  CFAllocatorRef            alloc,
  CFHTTPMessageRef          httpRequest,
  CFReadStreamRef           bodyStream,
  CFURLRequestCachePolicy   cachePolicy,
  CFTimeInterval            timeout,
  CFURLRef                  mainDocumentURL)                  AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyHTTPRequestMethod()
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Result:
 *    The HTTP method for this request; this will be "GET" if no method
 *    has been configured on the request
 *  
 */
APPLE_WIN_API CFStringRef 
CFURLRequestCopyHTTPRequestMethod(CFURLRequestRef request)    AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyHTTPRequestBody()
 *  
 */
APPLE_WIN_API CFDataRef 
CFURLRequestCopyHTTPRequestBody(CFURLRequestRef request)      AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyHTTPHeaderFieldValue()
 *  
 */
APPLE_WIN_API CFStringRef 
CFURLRequestCopyHTTPHeaderFieldValue(
  CFURLRequestRef   request,
  CFStringRef       headerField)                              AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyAllHTTPHeaderFields()
 *  
 */
APPLE_WIN_API CFDictionaryRef 
CFURLRequestCopyAllHTTPHeaderFields(CFURLRequestRef request)  AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyHTTPRequest()
 *  
 *  Discussion:
 *    Intended to grab a copy of the final, to-be-transmitted HTTP
 *    request; caller cannot mutate the returned CFHTTPMessageRef and
 *    expect the CFURLRequest to have mutated
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API CFHTTPMessageRef 
CFURLRequestCopyHTTPRequest(CFURLRequestRef request)          AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyHTTPRequestBodyStream()
 *  
 */
APPLE_WIN_API CFReadStreamRef 
CFURLRequestCopyHTTPRequestBodyStream(CFURLRequestRef request) AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyAdditionalHTTPCookies()
 *  
 */
APPLE_WIN_API CFArrayRef 
CFURLRequestCopyAdditionalHTTPCookies(CFURLRequestRef request) AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCopyHTTPCookieStorage()
 *  
 */
APPLE_WIN_API CFHTTPCookieStorageRef 
CFURLRequestCopyHTTPCookieStorage(CFURLRequestRef request)    AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestGetHTTPCookieStorageAcceptPolicy()
 *  
 */
APPLE_WIN_API CFHTTPCookieStorageAcceptPolicy 
CFURLRequestGetHTTPCookieStorageAcceptPolicy(CFURLRequestRef request) AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;



/*
 *  CFURLRequestShouldHandleHTTPCookies()
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Result:
 *    Whether this request, when processed, should automatically
 *    consult the shared cookie store and automatically update that
 *    store
 *  
 */
APPLE_WIN_API Boolean 
CFURLRequestShouldHandleHTTPCookies(CFURLRequestRef request)  AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/* Mutable requests */

/*
 *  CFMutableURLRequestRef
 *  
 *  Discussion:
 *    The type for a mutable request for a URL
 */
typedef struct _CFURLRequest*           CFMutableURLRequestRef;
/*
 *  CFURLRequestCreateMutable()
 *  
 *  Summary:
 *    Variant of CFURLRequestCreate that returns a mutable object
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API CFMutableURLRequestRef 
CFURLRequestCreateMutable(
  CFAllocatorRef            alloc,
  CFURLRef                  URL,
  CFURLRequestCachePolicy   cachePolicy,
  CFTimeInterval            timeout,
  CFURLRef                  mainDocumentURL)                  AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestCreateMutableCopy()
 *  
 *  Summary:
 *    Creates a mutable copy of an existing request
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API CFMutableURLRequestRef 
CFURLRequestCreateMutableCopy(
  CFAllocatorRef    alloc,
  CFURLRequestRef   origRequest)                              AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/* Simple accessors to configure a mutable request*/
/*
 *  CFURLRequestSetURL()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetURL(
  CFMutableURLRequestRef   mutableRequest,
  CFURLRef                 url)                               AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetCachePolicy()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetCachePolicy(
  CFMutableURLRequestRef    mutableRequest,
  CFURLRequestCachePolicy   cachePolicy)                      AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetTimeoutInterval()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetTimeoutInterval(
  CFMutableURLRequestRef   mutableRequest,
  CFTimeInterval           timeoutInterval)                   AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetMainDocumentURL()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetMainDocumentURL(
  CFMutableURLRequestRef   mutableRequest,
  CFURLRef                 mainDocumentURL)                   AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetProxySettings()
 *  
 *  Discussion:
 *    Configures any custom proxy settings
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Parameters:
 *    
 *    mutableRequest:
 *      The request to be changed
 *    
 *    proxySettings:
 *      A dictionary setting the desired SSL properties; the format
 *      should match the dictionary used by SystemConfiguration. A NULL
 *      dictionary means use the default proxy settings for the system;
 *      non-NULL but empty means no proxies at all.
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetProxySettings(
  CFMutableURLRequestRef   mutableRequest,
  CFDictionaryRef          proxySettings)                     AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetSSLProperties()
 *  
 *  Discussion:
 *    Configures how any SSL communication with the server will be
 *    performed, including how the server's certificate is verified.
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Parameters:
 *    
 *    mutableRequest:
 *      The request to be changed
 *    
 *    sslProperties:
 *      A dictionary setting the desired SSL properties; the format
 *      should match the dictionary used by CFSocketStream. See
 *      CFNetwork/CFSocketStream.h for a full description of the
 *      appropriate keys and values
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetSSLProperties(
  CFMutableURLRequestRef   mutableRequest,
  CFDictionaryRef          sslProperties)                     AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/* Mutable HTTP requests */
/*
 *  CFURLRequestCreateMutableHTTPRequest()
 *  
 *  Summary:
 *    Variant of CFURLRequestCreateHTTPRequest that returns a mutable
 *    object.
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API CFMutableURLRequestRef 
CFURLRequestCreateMutableHTTPRequest(
  CFAllocatorRef            alloc,
  CFHTTPMessageRef          httpRequest,
  CFReadStreamRef           bodyStream,
  CFURLRequestCachePolicy   cachePolicy,
  CFTimeInterval            timeout,
  CFURLRef                  mainDocumentURL)                  AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/* Simple accessors for setting the HTTP properties of a request */
/*
 *  CFURLRequestSetHTTPRequestMethod()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetHTTPRequestMethod(
  CFMutableURLRequestRef   mutableHTTPRequest,
  CFStringRef              httpMethod)                        AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetHTTPHeaderFieldValue()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetHTTPHeaderFieldValue(
  CFMutableURLRequestRef   mutableHTTPRequest,
  CFStringRef              httpHeaderField,
  CFStringRef              httpHeaderFieldValue)              AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetMultipleHTTPHeaderFields()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetMultipleHTTPHeaderFields(
  CFMutableURLRequestRef   mutableHTTPRequest,
  CFDictionaryRef          headerFields)                      AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestAppendHTTPHeaderFieldValue()
 *  
 *  Discussion:
 *    Extends the value of an HTTP header field.  This is useful if you
 *    want to simply add an additional (comma-separated) value to an
 *    HTTP header.
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API void 
CFURLRequestAppendHTTPHeaderFieldValue(
  CFMutableURLRequestRef   mutableHTTPRequest,
  CFStringRef              httpHeaderField,
  CFStringRef              additionalHttpHeaderFieldValue)    AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/* Note that setting the body data is mutually exclusive with setting the body stream; setting either one will clear the other */
/*
 *  CFURLRequestSetHTTPRequestBody()
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetHTTPRequestBody(
  CFMutableURLRequestRef   mutableHTTPRequest,
  CFDataRef                httpBody)                          AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetHTTPRequestBodyStream()
 *  
 *  Discussion:
 *    Sets the stream from which the body of the HTTP should come.  The
 *    stream should be unopened, unscheduled, and without a client. 
 *    When the HTTP server is contacted, the loader will set the client
 *    and schedule the stream for its own convenience.  The bytes of
 *    the stream will then be read and sent to the HTTP server as the
 *    body of the request.
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetHTTPRequestBodyStream(
  CFMutableURLRequestRef   mutableHTTPRequest,
  CFReadStreamRef          bodyStream)                        AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetShouldHandleHTTPCookies()
 *  
 *  Discussion:
 *    Sets whether the request should retrieve cookies from the shared
 *    cookie store before transmission and update the shared cookie
 *    store from the server's response
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetShouldHandleHTTPCookies(
  CFMutableURLRequestRef   mutableHTTPRequest,
  Boolean                  shouldHandleCookies)               AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetAdditionalHTTPCookies()
 *  
 *  Discussion:
 *    Sets additional cookies (to be used in addition to any cookies in
 *    the shared cookie store) on the request
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetAdditionalHTTPCookies(
  CFMutableURLRequestRef   mutableHTTPRequest,
  CFArrayRef               additionalCookies)                 AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetHTTPCookieStorage()
 *  
 *  Discussion:
 *    Sets cookie storage (to be used instead of default storage) on
 *    the request
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetHTTPCookieStorage(
  CFMutableURLRequestRef   req,
  CFHTTPCookieStorageRef   cookieStorage)                     AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFURLRequestSetHTTPCookieStorageAcceptPolicy()
 *  
 *  Discussion:
 *    Sets cookie acceptance policy on the request
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 */
APPLE_WIN_API void 
CFURLRequestSetHTTPCookieStorageAcceptPolicy(
  CFMutableURLRequestRef            req,
  CFHTTPCookieStorageAcceptPolicy   cookieStorageAcceptPolicy) AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


/*
 *  CFReadStreamCreateWithFormArray()
 *  
 *  Discussion:
 *    Constructs and returns a read stream for the given form data
 *  
 *  Mac OS X threading:
 *    Thread safe
 *  
 *  Parameters:
 *    
 *    alloc:
 *      The allocator to use when creating the stream
 *    
 *    formArray:
 *      A CFArray whose elements are all either CFDatas or CFStrings.
 *      The CFStrings are interpreted as paths for files to be loaded;
 *      the CFDatas are interpreted as raw bytes. Note that if the
 *      array contains elements that are neither CFDatas nor CFStrings,
 *      or if the CFStrings are not valid paths (or do not point to
 *      valid files), this function will still return a valid read
 *      stream.  However, at some point, reading from that stream will
 *      produce an error.
 *  
 *  Result:
 *    A read stream which, when read, returns the contents of formArray
 *    in order.  For CFDatas, it returns the bytes of the data; for
 *    strings, it returns the bytes of the file at the path for the
 *    string.
 *  
 */
APPLE_WIN_API CFReadStreamRef 
CFReadStreamCreateWithFormArray(
  CFAllocatorRef   alloc,
  CFArrayRef       formArray)                                 AVAILABLE_MAC_OS_X_VERSION_10_5_AND_LATER;


#if PRAGMA_ENUM_ALWAYSINT
    #pragma enumsalwaysint reset
#endif

#ifdef __cplusplus
}
#endif

#endif /* __CFURLREQUESTPRIV__ */

