/*
 * Copyright (c) S/W Platform, Samsung Electronics, Inc.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Samsung Electronics, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Samsung Electronics. 
 */
/** 
 * @file		IChannel.h
 * @brief		This file define interface for URILoader IChannel.
 * @author	Sangheon Kim (sangheon77.kim@samsung.com)
 * @version	1.0
 * @date		2008.03.04	v1.0		Created
 */

#ifndef __IChannel_Defined__
#define __IChannel_Defined__


#include "URILoader.h"
#include "IRequest.h"
#include "IURI.h"
#include "IChannelEventListener.h"
#include "IChannelEventSender.h"
#include "IStreamListener.h"
#include "ICookieStorage.h"
#include "ICacheStorage.h"

namespace  URILoader
{



/**
 * The IChannel interface allows clients to construct "GET" requests for
 * specific protocols, and manage them in a uniform way.  Once a channel is
 * created (via newChannel), parameters for that request may
 * be set by using the channel attributes, or by QI'ing to a subclass of
 * IChannel for protocol-specific parameters.  Then, the URI can be fetched
 * by calling IChannel::open or nsIChannel::asyncOpen.
 *
 * After a request has been completed, the channel is still valid for accessing
 * protocol-specific results.  For example, QI'ing to IHttpChannel allows
 * response headers to be retrieved for the corresponding http transaction. 
 *
 */



class IChannel : public IRequest
{
public:
	/* Constants */

	static const unsigned long LOAD_DOCUMENT_URI = 1 << 16;

	static const unsigned long LOAD_RETARGETED_DOCUMENT_URI = 1 << 17;

	static const unsigned long LOAD_REPLACE = 1 << 18;

	static const unsigned long LOAD_INITIAL_DOCUMENT_URI = 1 << 19;

	static const unsigned long LOAD_TARGETED = 1 << 20;

	static const unsigned long LOAD_CALL_CONTENT_SNIFFERS = 1 << 21;
	/* Properties */
	/**
	 * The original URI used to construct the channel. This is used in the case
	 * of a redirect or URI "resolution" (e.g. resolving a resource: URI to a
	 * file: URI) so that the original pre-redirect URI can still be obtained. 
	 */
	virtual Result GetOriginalURI( IURI **ppValue ) = 0;
	virtual Result SetOriginalURI( IURI *pURI ) = 0;

	/**
	 * The URI corresponding to the channel.  Its value is immutable.
	 */
	virtual Result GetURI( IURI **ppURI ) = 0;

	/**
	 * The owner, corresponding to the entity that is responsible for this
	 * channel.  Used by the security manager to grant or deny privileges to
	 * mobile code loaded from this channel.
	 */
	virtual Result GetChannelEventListener( IChannelEventListener ** pValue ) = 0;
	virtual Result SetChannelEventListener( IChannelEventSender* aValue ) = 0;


	virtual Result GetCacheStorage( ICacheStorage **pValue ) = 0;
	virtual Result SetCacheStorage( ICacheStorage *aValue ) = 0;

	
	virtual Result GetCookieStorage( ICookieStorage **pValue ) = 0;
	virtual Result SetCookieStorage( ICookieStorage *aValue ) = 0;

	virtual Result GetContentType( String *pValue ) = 0;
	virtual Result SetContentType( String *aValue ) = 0;
	

	/**
	 * The character set of the channel's content if available and if applicable.
	 * This attribute only applies to textual data.
	 *
	 * The value of the contentCharset attribute is a mixedcase string.
	 */
	virtual Result GetContentCharset( String *pValue ) = 0;
	virtual Result SetContentCharset(  String *aValue ) = 0;

	/**
	 * The length of the data associated with the channel if available.  A value
	 * of -1 indicates that the content length is unknown.
	 */
	virtual Result GetContentLength( long *pValue ) = 0;
	virtual Result SetContentLength( long aValue ) = 0;

	/* Funtions */
	/**
	 * Open this channel Asynchronously. Data is fed to the specified stream
	 * listener as it becomes available. The stream listener's methods are
	 * called on the thread that calls asyncOpen and are not called until
	 * after asyncOpen returns. If asyncOpen returns successfully, the
	 * channel promises to call at least onStartRequest and onStopRequest.
	 */
	virtual Result AsyncOpen( IStreamListener *aListener, void* aContext) = 0;

};


} // namespace URILoader

#endif //_IChannel_Defined
