/*
 * Copyright (c) S/W Platform, Samsung Electronics, Inc.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Samsung Electronics, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Samsung Electronics. 
 */
/** 
 * @file		IHttpChannel..h
 * @brief		Define interface of URILoader IHttpChannel
 * @author	Sangheon Kim (sangheon77.kim@samsung.com)
 * @version	1.0
 * @date		2008.03.05	v1.0		Created
 */
#ifndef _IHttpChannel_Defined
#define _IHttpChannel_Defined

#include "URILoader.h"
#include "ILoadGroup.h"
#include "IChannel.h"

namespace URILoader
{


/**
 * IHttpChannel
 *
 * This interface allows for the modification of HTTP request parameters and
 * the inspection of the resulting HTTP response status and headers when they
 * become available.
 */
class IHttpChannel : public IChannel
{
public:

	
	/***************************************************************************
	*
	* Properties
	*
	****************************************************************************/

	/**************************************************************************
	 * REQUEST CONFIGURATION - Modifying request parameters after asyncOpen has been called is an error. */

	/**
	 * Set/get the HTTP request method (default is "GET").  Setter is case
	 * insensitive; getter returns an uppercase string.
	 */
	virtual Result GetRequestMethod( String *pValue ) = 0;
	virtual Result SetRequestMethod( String aValue ) = 0;
	

	/**
	 * Get/set the HTTP referrer URI.  This is the address (URI) of the
	 * resource from which this channel's URI was obtained (see RFC2616 section
	 * 14.36). */
	 
	virtual Result GetReferrer( IURI **ppValue )  = 0;
	virtual Result SetReferrer( IURI *aValue ) = 0;

	/** RESPONSE INFO
	 * Accessing response info before the onStartRequest event is an error.
	 */

	/**
	 * Get the HTTP response code (e.g., 200).
	 */
	virtual Result GetResponseStatus( unsigned long *pValue ) = 0;

	/**
	 * Get the HTTP response status text (e.g., "OK").
	 */

	/**
	 * Returns true if the HTTP response code indicates success.  The value of
	 * IRequest::status will be OK even when processing a 404 response
	 * because a 404 response may include a message body that (in some cases)
	 * should be shown to the user.
	 */
	virtual Result GetRequestSucceeded( BOOL *pValue ) = 0; 

	/***************************************************************************
	*
	* Funtions
	*
	****************************************************************************/

	/**
	 * Get the value of a particular request header.
	 * @param aHeader: The case-insensitive name of the request header to query (e.g.,
	 *        "Cache-Control").

	 */
	// Temporary fix - sumdra
	virtual Result GetRequestHeader( String *aHeader, String *pValue ) = 0;
	  //virtual Result GetRequestHeader( void **pValue ) = 0;

	/**
	 * Set the value of a particular request header.
	 *
	 */
	// Temporary fix - sumdra
	virtual Result SetRequestHeader( String *aHeader,
	                      String *aValue,
	                      BOOL aMerge) = 0;
	 //virtual Result SetRequestHeader( void * pHeader) ;

	/**
	 * Call this method to visit all request headers.  Calling setRequestHeader
	 * while visiting request headers has undefined behavior.  Don't do it!
	 */
	//virtual  Result VisitRequestHeaders( IHttpHeaderVisitor* aVisitor) = 0;



	/**
	 * Get the value of a particular response header.
	 *
	 * @param aHeader
	 *        The case-insensitive name of the response header to query (e.g.,
	 *        "Set-Cookie").
	 * @return the value of the response header.
	 * @returns NS_ERROR_NOT_AVAILABLE if called before the response
	 *         has been received (before onStartRequest) or if the header is
	 *         not set in the response.
	 */
	virtual Result GetResponseHeader(  String *aHeader, String *pValue ) = 0;

	/**
	 * Set the value of a particular response header.
	 *
	 * This method allows, for example, the HTML content sink to inform the HTTP
	 * channel about HTTP-EQUIV headers found in HTML <META> tags.
	 */
	virtual Result SetResponseHeader( String *aHeader,
	                       String aValue,
	                       BOOL aMerge) =0;

	/**
	 * Call this method to visit all response headers.  Calling
	 * setResponseHeader while visiting response headers has undefined
	 * behavior.  Don't do it!
	 */
	virtual Result GetResponseHeaderMap( HeaderMap* aVisitor) = 0;

	/**
	*
	*/
	virtual Result GetResponseHeaderString( char **pszBuf  ) = 0;

	/**
	 * Returns true if the server sent a "Cache-Control: no-store" response
	 * header.
	 */
	virtual Result IsNoStoreResponse( BOOL *pValue ) = 0;

	/**
	 * Returns true if the server sent the equivalent of a "Cache-control:
	 * no-cache" response header.  Equivalent response headers include:
	 * "Pragma: no-cache", "Expires: 0", and "Expires" with a date value
	 * in the past relative to the value of the "Date" header.
	 */
	virtual Result IsNoCacheResponse( BOOL *pValue ) = 0;

	//virtual Result GetResponseType(int *pValue) = 0;


	virtual Result SetUploadStream(const char* pData, int len, String contentType ) = 0;

	// Get and Set Caller Client ID 
	virtual Result SetCallerClientID(int clientID ) = 0;

	virtual Result GetCallerClientID(int* clientID) = 0;


}; // IHttpChannel


} // namespace 

#endif //_IHttpChannel_Defined
