/*
 * Copyright (c) S/W Platform, Samsung Electronics, Inc.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Samsung Electronics, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Samsung Electronics. 
 */
/** 
 * @file		WebPageInfo.cpp
 * @brief		implement DOM level 2 Core
 * @author	Sangheon Kim (sangheon77.kim@samsung.com)
 * @version	1.0
 * @date		2008.03.18	v1.0		Created
 */

#include "WebDOMCore.h"
#include "WebDOMHtml.h"

#include "Node.h"
#include "Element.h"
#include "Document.h"
#include "String.h"
#include "CString.h"

namespace WebControl {

IDOMNode* DOMNode::New( WebCore::Node* pNode )
{
    if ( pNode == NULL )
        return NULL;

	IDOMNode *pDOMNode;
    WebCore::Node::NodeType nodeType = pNode->nodeType();

    switch (nodeType) {

        case WebCore::Node::ELEMENT_NODE: 
        {
            pDOMNode = DOMElement::New(static_cast<WebCore::Element*>(pNode) );
        }
        break;

        case WebCore::Node::DOCUMENT_NODE:
        {
            pDOMNode = DOMDocument::New( pNode->document() );
        }
        break;

        default:
        {
            pDOMNode = new DOMNode(pNode); 
        }
        break;
    }

	return pDOMNode;
}


DOMNode::DOMNode( WebCore::Node* pNode )
: m_pNode( pNode )
{
}


/*
 * @desc gets valud of node
 * @param out_pszBuffer
 * @param in_nBufferLen
 * @return HRESULT TRUE/FALSE
 */
HRESULT DOMNode::getValue( char* out_pszBuffer, int in_nBufferLen )
{
	WebCore::String nodeValue = m_pNode->nodeValue();

	if( in_nBufferLen - 1 < nodeValue.length() )
		return HRESULT_NO_MEMORY;
	else
	{
		AcStrcpy( out_pszBuffer , nodeValue.latin1().data() );
		return HRESULT_OK;
	}
}

/*
 * @desc gets name of node
 * @param out_pszBuffer
 * @param in_nBufferLen
 * @return HRESULT TRUE/FALSE
 */
HRESULT DOMNode::getName( char* out_pszBuffer, int in_nBufferLen )
{
	WebCore::String nodeName = m_pNode->nodeName();

	if( in_nBufferLen - 1 < nodeName.length() )
		return HRESULT_NO_MEMORY;
	else
	{
		AcStrcpy( out_pszBuffer , nodeName.latin1().data() );
		return HRESULT_OK;
	}
} 

/*
 * @desc gets the type of node
 * @param
 * @return DomNodeType
 */
HRESULT DOMNode::getNodetype( DomNodeType *out_type )
{
	DomNodeType nodeType = UNKNOWN_NODE;
	WebCore::Node::NodeType type = m_pNode->nodeType();

	switch( type )
	{
	case WebCore::Node::ELEMENT_NODE:
		nodeType = ELEMENT_NODE;
		break;

	case WebCore::Node::ATTRIBUTE_NODE:
		nodeType = ATTRIBUTE_NODE;
		break;

	case WebCore::Node::TEXT_NODE:
		nodeType = TEXT_NODE;
		break;

	case WebCore::Node::CDATA_SECTION_NODE:
		nodeType = CDATA_SECTION_NODE;
		break;

	case WebCore::Node::ENTITY_REFERENCE_NODE:
		nodeType = ENTITY_REFERENCE_NODE;
		break;

	case WebCore::Node::ENTITY_NODE:
		nodeType = ENTITY_NODE;
		break;

	case WebCore::Node::PROCESSING_INSTRUCTION_NODE:
		nodeType = PROCESSING_INSTRUCTION_NODE;
		break;

	case WebCore::Node::COMMENT_NODE:
		nodeType = COMMENT_NODE;
		break;

	case WebCore::Node::DOCUMENT_NODE:
		nodeType = DOCUMENT_NODE;
		break;

	case WebCore::Node::DOCUMENT_TYPE_NODE:
		nodeType = DOCUMENT_TYPE_NODE;
		break;

	case WebCore::Node::DOCUMENT_FRAGMENT_NODE:
		nodeType = DOCUMENT_FRAGMENT_NODE;
		break;

	case WebCore::Node::NOTATION_NODE:
		nodeType = NOTATION_NODE;
		break;

	default:
		*out_type = nodeType;
		return HRESULT_NOK;
	}

	*out_type = nodeType;
	return HRESULT_OK;

}

/*
 * @desc checks whether the node has attribute
 * @param
 * @return HRESULT TRUE/FALSE
 */
HRESULT DOMNode::hasAttribute()
{
	if( m_pNode->hasAttributes() )
		return HRESULT_TRUE;
	else
		return HRESULT_FALSE;
}

WebCore::Node* DOMNode::getWebCoreNode()
{
	return m_pNode;
}		



IDOMDocument* DOMDocument::New( WebCore::Document* pDoc )
{
    if (!pDoc)
        return NULL;

    IDOMDocument* pDOMDoc = 0;

    if (pDoc->isHTMLDocument()) 
	{
        pDOMDoc = new DOMHtmlDocument(pDoc);
     } 
	else 
	{
        pDOMDoc = new DOMDocument(pDoc);        
    }

	return pDOMDoc;
}

DOMDocument::DOMDocument( WebCore::Document* pDoc )
: m_DOMNode( pDoc ), m_pDoc( pDoc )
{

}

// DOMNode funtions
// Not to use multiple inheritance , use embeded DOMNode concept
HRESULT DOMDocument::getValue( char* out_pszBuffer, int in_nBufferLen )
{
	return m_DOMNode.getValue( out_pszBuffer, in_nBufferLen );
}
HRESULT DOMDocument::getName( char* out_pszBuffer, int in_nBufferLen )
{
	return m_DOMNode.getName( out_pszBuffer, in_nBufferLen );
}
HRESULT DOMDocument::getNodetype( DomNodeType *out_type )
{
	return m_DOMNode.getNodetype( out_type);
}
HRESULT DOMDocument::hasAttribute()
{
	return m_DOMNode.hasAttribute();
}


HRESULT DOMDocument::getDocType( IDOMDocumentType **out_ppDocType )
{
	SysNeverGetHere();
	return HRESULT_NOTSUPPORT;
}








IDOMElement* DOMElement::New(WebCore::Element* pElement )
{
    if (!pElement)
        return NULL;

    IDOMElement* pDOMElement = NULL;

	// [todo] in the future
    //if (e->hasTagName(formTag)) {
    //    DOMHTMLFormElement* newElement = new DOMHTMLFormElement(e);
    //    hr = newElement->QueryInterface(IID_IDOMElement, (void**)&domElement);
    //} else if (e->hasTagName(selectTag)) {
    //    DOMHTMLSelectElement* newElement = new DOMHTMLSelectElement(e);
    //    hr = newElement->QueryInterface(IID_IDOMElement, (void**)&domElement);
    //} else if (e->hasTagName(optionTag)) {
    //    DOMHTMLOptionElement* newElement = new DOMHTMLOptionElement(e);
    //    hr = newElement->QueryInterface(IID_IDOMElement, (void**)&domElement);
    //} else if (e->hasTagName(inputTag)) {
    //    DOMHTMLInputElement* newElement = new DOMHTMLInputElement(e);
    //    hr = newElement->QueryInterface(IID_IDOMElement, (void**)&domElement);
    //} else if (e->hasTagName(textareaTag)) {
    //    DOMHTMLTextAreaElement* newElement = new DOMHTMLTextAreaElement(e);
    //    hr = newElement->QueryInterface(IID_IDOMElement, (void**)&domElement);
    //if (e->isHTMLElement()) {
    //    DOMHTMLElement* newElement = new DOMHTMLElement(e);
    //    hr = newElement->QueryInterface(IID_IDOMElement, (void**)&domElement);
    //} else 
	
    pDOMElement = new DOMElement(pElement);
    
    return pDOMElement;
}

DOMElement::DOMElement( WebCore::Element* pElement )
: m_DOMNode( pElement ), m_pElement( pElement )
{
}

// DOMNode funtions
// Not to use multiple inheritance , use embeded DOMNode concept
HRESULT DOMElement::getValue( char* out_pszBuffer, int in_nBufferLen )
{
	return m_DOMNode.getValue( out_pszBuffer, in_nBufferLen );
}
HRESULT DOMElement::getName( char* out_pszBuffer, int in_nBufferLen )
{
	return m_DOMNode.getName( out_pszBuffer, in_nBufferLen );
}
HRESULT DOMElement::getNodetype( DomNodeType *out_type )
{
	return m_DOMNode.getNodetype( out_type);
}
HRESULT DOMElement::hasAttribute()
{
	return m_DOMNode.hasAttribute( );
}


/*
 * @desc gets the tag name
 * @param out_pszBuffer
 * @param in_nBufferLen
 * @return HRESULT TRUE/FALSE
 */
HRESULT DOMElement::getTagName( char* out_pszBuffer, int in_nBufferLen )
{
	out_pszBuffer[0] = 0;

    if (!m_pElement)
        return HRESULT_NOK;

    //WebCore::String nameString(in_pszName);
    WebCore::String attrValueString = (WebCore::String) m_pElement->tagName();
    
    if( in_nBufferLen <= attrValueString.length())
    {
			return HRESULT_NO_MEMORY;
    }
    else
    {
        AcStrcpy(out_pszBuffer, attrValueString.latin1().data());
		return HRESULT_OK;
    }
	
}
/*
 * @desc gets the DOM attribute
 * @param out_pszBuffer
 * @param in_nBufferLen
 * @param in_pszName
 * @return HRESULT TRUE/FALSE
 */
HRESULT DOMElement::getAttribute( char* out_pszBuffer, int in_nBufferLen, char* in_pszName )
{
	out_pszBuffer[0] = 0;

    if (!m_pElement)
        return HRESULT_NOK;

    WebCore::String nameString(in_pszName);
    WebCore::String& attrValueString = (WebCore::String&) m_pElement->getAttribute(nameString);
    
    if( in_nBufferLen <= attrValueString.length())
    {
			return HRESULT_NO_MEMORY;
    }
    else
    {
        AcStrcpy(out_pszBuffer, attrValueString.latin1().data());
		return HRESULT_OK;
    }
}
/*
 * @desc get IDOMAttr
 * @param in_pszName
 * @return HRESULT TRUE/FALSE
 */
HRESULT DOMElement::getAttributeNode( char* in_pszName, IDOMAttr **out_ppDOMAttr ) 
{

	SysNeverGetHere();
	return HRESULT_NOTSUPPORT;
}



} // namespace WebFrame